package de.jaetzold.art;

import java.beans.VetoableChangeListener;

/**	Classes that represent a output-device should implement this interface.
	Implementations of this interface represent specific interpretations of the value this Actuator accepts and delivers.
	*/
public interface Actuator extends Sensor {
	// TBD: think about this
	// public void connectWith(ActuatorPort port);
	
	/**	Set this Actuator to the value given as a parameter.
		If this Actuator is connected with an ActuatorPeer the Peer is informed about this, otherwise an Exception might be thrown.
		@exception de.jaetzold.util.NotConnectedException if this Actuator has no Peer or the Peer itself isn't connected (e.g. because the connection to the hardware went down)
		@param value the value this Actuator should have.
		@see #setValue(double,boolean)
		@see #setWaitForCompletion(boolean)
		*/
	public void setValue(double value);
	/**	Set this Actuator to the Value given as a parameter.
		If waitForCompletion is false this method may return before the new value has been sent to the hardware. This way the method could return almost immediatly and informing the peer and thus the hardware may happen asynchronous.
		If for some reason the value is not o.k. for this Actuator it is not required to accept the value and it is also not required to report that through an Exception.
		It may fail for example if one of the registered <code>VetoableChangeListener</code>-instances vetoes. In this case the method should return without changing anything.
		@exception de.jaetzold.util.NotConnectedException if this Actuator has no Peer or the Peer itself isn't connected (e.g. because the connection to the hardware went down)
		@param value the value this Actuator should have.
		@param waitForCompletion whether this method may set the value asynchronous.
		@see #setWaitForCompletion(boolean)
		@see #setValue(double)
		@see #isConnected()
		@see #addVetoableChangeListener(VetoableChangeListener)
		*/
	public void setValue(double value, boolean waitForCompletion);
	
	// whether setting the value should happen synchronous
	// @see ActuatorPeer#setValue(double,boolean)
	/**	Controls when state-changing Methods return.
		If waitMode is true, every Method which changes the state of the Actuator should per default wait until the update in the Hardware-Interface had happened. This does not block the Method from returning because of a timeout. Setting this to false means the calls can happen asynchronous (but retain order).
		@see #setValue(double)
		@see de.jaetzold.art.platform.ActuatorPeer#setValue(double,boolean)
		@param waitMode whether state-changing methods may return immediately
		*/
	public void setWaitForCompletion(boolean waitMode);
	/**	Returns the default for synchronous or asynchronous behavior of state-changing methods.
		@see #setWaitForCompletion(boolean)
		*/
	public boolean getWaitForCompletion();
	/**	Controls the behaviour of value-returning Methods. If it is true, every
		Method returns the value this Actuator wants the Hardware-Output to resemble.
		If it is false the value the Peer thinks this Hardware-Output really has is returned.
		@see #getValue()
		@param returnMode whether value-returning methods should assume that the value last set is the one the Actuator really has.
		*/
	public void setLocalReturn(boolean returnMode);
	/**	Returns the whether value-returning methods assume no one else is using this Actuator.
		@see #setLocalReturn(boolean)
		*/
	public boolean getLocalReturn();
	
	/**	For Listeners which can refuse some values to be set.
		This is for now not really used and tested in ART, but at least <code>RawActuator</code> makes takes these listeners into account.
		@see RawActuator#setValue(double)
		*/
	public void addVetoableChangeListener(VetoableChangeListener listener);
	/**	For Listeners which can refuse some values to be set.
		@see #addVetoableChangeListener(VetoableChangeListener)
		*/
	public void removeVetoableChangeListener(VetoableChangeListener listener);
}
