package de.jaetzold.art;

import de.jaetzold.art.platform.AngleSensorPeer;

// TBD: Is it better/sufficient not to provide an AngleSensor but an AngleOutgoingConversion?

// TBD: Is it better to move all special and conversioning-Stuff in a special Peer which accepts another peer as target and wraps the missing functionality?

// TBD: use Peer
// TBD: deliver Port with a real AngleSensorPeer
/**	A class for measuring an angle. This angle can be greater than 360 degree and is typically used for measuring the rotation-position of an axle. Negative Values are also possible.
The method <code>getValue()</code> returns the angle measured in whole turns.
For standard Fischertechnik and Lego an AngleSensor usually gets connected with a CountSensor because angles are measured by counting state-changes.
*/
public class AngleSensor extends CountSensor {
	/**	Constructs an AngleSensor whithone count being one revolution.
		*/
	public AngleSensor() {
		this(1);
	}
	/**	Constructs an AngleSensor whith a scale. The scale is <code>unitsPerTurn</code> counts making up one revolution.
		*/
	public AngleSensor(double unitsPerTurn) {
		this(unitsPerTurn, 1);
	}
	/**	Constructs an AngleSensor whith a scale. The scale is <code>unitsPerTurn/fraction</code> counts making up one revolution.
		*/
	public AngleSensor(double unitsPerTurn, double fraction) {
		setUnitsPerTurnFraction(unitsPerTurn, fraction);
	}
	
	/**	Converts a count-value to an angle according to its scale.
		*/
	public synchronized double convertUnitsToAngle(double units) {
		return (units*angleUnitsFraction)/angleUnits;
	}
	/**	Converts an angle to a count-value according to its scale.
		*/
	public synchronized int convertAngleToUnits(double angle) {
		return (int)Math.round((angle*angleUnits)/angleUnitsFraction);
	}
	
	// the Sensor internally stores the count because that is the most exact
	// value. This value is converted to turns only when asked for it.
	protected double convertToOutgoing(double value) {
		double units = super.convertToOutgoing(value);
		return convertUnitsToAngle(units);
	}
	
	/**	Calibrates this Sensor to be at the given angle.
		@see CountSensor#setCount(double)
		*/
	public void setAngle(double angle) {
		super.setCount(convertAngleToUnits(angle));
	}
	
	/** Returns the angle relative to the starting-position. The angle is computed as whole turns of 360 degree and a fraction of them (2*PI radians = 360 degree = 1 turn).
	*/
	public double getAngleTurns() {
		return getValue();
	}
	/** Returns the angle relative to the starting-position. The angle is measured in radians (2*PI radians = 360 degree = 1 turn).
	*/
	public double getAngleRadians() {
		return getAngleTurns()*2*Math.PI;
	}

	/** Returns the angle relative to the starting-position. The angle is measured in degrees (2*PI radians = 360 degree = 1 turn).
	*/
	public double getAngleDegree() {
		return getAngleTurns()*360;
	}
	
	/** The count-value is multiplied by <code>angleUnits/angleUnitsFraction</code> when computing the angle in whole turns.
		*/
	protected double angleUnits;
	/** The count-value is multiplied by <code>angleUnits/angleUnitsFraction</code> when computing the angle in whole turns.
		*/
	protected double angleUnitsFraction;
	/** Sets the Units in which this AngleSensor measures one turn. If the actual hardware-sensor for the angle is implemented by counting some changes in a state (e.g. standard LEGO or Fischertechnik), it would be best, to set this value equal to (or to a multiple of) the number of state-changes making a whole turn of 360 degree.
		If the hardware-sensor is implemented by some device which instead measures the angle in some analog form, the angleUnits value is also used to tell this Sensor which value makes up one revolution. But since it is more likely in this case, that one revolution is not measured by a whole number, it might be necessary to use another method: <code>setCountsPerTurnFraction(double, double)</code>.
		It is actually possible to set <code>countsPerTurn</code> to a negative value, but that would have the same effect as reversing the sensors value and is therefore not encouraged.
		@see RawSensor#setReversed(boolean)
		@see #setUnitsPerTurnFraction(double,double)
		*/
	public void setUnitsPerTurn(double unitsPerTurn) {
		setUnitsPerTurnFraction(unitsPerTurn, 1);
	}
	/**	Returns the count-units that make one revolution.
		@see #setUnitsPerTurn(double)
		*/
	public double getUnitsPerTurn() {
		return angleUnits/angleUnitsFraction;
	}
	/** Sets the Units in which this AngleSensor measures one turn. This method can set angleUnit sometimes more intuitively and precisely.
		@see #setUnitsPerTurn(double)
		*/
	public synchronized void setUnitsPerTurnFraction(double numerator, double denominator) {
		angleUnits = numerator;
		angleUnitsFraction = denominator;
	}
	/**	Returns the count-units that make one revolution.
		The value is returned as an array of double with the numerator at position 0 and the denominator at position 1.
		@see #getUnitsPerTurn()
		*/
	public double[] getUnitsPerTurnFraction() {
		return new double[]{angleUnits, angleUnitsFraction};
	}
	
	// TBD: combine that with the granularity of the port
	public double getPrecision() {
		return angleUnitsFraction/(2*angleUnits);
	}
	
	public boolean getDeliversFloatingPoint() {
		return true;
	}
}
