package de.jaetzold.art;

import de.jaetzold.util.BinaryOperator;
import de.jaetzold.util.NotConnectedException;

import java.beans.PropertyChangeListener;

public class CombinedSensor extends RawSensor {
	public CombinedSensor() {
	}
	public CombinedSensor(Sensor operand, BinaryOperator operator) {
		setOperand(operand);
		setOperator(operator);
	}
	
	private Sensor operand;
	/**	It's the second operand to operator, the first is this sensor's value itself. A CombinedSensor is best connected to the Port which genereates the most events and gets the Sensor as operator which generates less events.
		*/
	public void setOperand(Sensor operand) {
		Sensor localOperand = this.operand;
		if(localOperand != null) {
			localOperand.removeSensorListener(operandListener);
		}
		this.operand = operand;
		if(changes.hasListeners("dummyProperty") && operand != null) {
			operand.addSensorListener(operandListener);
		}
	}
	/** Beware, the used operand can change over time.
		*/
	public Sensor getOperand() {
		return operand;
	}
	
	private BinaryOperator operator;
	public void setOperator(BinaryOperator operator) {
		this.operator = operator;
	}
	public BinaryOperator getOperator() {
		return operator;
	}
	
	private SensorListener operandListener =
		new SensorListener() {
			public void processEvent(SensorEvent se) {
				// a call to "update()" was the first solution
				// but that didn't necessarily took the event's
				// value to combine and also update() can generate a new Event
				// which results in an update() ...(see BaseSensorPeer.update())
				double newCombinedValue;
				double oldCombinedValue;
				synchronized(operandEventMonitor) {
					operandEventValue = se.getValue();
					operandEvent = true;
					newCombinedValue = getValue();
					oldCombinedValue = lastProcessedEventValue;
					lastProcessedEventValue = newCombinedValue;
					operandEvent = false;
				}
				SensorEvent combinedEvent =
								new SensorEvent(CombinedSensor.this,
												se.getEventDescriptor(),
												se.getTimestamp(),
												oldCombinedValue,
												newCombinedValue,
												getDeliversFloatingPoint(),
												se.getPropagationId());
				CombinedSensor.this.processEvent(combinedEvent);
			}
		};
	
	public void addPropertyChangeListener(PropertyChangeListener l) {
		super.addPropertyChangeListener(l);
		Sensor localOperand = operand;
		if(changes.hasListeners("dummyProperty") && localOperand != null) {
			localOperand.addSensorListener(operandListener);
		}
	}
	public void removePropertyChangeListener(PropertyChangeListener l) {
		super.removePropertyChangeListener(l);
		Sensor localOperand = operand;
		if(!changes.hasListeners("dummyProperty") && localOperand != null) {
			localOperand.removeSensorListener(operandListener);
		}
	}
	
	private boolean operandEvent;
	private double operandEventValue;
	private Object operandEventMonitor = new Object();
	protected double convertToIncoming(double value)
	{
		double result = super.convertToIncoming(value);
		
		BinaryOperator localOperator = operator;
		Sensor localOperand = operand;
		if(localOperand == null || localOperator == null) {
			throw new NotConnectedException("Need operator and operand first.");
		}
		double operandValue;
		if(operandEvent) {
			// in case this incomingConversion is done because
			// the operand got an event
			operandValue = operandEventValue;
		} else {
			operandValue = localOperand.getValue();
		}
		result = localOperator.combine(result, operandValue);

		return result;
	}
	
	// beware, an Event from 1->2 and one from 2->3 can come in any order.
	// but the synchronized should make sure that if current is 1
	// and there comes an event from 2->3, the event from 1->2 will still
	// come, without being "garbled" to be one from 2->2 or even 3->2
	private double lastProcessedEventValue;
	public void processEvent(SensorEvent se) {
		SensorEvent newEvent = se;
		synchronized(operandEventMonitor) {
			if(!operandEvent) {
				// since RawSensor takes the cached value as the old one
				// it maybe has to be changed here.
				// TBD: provide a method in RawSensor for retrieving the old
				// Event-Value which can then be overriden by subclasses?
				if(lastProcessedEventValue != se.getOldValue()) {
					newEvent = new SensorEvent(se.getSource(),
												se.getEventDescriptor(),
												se.getTimestamp(),
												lastProcessedEventValue,
												se.getValue(),
												se.isFloatingPoint(),
												se.getPropagationId());
				}
				lastProcessedEventValue = se.getValue();
			}
		}
		super.processEvent(newEvent);
	}
	
	public double getPrecision() {
		double result = super.getPrecision();
		if(!precisionSet) {
			try {
				result += operand.getPrecision();
			} catch(Exception e) {
			}
		}
		return result;
	}
	
	public boolean isIncomingConversionCacheable() {
		// important, because the incoming conversion
		// takes the operand's value into account
		return false;
	}
	
	public boolean isConnected() {
		Sensor localOperand = operand;
		if(localOperand != null) {
			return operator != null
					&& super.isConnected()
					&& localOperand.isConnected();
		} else {
			return false;
		}
	}
	
	public boolean getDeliversFloatingPoint() {
		// because we don't know what the operator does:
		return true;
	}
}
