package de.jaetzold.art;

import de.jaetzold.art.platform.CountSensorPeer;
import de.jaetzold.art.platform.ForwardingSensorPeer;
import de.jaetzold.art.platform.SensorPeer;
import de.jaetzold.art.platform.BaseSensorPort;

import de.jaetzold.util.StateDecider;
import de.jaetzold.util.StateDifferenceStateDecider;

// TBD: change StateDeciders to BinaryOperators?
// maybe getDefaultStateDecider?
// TBD: validate concept for delivering class-specific values (as in getCountValue())

/**	This class represents a Sensor which counts state-transitions. For associating state-transitions with a value to count by itself (without a <code>CountSensorPeer</code>) it can have a <code>StateDecider</code>.
	@see de.jaetzold.util.StateDecider
	*/
public class CountSensor extends StateSensor {
	private double countValue;
	private double lastState;
	private synchronized double computeCountFromState(double state)
	{
		if(countSensorPeer != null) {
			countValue = state;
		} else {
			StateDecider localDecider = countDecider;
			if(localDecider == null) {
				localDecider = defaultCountDecider;
			}
			if(isReversed()) {
				countValue -= localDecider.nextStateValue(lastState, state);
			} else {
				countValue += localDecider.nextStateValue(lastState, state);
			}
			lastState = state;
		}
		return countValue;
	}
		
	protected double convertToIncoming(double value) {
		double state = super.convertToIncoming(value);
		return computeCountFromState(state);
	}
	
	// Note: overriding the normal getValue() isn't necessary because
	// convertToIncoming(double) does return the count-value.
	/**	Returns the count-value this <code>CountSensor</code> currently has.
		*/
	public double getCountValue() {
		getValue();	// update countValue
		return countValue;
	}

	/**	Sets the count-value this <code>CountSensor</code> currently has. This way it can be calibrated to a specific point to count from.
		@see #resetCount()
		*/
	public void setCount(double count) {
		CountSensorPeer localPeer = countSensorPeer;
		if(localPeer != null) {
			localPeer.setCount(isReversed() ? count*-1 : count);
		} else {
			double oldCount = countValue;
			countValue = count;
			if(oldCount != count) {
				update();
			}
		}
	}

	/**	Resets the count-value this <code>CountSensor</code> currently has to 0.
		@see #setCount(double)
		*/
	public void resetCount() {
		setCount(0);
	}
	
	// beware for countValue when reconnecting
	private CountSensorPeer countSensorPeer;
	public synchronized void connectWith(SensorPort port) {
		super.connectWith(port);
		if(getPeer() instanceof CountSensorPeer) {
			countSensorPeer = (CountSensorPeer)getPeer();
		} else {
			// should'nt be necessary
			countSensorPeer = null;
		}
	}
	
	public synchronized void disconnect() {
		countSensorPeer = null;
		super.disconnect();
	}
		
	public boolean getDeliversFloatingPoint() {
		if(countSensorPeer != null) {
			return super.getDeliversFloatingPoint();
		} else {
			// as long as the countDecider can't be asked for 
			return true;
		}
	}
	
	public boolean isIncomingConversionCacheable() {
		StateDecider localDecider = countDecider;
		if(countSensorPeer != null) {
			return super.isIncomingConversionCacheable();
		} else if(localDecider != null) {
			return localDecider.isCacheable()
					&& super.isIncomingConversionCacheable();
		} else {
			return defaultCountDecider.isCacheable()
					&& super.isIncomingConversionCacheable();
		}
	}
	
	// TBD: maybe instead of having an own decider use and set the one in the superclass
	private StateDecider countDecider;
	private StateDecider defaultCountDecider =
									new AbsoluteStateDifferenceStateDecider();
	
	/**	Sets the <code>StateDecider</code> used to associate state-transitions with a value to count by ourselves.
		If there is a new value from the peer the given decider is used to calculate the value which should be added to the current count-value. The method <code>StateDecider.nextStateValue(double,double)</code> is called with the last state as the first argument and the new state as the second. Note: states are computed by the superclass.
		<p>
		Beware, as long as the peer is a <code>CountSensorPeer</code> the decider isn't called.
		@see de.jaetzold.art.platform.CountSensorPeer
		@see StateSensor#setStateDecider(StateDecider)
		*/
	public void setCountDecider(StateDecider countDecider) {
		this.countDecider = countDecider;
	}
	
	/**	This is the <code>StateDecider</code> used as the default in <code>CountSensor</code>.
		It returns the absolute value fo the result of it's superclass <code>StateDifferenceStateDecider</code>.
		*/
	public static class AbsoluteStateDifferenceStateDecider
			extends StateDifferenceStateDecider {
		public double nextStateValue(double lastState, double newState) {
			double result = Math.abs(super.nextStateValue(lastState, newState));
			return result;
		}
	}
	
	// A special Peer is required so that getValue() really delivers the count
	private SensorPeer countSensorPortPeer =
		new ForwardingSensorPeer(this) {
			private double oldCount;
			public double getValue() {
				oldCount = getCountValue();
				return oldCount;
			}
			
			public void processEvent(SensorEvent se) {
				SensorEvent countEvent = se;
				if(se.getSource() != this) { // otherwise from update()
					countEvent =
						new SensorEvent(this,
										se.getEventDescriptor(),
										se.getTimestamp(),
										oldCount,
										getCountValue(),
										this.getDeliversFloatingPoint(),
										se.getPropagationId());
				}
				oldCount = se.getNewValue();
				super.processEvent(countEvent);
			}
		};
	
	/**	A SensorPort for the count-value.
		This is only useful if a subclass affects the value the <code>getValue()</code>-method returns. This way the original count-value can always be acessed.
		@see RawSensor.ValueSensorPort
		@see #getCountSensorPort()
		*/
	public class CountSensorPort extends BaseSensorPort {
		public CountSensorPort() {
		}
		public SensorPeer getPeerFor(Sensor sensor) {
			return countSensorPortPeer;
		}
	}
	
	/**	Returns a SensorPort for the count-value.
		@see RawSensor#getSensorPort()
		@see CountSensor.CountSensorPort
		*/
	public CountSensorPort getCountSensorPort() {
		return new CountSensorPort();
	}
}
