package de.jaetzold.art;

import de.jaetzold.art.platform.LightSensorPeer;

import de.jaetzold.util.NotConnectedException;
import de.jaetzold.util.Conversion;

// TBD: what if not the LightSensor can scale, but may be connected to a Port which can scale?

/**	This class represents values which are associated with some kind of light-level. It is a sensor for a particular unit of measure.
	*/
public class LightSensor extends RawSensor {
	private Conversion scalingConversion;
	/** This is a delegate used for scale-conversion.
		*/
	public void setScalingConversion(Conversion scalingConversion) {
		this.scalingConversion = scalingConversion;
	}
	/** Returns the delegate used for scale-conversion.
		*/
	public Conversion getScalingConversion() {
		return scalingConversion;
	}
	
	protected double convertToIncoming(double value)
	{
		double result = super.convertToIncoming(value);
		
		// compute the lightLevel by ourselves if the peer doesn't
		if(lightSensorPeer == null) {
			Conversion localConversion = scalingConversion;
			if(localConversion != null) {
				result = localConversion.convert(result);
			} else {
				throw new NotConnectedException("Need either a LightSensorPeer or a scaling Conversion.");
			}
		}
		
		return result;
	}
	
	private double measurableMin;
	/** This is the value below which the sensor isn't accurate enough anymore.
		If the peer is a <code>LightSensorPeer</code> the value is taken from it until it is set explicitly.
		@see #setMeasurableMin(double)
		@see #getPrecision()
		*/
	public double getMeasurableMin() {
		LightSensorPeer localPeer = lightSensorPeer;
		if(localPeer != null) {
			return localPeer.getMeasurableMin();
		}
		return measurableMin;
	}
	/** This sets the value below which the sensor isn't accurate enough anymore.
		@see #getPrecision()
		*/
	public void setMeasurableMin(double min) {
		this.measurableMin = min;
	}
	
	private double measurableMax;
	/** This is the value above which the sensor isn't accurate enough anymore.
		If the peer is a <code>LightSensorPeer</code> the value is taken from it otherwise it must be set explicitly.
		@see #setMeasurableMax(double)
		@see #getPrecision()
		*/
	public double getMeasurableMax() {
		LightSensorPeer localPeer = lightSensorPeer;
		if(localPeer != null) {
			return localPeer.getMeasurableMax();
		}
		return measurableMax;
	}
	/** This sets the value above which the sensor isn't accurate enough anymore.
		@see #getPrecision()
		*/
	public void setMeasurableMax(double max) {
		this.measurableMax = max;
	}

	private LightSensorPeer lightSensorPeer;
	public synchronized void connectWith(SensorPort port) {
		super.connectWith(port);
		if(getPeer() instanceof LightSensorPeer) {
			lightSensorPeer = (LightSensorPeer)getPeer();
		} else {
			lightSensorPeer = null;
		}
	}
	
	public synchronized void disconnect() {
		lightSensorPeer = null;
		super.disconnect();
	}
	// this sensor only has int as value, but let subclasses
	// and outgoing delegates do what they want (obsolete)
	public boolean getDeliversFloatingPoint() {
		boolean result;
		Conversion localOutgoingConversion = getOutgoingConversionDelegate();

		if(localOutgoingConversion == null) {
			Conversion localScalingConversion = scalingConversion;
			if(lightSensorPeer == null && localScalingConversion != null) {
				result = localScalingConversion.getDeliversFloatingPoint();
			} else {
				// that asks the peer
				result = super.getDeliversFloatingPoint();
			}
		} else {
			result = localOutgoingConversion.getDeliversFloatingPoint();
		}

		return result;
	}
	
	public boolean isIncomingConversionCacheable() {
		boolean result = super.isIncomingConversionCacheable();
		
		Conversion localConversion = scalingConversion;
		if(lightSensorPeer == null && localConversion != null) {
			result = localConversion.isCacheable() && result;
		}
		return result;
	}
}
