package de.jaetzold.art;

import de.jaetzold.art.platform.MotorPeer;

import de.jaetzold.util.Debug;

// TBD: MotorListener which gets informed seperately about on, power, backward
//      (and the corresponding properties ...)
/**	A class which represents a simple Motor. 
	It can be switched on or off and the direction, as well as the level of the
	power supplied to the motor can be set.
	This particular implementation can just be used as some kind of Actuator-frontend, because it can give access to all of its abilities when only connected to a simple ActuatorPeer. This class provides some State and Methods which are more convenient to deal with, than with only a single value.
	However, an Interface can provide a special MotorPeer for this type of Motor, but a normal ActuatorPeer should be sufficient.
	@see de.jaetzold.art.platform.ActuatorPeer
	@see de.jaetzold.art.platform.MotorPeer
	@see Actuator
	*/
public class Motor extends RawActuator {	
	private Debug debug = new Debug("de.jaetzold.art.Motor");
	{ if(debug.debug) debug.objectIdentifier = debug.getIdentifierFor(this); }
	
	private double computeSingleValue() {
		return on?(power)*(backward?-1:1):0;
	}
	/**	Computes the power, direction and the on/off state from the given value.
		A positive value means forward, a negative backward, each with the power set to abs(value). If, and only if a value of zero is given the motor is switched off (retaining any previous value for power and backward), otherwise on. This Method calls setPower(double), setBackward(boolean) and setOn(boolean) with the extracted values in that order.
		@see #setPower(double)
		@see #setBackward(boolean)
		@see #setOn(boolean)
		*/
	public synchronized void setValue(double value) {
		if(motorPeer != null) {
			// when the peer changes in-between that's bad luck
			// (and one couldn't expect proper values anyway)
			super.setValue(value);
		} else {
			if(value != 0) {
				double newPower = Math.abs(value);
				if(power != newPower) {
					setPower(newPower);
				}
				if((value<0 && !backward)
						|| (value>0 && backward)) {
					reverse();
				}
				on();
			} else {
				off();
			}
			if(debug.debug) {
				if(computeSingleValue() != value) {
					debug.printError("Error in computing Motor-state to or from a single value: "
										+computeSingleValue() +"!=" +value);
				}
			}
		}
		// the calls to setPower(), reverse() and on/off()
		// already call super.setValue()
	}
	
	/**	The Peer for this Motor if it is a MotorPeer. If it is only an ActuatorPeer a Motor uses it's Peer only by calling methods in the superclass RawActuator.
		@see RawActuator#actuatorPeer
		*/
	private MotorPeer motorPeer;
	/** Connects this Object with the given Port.
		This sets the peer for this Object and is therefore required before
		it can be properly used.
		*/
	public synchronized void connectWith(ActuatorPort port) {
		super.connectWith(port);
		if(getPeer() instanceof MotorPeer) {
			motorPeer = (MotorPeer)getPeer();
		} else {
			motorPeer = null;
		}
	}
	public synchronized void disconnect() {
		motorPeer = null;
		super.disconnect();
	}
	
	/**	The power that the output should supply to the motor if it is on. */
	protected double power = 1;
	/**	Returns the power the output does or should supply to the motor.
		Which of these values is returned is influenced by <code>localReturnMode</code>.
		@see #localReturn
		*/
	public double getPower() {
		if(localReturn) {
			return power;
		} else {
			MotorPeer localPeer = motorPeer;
			if(localPeer != null) {
				return localPeer.getPower();
			} else {
				double isValue = getValue();
				return isValue!=0?Math.abs(isValue):power;
			}
		}
	}
	/**	Sets the power the output should supply to the motor to the given value.
		This usually has an effect to the speed of the motor, but it's real purpose is more to control the torque.
		Setting the power to 0 has the effect that the real Motor turns off, but it is still considered to be in the on-state.
		Dependig on the value of <code>waitForCompletion</code> this Method returns immediately after it has been registerd that this is now the state the motor should have, or it returns after the state has been set.
		@see #waitForCompletion
		*/
	public synchronized void setPower(double power) {
		this.power = Math.max(0.0, power);
		if(debug.debug) {
			debug.printInfo("Got power " +power +", setting to " +this.power);
		}
		if(motorPeer != null) {
			motorPeer.setPower(this.power);
		} else {
			// _must_ be super because this.setValue() might in turn call setPower()
			super.setValue(computeSingleValue());
		}
	}
	
	/**	Whether the output should reverse its polarity when supplying
		power to the motor.
		*/
	protected boolean backward = false;
	/**	Returns whether the polarity the output does or should supply to the motor
		will or should be reversed or not.
		Which of these values is returned is influenced by <code>localReturnMode</code>.
		@see #localReturn
		*/
	public boolean getBackward() {
		if(localReturn) {
			return backward;
		} else {
			MotorPeer localPeer = motorPeer;
			if(localPeer != null) {
				return localPeer.getBackward();
			} else {
				double isValue = getValue();
				return isValue!=0?isValue<0:backward;
			}
		}
	}
	/**	Controls whether the output should reverse its default polarity.
		Dependig on the value of <code>waitMode</code> this Method returns
		immediately after it has been registerd that this is now the state the
		motor should have, or it returns after the state has been set.
		@see #waitForCompletion
		*/
	public synchronized void setBackward(boolean backward) {
		this.backward = backward;
		// should update setPowerPercent() ?
		if(motorPeer != null) {
			motorPeer.setBackward(this.backward);
		} else {
			// _must_ be super because this.setValue() might in turn call setBackward()
			super.setValue(computeSingleValue());
		}
	}

	/** Whether the output should be supplied with any power at all. */
	protected boolean on = true;
	/**	Returns whether the motor should be or is supplied with power.
		Which of these values is returned is influenced by <code>localReturnMode</code>.
		@see #localReturn
		*/
	public boolean isOn() {
		if(localReturn) {
			return on;
		} else {
			MotorPeer localPeer = motorPeer;
			if(localPeer != null) {
				return localPeer.isOn();
			} else {
				double isValue = getValue();
				return isValue!=0;
			}
		}
	}
	/**	Turns the motor on or of depending on the given value.
		Dependig on the value of <code>waitMode</code> this Method returns
		immediately after it has been registerd that this is now the state the
		motor should have, or it returns after the state has been set.
		@see #waitForCompletion
		*/
	public synchronized void setOn(boolean on) {
		this.on = on;
		if(motorPeer != null) {
			motorPeer.setOn(this.on);
		} else {
			// _must_ be super because this.setValue() might in turn call setOn()
			super.setValue(computeSingleValue());
		}
	}
	
	// some more telling methods
	/** Sets the direction of the motor to the opposite of its default-direction. 
		@see #setBackward(boolean)
		*/
	public void backward() {
		setBackward(true);
	}
	/** Sets the direction of the motor to its default-direction. 
		@see #setBackward(boolean)
		*/
	public void forward() {
		setBackward(false);
	}
	/**	Switches the Motor on.
		@see #setOn(boolean)
		*/
	public void on() {
		setOn(true);
	}
	/**	Switches the Motor off.
		@see #setOn(boolean)
		*/
	public void off() {
		setOn(false);
	}
	/**	Reverse the actual direction of the motor, whatever it is now.
		Because this Method depends on the actual state of the Motor it depends on the value of localReturnMode. If localReturnMode is set to false it is possible e.g. that two calls to reverse() have the same effect as one.
		This can happen if the value from the first call still hasn't been set in the hardware-interface when the second call arrives.
		*/
	public void reverse() {
		setBackward(!getBackward());
	}
}
