package de.jaetzold.art;

// 1. idea: Ports are available for any data in any special format. which data and which format is selected by connecting to the right port.
// 2. idea: Ports are available for any type of data. The format is selected by what Sensor/Actuator gets connected. To connect a Sensor/Actuator to a different than its associated format it has to be connected afterwards.
/* pro 1:
which data in which format is more directly and transparent to the user
/* pro 2:
offers more flexibility when creating new sensor/actuator-types
*/
// so far: implemented idea 2
/** A Port describes what data in which format an interface can deliver.
	A Port usually knows how to create a Peer for Objects if they are needed but this interface leaves it open which type the Peer might be and which method should be used for it. Subclasses can specify which types of Peers they create under which circumstances. A usual Port consists therefore mainly of a method for retrieving a peer associated with the port, methods which give information about the capabilities and attributes of the port as well as methods for connecting and disconnecting other Objects.
	*/
public interface Port {
	/**	Returns a numeric id for this Port which is maybe easier to interpret by code than the result of <code>getPortID()</code>.
		@see #getPortID()
		*/
	public int getPortNumber();
	/**	Returns some Object describing this instance of a Port.
		It describes the Port in the sense that a call to <code>conformsTo(Object)</code> with it as a parameter will yield true as a result.The current implementations use a String which is equal to the names printed on the hardware-interface or used in its manual.
		@see #getPortNumber()
		@see #conformsTo(Object)
		@see RobotInterface#getPort(Object)
		*/
	public Object getPortID();
	/**	Returns an Array consisting of all Ports which cannot be used together
		with this port at a given time.
		*/
		/* That means if more than one of these Ports should be connected, only one connection can be enabled at the same time.
		TBD: is it possible to enable/disable Ports directly? can they forward this information to any connected Sensors/Actuators?
		@see Sensor#enable()
		@see Sensor#disable()
		@see Actuator#enable()
		@see Actuator#disable()
		*/
	public Port[] getExcludedPorts();
	/**	Returns the minimum value this port can really put out or read in. 
		If Port is an ActuatorPort any value less than that will produce the same output as the value returned by this method. If this Port is a SensorPort then this Method returns the minimum Value you could expect to get from this Port. This value is only correct if the Sensor connected to this Port gets no special type of Peer.
		*/
	public double getMinimumValue();
	/**	Returns the maximum value this port can really put out or read in. 
		If Port is an ActuatorPort any value greater than that will produce the same output as the value returned by this method. If this Port is a SensorPort then this Method returns the maximum Value you could expect to get from this Port. This value is only correct if the Sensor connected to this Port gets no special type of Peer.
		*/
	public double getMaximumValue();
	/**	Returns the number of different values which this port can put out or read in. That means this value will always be greater than 1, with a value of 2 meaning that an actuator for this port can only be switched on or off and a Sensor will only deliver two different values. This value is only correct if the Sensor connected to this Port gets no special type of Peer.
		*/
	public long getGranularity();
	
	/** Tells whether the given Object somehow identifies this Port.
		The provided implementations for Fischertechnik and Mindstorms use Strings which are equal to the Names printed on the hardware-interface to identify Ports.
		*/
	public boolean conformsTo(Object identifier);
}
