package de.jaetzold.art;

import de.jaetzold.art.platform.ActuatorPeer;

import de.jaetzold.util.Conversion;
import de.jaetzold.util.Debug;

import java.util.Hashtable;
import java.util.Enumeration;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeEvent;
import java.beans.VetoableChangeListener;
import java.beans.VetoableChangeSupport;
import java.beans.PropertyVetoException;

// TBD: maybe rename
// convertToIncoming			-> convertPeerToInternal
// convertToOutgoing			-> convertInternalToExternal
// convertToActuatorIncoming	-> convertExternalToInternal
// convertToActuatorOutgoing	-> convertInternalToPeer

/**	This class represents a basic output-device of a hardware-interface.
	Each output device, such as an Actuator or a Motor needs to be connected to
	an ActuatorPort first. Through connecting with a Port this Object gets its
	peer which is required for proper use.
	@see #connectWith(ActuatorPort)
	*/
public class RawActuator extends RawSensor implements Actuator {
	private Debug debug = new Debug("de.jaetzold.art.RawActuator");
	{ if(debug.debug) debug.objectIdentifier = debug.getIdentifierFor(this); }

	protected double convertToActuatorOutgoing(double internalValue) {
		double result = reverseConversion(internalValue);
		if(debug.debug) {
			debug.printInfo("convertToActuatorOutgoing(" +internalValue +")"
						+(reversed ? ", reversed: " +(result) : ""));
		}
		Conversion localConversion = actuatorOutgoingConversionDelegate;
		if(localConversion != null) {
			result = localConversion.convert(result);
		}
		return result;
	}
	
	protected double convertToActuatorIncoming(double externalValue) {
		double result = externalValue;
		Conversion localConversion = actuatorIncomingConversionDelegate;
		if(localConversion != null) {
			result = localConversion.convert(result);
		}
		return result;
	}
	
	private Conversion actuatorIncomingConversionDelegate;
	public void setActuatorIncomingConversionDelegate(Conversion delegate) {
		this.actuatorIncomingConversionDelegate = delegate;
	}
	public Conversion getActuatorIncomingConversionDelegate() {
		return actuatorIncomingConversionDelegate;
	}
	private Conversion actuatorOutgoingConversionDelegate;
	public void setActuatorOutgoingConversionDelegate(Conversion delegate) {
		this.actuatorOutgoingConversionDelegate = delegate;
	}
	public Conversion getActuatorOutgoingConversionDelegate() {
		return actuatorOutgoingConversionDelegate;
	}

	private double localValue;
	public void setValue(double value) {
		setValue(value, waitForCompletion);
	}
	public synchronized void setValue(double value, boolean waitMode) {
		double oldValue = this.localValue;
		double newValue = convertToActuatorIncoming(value);
		// maybe the connected port isn't capable of reversing the polarity
		// (we have no listener by now)
		try {
			vetos.fireVetoableChange("value",
                    			   new Double(oldValue),
                    			   new Double(newValue));
		} catch(PropertyVetoException pve) {
			if(debug.debug) {
				debug.printInfo("setValue(" +value +") has been vetoed: " +pve);
			}
			return;
		}
		// No-one vetoed, so go ahead and make the change.

		this.localValue = newValue;
		
		if(actuatorPeer != null) {
			actuatorPeer.setValue(convertToActuatorOutgoing(this.localValue),
								waitMode);
		}
		// no NotConnectedException necessary here, because
		// an Actuator isn't required to set the value
		// if one wants to be sure: test with getValue()
		// (with localReturn=false)
		if(localReturn && oldValue != newValue) {
			processEvent(new SensorEvent(this,
										propertyChangeEventDescriptor,
										System.currentTimeMillis(),
										oldValue,
										newValue,
										true,
										null));
		}
	}

	// maybe add another property like remoteValue
	// and make Motors PropertyChangeListener of their peers as well
	// if a remoteValue changes the Motor is informed and sets its
	// own copy of remoteValue and then informs all his PropertyListeners
	// of that change.

	public double getValue() {
		if(localReturn) {
			return localValue;
		} else {
			return super.getValue();
		}
	}
	
	protected VetoableChangeSupport vetos =
				new VetoableChangeSupport(this); 
	public void addVetoableChangeListener(VetoableChangeListener l) {
		vetos.addVetoableChangeListener(l);
	}
	public void removeVetoableChangeListener(VetoableChangeListener l) {
		vetos.removeVetoableChangeListener(l);
	}
	
	private ActuatorPeer actuatorPeer;
	public synchronized void disconnect() {
		actuatorPeer = null;
		super.disconnect();
	}
	
	/** Connects this Object with the given Port.
		This sets the peer for this Object and is therefore required before
		it can be properly used.
		*/
	public synchronized void connectWith(ActuatorPort port) {
		super.connectWith(port);
		if(getPeer() instanceof ActuatorPeer) {
			actuatorPeer = (ActuatorPeer)getPeer();
		} else {
			disconnect();
			throw new IllegalArgumentException("No listen-only Actuators yet.");
			//actuatorPeer = null;
		}
	}
	
	public void connectWith(SensorPort port) {
		if(port instanceof ActuatorPort) {
			if(debug.debug) {
				debug.printInfo("Got port as SensorPort which is instance of ActuatorPort, connecting ...");
			}
			connectWith((ActuatorPort) port);
		} else {
			throw new IllegalArgumentException("An ActuatorPort is required for connectWith()");
		}
	}
	
	private boolean reversed;
	/**	Controls whether this Actuator should treat negative values as positive and the other way round. That way this state defines what "+" for the algorithm means for the robot.
		*/
	public void setReversed(boolean reversed) {
		this.reversed = reversed;
		setValue(this.localValue);
		
		super.setReversed(reversed);
	}

// The following could be useful some time for subsumption
/*	/**	Controls whether this Actuators state is considered important or not.
		That means, if importantMode is false, this Actuators state will never be
		reflected in the Robot if there is another Actuator for the same output with
		importantMode set to true.
		*/
/*	protected boolean importantMode = true;
	public boolean getImportantMode() {
		return importantMode;
	}
	public void setImportantMode(boolean importantMode) {
		this.importantMode = importantMode;
	}*/
	/**	This is true by default.
		@see #setWaitForCompletion(boolean)
		*/
	protected boolean waitForCompletion = true;
	public boolean getWaitForCompletion() {
		return waitForCompletion;
	}
	public void setWaitForCompletion(boolean waitMode) {
		this.waitForCompletion = waitMode;
	}

	/**	This is false by default.
		@see #setLocalReturn(boolean)
		*/
	protected boolean localReturn = false;
	public boolean getLocalReturn() {
		return localReturn;
	}
	public void setLocalReturn(boolean returnMode) {
		this.localReturn = returnMode;
	}
}
