package de.jaetzold.art;

/**	Basic Java-interface for classes which represent a hardware-interface in ART.
	Instances of RobotInterface can be retrieved from a RobotInterfaceFactory
	@see RobotInterfaceFactory
	*/
public interface RobotInterface {
	/**	Returns a RobotInterfaceDefinition describing this instance of a RobotInterface.
		It describes the RobotInterface in the sense that a call to <code>conformsTo(RobotInterfaceDefinition)</code> with it as a parameter will yield true as a result. There should be no other instances of RobotInterface at a time which conform to this definition. However there may other definitions possible to which more than one RobotInterface conforms to, but they shouldn't be returned by this method. The current implementations use a RobotInterfaceStringDefinition.
		@see RobotInterfaceStringDefinition
		@see #conformsTo(RobotInterfaceDefinition)
		@see RobotInterfaceFactory#getInterface(RobotInterfaceDefinition)
		*/
	public RobotInterfaceDefinition getInterfaceDefinition();
	/** Tells whether the given RobotInterfaceDefinition identifies this RobotInterface.
		The provided implementations for Fischertechnik and Mindstorms use Strings.
		@see RobotInterfaceStringDefinition
		@see #getInterfaceDefinition()
		@see RobotInterfaceFactory#getInterface(RobotInterfaceDefinition)
		*/
	public boolean conformsTo(RobotInterfaceDefinition definition);
	/**	Tells the state of the physical connection to the hardware-interface. It may be the case that this state is not distinguishable from the one defined by <code>isAlive()</code>.
		@see #isAlive()
		@see de.jaetzold.art.platform.SensorPeer#isConnected()
		@see de.jaetzold.art.Sensor#isConnected()
		*/
	public boolean isConnected();
	/**	Tells the state of the logical connection to the hardware-interface. If it is true that means the hardware-interface still behaves as expected.
		@see #isConnected()
		*/
	public boolean isAlive();
	// maybe both through definition of CONNECTION_STATE's
	/** Returns a description of all ActuatorPorts this Interface provides.
		The Ports in the returned Array can then be used to connect some Actuator- or Sensor-Objects with a Peer that represents some specific type of hardware connected to the hardware-interface.
		@see #getSensorPorts()
		@see ActuatorPort#connectWith(Actuator)
	*/
	public ActuatorPort[] getActuatorPorts();

	/** Returns the SensorPort with the specified index.
		This method is only for convenience when using a RobotInterface as a JavaBean.
		@see #getActuatorPorts()
		@throws ArrayIndexOutOfBoundsException if the specified index is not a legal index into the array of SensorPorts.
		*/
	public ActuatorPort getActuatorPorts(int index);
		
	/** Returns a description of all SensorPorts this Interface provides. The
	description consists of the portnumber and some other information describing
	the capabilities of the sensorport. (pollingfreq., settable pollingfreq.?,
	hw-events?, scaling (which voltage/resistance for which value),
	maximum-value/voltage?, Powered?, counting?, excludes use of other
	sensors?, asf.).
	*/
	/** Returns a description of all SensorPorts this Interface provides.
		The Ports in the returned Array can then be used to connect some Sensor-Objects with a Peer that represents some specific type of hardware connected to the hardware-interface.
		@see #getActuatorPorts()
		@see SensorPort#connectWith(Sensor)
	*/
	public SensorPort[] getSensorPorts();
	
	/** Returns the SensorPort with the specified index.
		This method is only for convenience when using a RobotInterface as a JavaBean.
		@see #getActuatorPorts()
		@exception ArrayIndexOutOfBoundsException if the specified index is not a legal index into the array of SensorPorts.
		*/
	public SensorPort getSensorPorts(int index);
	
	/** Returns a Port which is identified by the given Object.
		Which identifiers are legal and their meaning is defined by the actual implementation of RobotInterface. The implementations provided for Fischertechnik and Mindstorms use Strings which are equal to the names printed on the hardware-interface.
		@see #getSensorPorts()
		@see #getActuatorPorts()
		@see Port#getPortID()
		@see Port#conformsTo(Object)
		*/
	public Port getPort(Object identifier);
}
