package de.jaetzold.art;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Properties;
import java.util.StringTokenizer;
import java.io.FileInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import de.jaetzold.util.Debug;
import de.jaetzold.util.NotConnectedException;

// TBD: add a method for "stopping" the interface.
// 1.: the interface can "exit" cleanly
// 2.: if the interface creates a non-daemon-thread it can at least
// keep the program running until a system.exit
// 3.: creating this thread can't be prevented in AWTInterface either - obsolete

// TBD: maybe create a new Factory with the InterfacePort parameters ...

/**	This is the class for creating instances of <code>RobotInterface</code>.
	To use it you will have to construct a new instance of <code>RobotInterfaceFactory</code> and then use any of the provided <code>getInterface(...)</code>-Methods for retrieving instances of <code>RobotInterface</code>.
	<p>
	Note: This class depends on a whitespace-separated list of fully qualified class-names given as the value of the property <code>de.jaetzold.art.RobotInterfaceFactory.platforms</code>
	<p> The property's value is taken from the System-properties and if it isn't defined there from the file <code>de.jaetzold.art.properties</code>
	@see RobotInterfaceDefinition
	@see RobotInterfaceStringDefinition
	@see RobotInterface
	*/
public class RobotInterfaceFactory {
	private static Debug debug =
		new Debug("de.jaetzold.art.RobotInterfaceFactory");
	// implement some stuff for having more factorys
	// RobotInterface-Implementation have to take care of the autodetect-stuff
	// Maybe let all RobotInterfaceFactorys use the same Array of
	// detected Interfaces? done, as the RobotInterface-Implementations are now
	// required to take care that they can deliver the instances more than once.
	/**	Contruct a new Factory for <code>RobotInterface</code>s.
		*/
	public RobotInterfaceFactory() {
	}
	
	/**	JavaBeans indexed property.
		@see #getInterfaces()
		*/
	public RobotInterface getInterfaces(int index) {
		return getInterfaces()[index];
	}
	/**	Automatically scans all known Ports for all known types of Interfaces.
		'known' means the ones given in the property <code>de.jaetzold.art.RobotInterfaceFactory.platforms</code>
		<p>
		If no interface could be found an array of length 0 is returned.
		@see #getInterfaces(RobotInterfaceDefinition[])
		*/
	// Check Classname on entry, to avoid problems with 'super.getInterfaces()'
	// Maybe also check whether a Thread is already executing this mehtod and do not allow any re-entrees, even by the same Thread because of possible recursions?
	public RobotInterface[] getInterfaces() {
		// null creates ALL which is an id which all implementations (can) accept.
		RobotInterfaceDefinition[] ports = 
				new RobotInterfaceDefinition[]{
						new RobotInterfaceStringDefinition(null)};
		return getInterfaces(ports);
	}
	/** Scans for <code>RobotInterface</code>s only for the given Port. If no interface could be found an array of length 0 is returned.
		@see #getInterfaces(RobotInterfaceDefinition[])
		*/
	public RobotInterface[] getInterfaces(RobotInterfaceDefinition port) {
		return getInterfaces(new RobotInterfaceDefinition[]{port});
	}
	
	/** Scans for <code>RobotInterface</code>s only for the given Ports.
		<code>RobotInterface</code>-classes have to implement a method with this exact same signature (and a compatible return-type) in order to provide automatic detection capability through <code>RobotInterfaceFactory</code>.
		<p>
		This method simply collects the <code>RobotInterface</code>s returned by these methods. In each call to this method the known classes are retrieved again from the property <code>de.jaetzold.art.RobotInterfaceFactory.platforms</code> and then they are asked again for <code>RobotInterface</code>s.
		<p> 
		All other <code>getInterface(...)</code>-methods rely on this method, so subclasses of <code>RobotInterfaceFactory</code>s only need to override this method to support all the various methods for retrieving a <code>RobotInterface</code>.
		<p>
		If no interface could be found an array of length 0 is returned.
		*/
	public RobotInterface[] getInterfaces(RobotInterfaceDefinition[] ports) {
		if(ports == null || ports.length == 0) {
			return getInterfaces();
			// calls getInterfaces(RobotInterfaceDefinition[]) in turn
		}
		// TBD: generate list of known classes dynamically
		Vector v = new Vector();
		RobotInterface[] interfaces;
		
		// always construct the list of drivers 'from scratch'
		// that way the used Drivers can be changed at any time
		// (e.g. by calling System.setProperty(String, String))
		Properties properties;	// the actual properties to use by this class
		Properties defaults = new Properties();
		try {
			defaults.load(new FileInputStream("de.jaetzold.art.properties"));
		} catch(FileNotFoundException fnfe) {
			if(debug.debug) {
				debug.printError("Properties-file \"de.jaetzold.art.properties\" not found.");
			}
		} catch(IOException ioe) {
			if(debug.debug) {
				debug.printError(1, "An io-error occured while loading the properties-file \"de.jaetzold.art.properties\":" +ioe);
			}
		}
		properties = new Properties(defaults);
		
		// let any System-Properties override the default-ones in our config-file
		// is there a better way to combine two properties?
		try {
			// Piped does not work (always?) within one Thread.
			ByteArrayOutputStream writeEnd = new ByteArrayOutputStream();
			System.getProperties().store(writeEnd, null);
			ByteArrayInputStream readEnd =
				new ByteArrayInputStream(writeEnd.toByteArray());
			properties.load(readEnd);
			writeEnd.close();
			readEnd.close();
		} catch(IOException ioe) {
			if(debug.debug) {
				debug.printError("Exception when copying the System-Properties: "+debug.stackTraceToString(ioe));
			}
		}

		String driverClasses = properties
			.getProperty("de.jaetzold.art.RobotInterfaceFactory.platforms");
		if(driverClasses == null || driverClasses.length()==0) {
			throw new NotConnectedException("No driver selected. Check value of property de.jaetzold.art.RobotInterfaceFactory.platforms");
		}
		StringTokenizer tokenizer = new StringTokenizer(driverClasses);
		while(tokenizer.hasMoreTokens()) {
			Class driverClass;
			String name = tokenizer.nextToken();
			try {
				driverClass = Class.forName(name);
			} catch(ClassNotFoundException cnfe) {
				debug.printInfo(0, cnfe.toString());
				continue;
			}
			// maybe next time better accomplish that with an extra Factory-Object
			// registered by Class.forName(String)
			Method factoryMethod;
			interfaces = null;
			try {
				factoryMethod =
					driverClass.getDeclaredMethod("getInterfaces",
								new Class[] {RobotInterfaceDefinition[].class});
				try {
					interfaces = (RobotInterface[])
						factoryMethod.invoke(null, new Object[]{ports});
				} catch(ClassCastException cce) {
					throw new NoSuchMethodException("Return type is wrong.");
				} catch(Exception e) {
					if(debug.debug) {
						debug.printError("call to getInterfaces on "
										+driverClass
										+" failed: "
										+debug.stackTraceToString(e));
					}
				}
			} catch(NoSuchMethodException nsme) {
				debug.printInfo(0,
								"Class " +driverClass
								+" is not a compatible driver class.");
			}
			
			// The driver class constructed some RobotInterface's
			if(interfaces != null) {
				for(int i=0; i<interfaces.length; i++) {
					if(debug.debug) {
						debug.printInfo("adding " +interfaces[i]);
					}
					v.addElement(interfaces[i]);
				}
			}
		}

		interfaces = new RobotInterface[v.size()];
		v.copyInto(interfaces);
		
		return interfaces;
	}
	/** Returns the first sucessfully detected Interface. If no interface could be found this method returns <code>null</code>.
		@see #getInterfaces(RobotInterfaceDefinition[])
	*/
	public RobotInterface getInterface() {
		RobotInterface[] interfaces = getInterfaces();
		if(interfaces == null || interfaces.length == 0) {
			return null;
		}
		return interfaces[0];
	}
	/** Scans only the given Port. If no interface could be found this method returns <code>null</code>.
		@see #getInterfaces(RobotInterfaceDefinition[])
		*/
	public RobotInterface getInterface(RobotInterfaceDefinition port) {
		return getInterface(new RobotInterfaceDefinition[]{port});
	}
	/** Scans only the given Ports. If no interface could be found this method returns <code>null</code>.
		@see #getInterfaces(RobotInterfaceDefinition[])
		*/
	public RobotInterface getInterface(RobotInterfaceDefinition[] ports) {
		RobotInterface[] interfaces = getInterfaces(ports);
		if(interfaces == null || interfaces.length == 0) {
			return null;
		}
		return interfaces[0];
	}
	/** Searches for an interface conforming to a RobotInterfaceStringDefinition constructed from the given String. If no interface could be found this method returns <code>null</code>.
		@see #getInterfaces(RobotInterfaceDefinition[])
		@see RobotInterfaceStringDefinition
		*/
	public RobotInterface getInterface(String id) {
		return getInterface(new RobotInterfaceStringDefinition(id));
	}
}
