package de.jaetzold.art;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

// its a propertychangelistener for its peer (through an adapter) as well as
// it can have some Propertychangelisteners itself
// TBD: define enable(), disable()? To be able to tell when the Sensor isn't used and therefore doesn't need to be configured and/or polled.

// remove the connected-stuff from this interface?
// maybe just leave isConnected() ?
// and add "enableEvents()"

/**	Classes that represent an input-device should implement this interface.
	Implementations of this interface represent specific interpretations of the value this Sensor delivers.
	*/
public interface Sensor extends PropertyChangeListener/*, SensorListener*/
{
	/**	Returns the value this Sensor represents as an int.
		This method is only for convenience. Better use <code>getValue()</code> because this method may involve casting with a loss of precision.
		@see #getValue()
		*/
	public int getIntValue();
	/**	Returns the value this Sensor represents as a double.
		*/
	public double getValue();
	
	/** Connects the Sensor with the given Port.
		The Sensor gets its peer from the Port so it is therefore required before the Sensor can be properly used.
		@see SensorPort
		*/
	public void connectWith(Port port);
	/**	Tells wether this Sensor has a Peer and whether the Peer is also connected.
		For example a Peer can be not connected if it is a Peer that only forwards calls to another Peer and that one is not connected or if the <code>RobotInterface</code> this Peer belongs to is not connected.
		@see de.jaetzold.art.platform.SensorPeer#isConnected()
		@see RobotInterface#isConnected()
		@see #connectWith(Port)
		*/
	public boolean isConnected();
	/**	Disconnects this Sensor from it's Peer and it's Port.
		This is implicitly called when the Sensor is already connected and asked to connect with another Port.
		@see #connectWith(Port)
		*/
	public void disconnect();

	/**	Adds the specified SensorListener to this Sensor so that it receives SensorEvents from it.
		The Events are sent whenever a change in the value is detected but are  also allowed to be sent when no change happened.
		@see #removeSensorListener(SensorListener)
		@see SensorListener#processEvent(SensorEvent)
		*/
	public void addSensorListener(SensorListener listener);
	/** Removes the specified SensorListener so that it no longer receives SensorEvents from this Sensor.
		This method performs no function, nor does it throw an exception, if the listener specified by the argument was not previously added to this Sensor. If listener is null, no exception is thrown and no action is performed.
		@see #addSensorListener(SensorListener)
		@see SensorListener#processEvent(SensorEvent)
	*/
	public void removeSensorListener(SensorListener listener);
	/**	A PropertyChangeListener receives PropertyChangeEvents whenever this Sensor receives a SensorEvent.
		The PropertyChangeEvent has the SensorEvent as it's new value and the last SensorEvent as the old value. The property name used for this is <code>value</code>.
		This mechanism in the Sensor is not really used for anything else right now but may be in future releases.
		@see #addSensorListener(SensorListener)
		@see SensorListener#processEvent(SensorEvent)
		*/
	public void addPropertyChangeListener(PropertyChangeListener l);
	/** Remove the specified listener.
		@see #addPropertyChangeListener(PropertyChangeListener)
		*/
	public void removePropertyChangeListener(PropertyChangeListener l);

	/**	This is the method through which a Sensor gets informed about changes in the hardware-value it resembles.
		The internal value of this sensor is set to the new value and a SensorEvent is generated with this Sensor as the source. This new SensorEvent is then passed to <code>processEvent(SensorEvent)</code>.
		@see #processEvent(SensorEvent) 
	*/
	public void propertyChange(PropertyChangeEvent event);

	/**	This Method is called when the value this Sensor resembles has changed and events are enabled.
		The given SensorEvent is then passed to any registered listeners, the SensorListeners as well as the PropertyChangeListeners.
		Unlike <code>propertyChange(PropertyChangeEvent)</code> this method is not interested in the new value but in the SensorEvent which resembles the change.
		@see #propertyChange(java.beans.PropertyChangeEvent)
		*/
	public void processEvent(SensorEvent event);
	
	/**	Returns a Port through which other Sensors can be connected to this Sensor.
		This way Sensor-Objects can be cascaded. A Sensor which is connected to the SensorPort returned by this method gets a Peer which represents the value of the Sensor the SensorPort came from. Events are also forwarded.
		Note the difference between <code>getSensorPort()</code> and <code>getPort()</code>.
		@see #getPort()
		*/
	public SensorPort getSensorPort();
	/**	Returns the Port to which this Sensor is connected.
		Note the difference between <code>getSensorPort()</code> and <code>getPort()</code>.
		@see #getPort()
		*/
	public Port getPort();
	
	/**	Causes a new SensorEvent for this Sensor to be generated and delivered.
		This Event is delivered by the usual Thread which delivers the Events. If this is not the case it is at least a different Thread than the one calling this method.
		@see #processEvent(SensorEvent)
		@see de.jaetzold.art.platform.SensorPeer#update()
		@see SensorEvent
		*/
	public void update();
	
	/**	Returns the maximum of the expectable error in the value.
		This is just a guess, what means that nothing <em>really</em> important should depend on it. It is meant to give an algorithm using this Sensor an <em>idea</em> of how precise the value is so that it either can refuse to use it or try to compensate if necessary.
		*/
	public double getPrecision();
}
