package de.jaetzold.art;

import de.jaetzold.art.platform.ServoPeer;

import de.jaetzold.util.StateDifferenceStateDecider;
import de.jaetzold.util.NotConnectedException;
import de.jaetzold.util.Debug;

// TBD: not only a check for the borders in setValue() but continuously to prevent e.g. a "forward" from moving the motor/servo behind the border

/**	A Servo gives control over the absolute position of an axle. It is calibrated so that positions are the same even when initializing in different starting-positions. A Servo has a moving range in which it should stay. That means the AngleSensor representing the position of the Servo won't have values out of this range.<p>
	When simulating a Servo without a special ServoPeer this class needs an AngleSensor and a BooleanSensor.<p>
	The BooleanSensor is associated with a specific position which the Servo is at, wehn the BooleanSensor is true. Together with a direction where the BooleanSensor can be "found" the calibration to an absolute position is done by just moving into that direction until the BooleanSensor becomes true. Usually this is achieved by placing the BooleanSensor on one end of the moving range.
	The AngleSensor is pretty much used in the same way as the CountSensor in the superclass StepperMotor. In fact it just scales the CountSensors value so that it represents the correct angle.
	@see ServoPeer
	@see AngleSensor
	@see BooleanSensor
	*/
public class Servo extends StepperMotor {
	private final Debug debug = new Debug("de.jaetzold.art.Servo");
	{ if(debug.debug) debug.objectIdentifier = debug.getIdentifierFor(this); }

	/** Constructs a new Servo.
		It is not specified whether it will have a ServoPeer or will be used as a simulating servo.
		*/
	public Servo() {
	}
	/** Constructs a new Servo.
		It uses the given parameters for simulating the usual Servo-behavior without having a ServoPeer.
		@param angleSensor the AngleSensor for measuring the relative position of the Servo
		@param positionSwitch the BooleanSensor for noticing the absolute position given by positionSwitchAngle
		@param positionSwitchAngle the absolute angle the positionSwitch is at
		@param positiveBorderAngle the positive border of the moving range
		@param negativeBorderAngle the negative border of the moving range
		@param searchDirection the direction in which to move for always reaching positionSwitchAngle. Positive values mean "forward".
		*/
	public Servo(AngleSensor angleSensor,
				BooleanSensor positionSwitch,
				double positionSwitchAngle,
				double positiveBorderAngle,
				double negativeBorderAngle,
				int searchDirection) {
		super();
		setPrecision(getPrecision()/angleSensor.getUnitsPerTurn());
		setAngleSensor(angleSensor);
		setPositionSwitch(	positionSwitch, 
							positionSwitchAngle,
							positiveBorderAngle,
							negativeBorderAngle,
							searchDirection);
	}
	
	private BooleanSensor positionSwitch;
	private double positionSwitchAngle;
	private double positiveBorderAngle;
	private double negativeBorderAngle;
	private int searchDirection;
	/** Sets the the BooleanSensor an other required parameters for calibrating this Servo to an absolute position. This is not required if the Servo is used in conjunction with a ServoPeer.
		@param positionSwitch the BooleanSensor for noticing the absolute position given by positionSwitchAngle
		@param positionSwitchAngle the absolute angle the positionSwitch is at
		@param positiveBorderAngle the positive border of the moving range
		@param negativeBorderAngle the negative border of the moving range
		@param searchDirection the direction in which to move for always reaching positionSwitchAngle. Positive values mean "forward".
		*/
	public void setPositionSwitch(	final BooleanSensor positionSwitch,
									final double positionSwitchAngle,
									double positiveBorderAngle,
									double negativeBorderAngle,
									int searchDirection) {
		this.positiveBorderAngle = positiveBorderAngle;
		this.negativeBorderAngle = negativeBorderAngle;
		
		double oldAngle = this.positionSwitchAngle;
		this.positionSwitchAngle = positionSwitchAngle;
		if(angleSensor != null) {
			double newAngle = positionSwitchAngle;
			try {
				newAngle = angleSensor.getValue()
										+(positionSwitchAngle-oldAngle);
			} catch(NotConnectedException nce) {
			}
			angleSensor.setAngle(newAngle);
		}
		
		this.positionSwitch = positionSwitch;
		this.searchDirection = searchDirection;
		// register Listener for positionSwitch which sets the position
		positionSwitch.addSensorListener(
			new SensorListener() {
				public void processEvent(SensorEvent se) {
					if(positionSwitch.convertToBoolean(se.getValue())) {
						if(debug.debug) {
							debug.printInfo("positionSwitch reached, setting angle to " +positionSwitchAngle);
						}
						angleSensor.setAngle(positionSwitchAngle);
					}
				}
			}
		);
		try {
			calibrate();
		} catch(NotConnectedException nce) {
		}
	}
	
	public synchronized void calibrate() {
		if(servoPeer != null) {
			servoPeer.calibrate();
		} else {
			// move to positionSwitch by ourselves
			if(positionSwitch != null) {
				if(debug.debug) {
					debug.printInfo("starting calibration by"
									+" moving to position-switch.");
				}
				off();
				SensorListener stop =
					new SensorListener() {
						public synchronized void processEvent(SensorEvent se) {
							if(positionSwitch
									.convertToBoolean(se.getValue())) {
								this.notifyAll();
							}
						}
					};
				if(!positionSwitch.getBooleanValue()) {
					synchronized(stop) {
						positionSwitch.addSensorListener(stop);

						if(searchDirection >= 0) {
							forward();
						} else {
							backward();
						}
						on();
						try {
							if(debug.debug) {
								debug.printInfo("calibrate: waiting for stop");
							}
							stop.wait();
							if(debug.debug) {
								debug.printInfo("calibrate: stop happened");
							}
						} catch(InterruptedException ie) {
						}
						positionSwitch.removeSensorListener(stop);
					}
				}
			}
		}
		
		super.calibrate();
	}
	
	private double desiredAngle;
	public void setValue(double value) {
		double allowedValue =
				Math.min(	Math.max(	value, 
										negativeBorderAngle),
							positiveBorderAngle);
		AngleSensor localSensor = angleSensor;
		if(debug.debug) {
			Double current = null;
			Double count = null;
			if(localSensor != null) {
				current = new Double(localSensor.getValue());
				count = new Double(localSensor
										.convertAngleToUnits(allowedValue));
			}
			debug.printInfo("Got value " +value
							+", are at " +current
							+", setting to " +allowedValue
							+", count " +count);
		}
		desiredAngle = allowedValue;
		super.setValue(desiredAngle);
	}
	
	private AngleSensor angleSensor;
	// don't return the real angleSensor because that can
	// change over time
	// private Sensor externalAngleSensor = new RawSensor();
	/**	Returns the AngleSensor representing the position of this Servo. It may be the one used for simulating the servo-behavior or the one directly from the ServoPeer.
		Keep in mind that it is possible that this Servo gets a new Peer in which case the AngleSensor returned by this method may not be valid anymore for this Servo. It is possible that future versions don't have this limitation.
		@see #setAngleSensor(AngleSensor)
		*/
	public AngleSensor getAngleSensor() {
		return angleSensor;
	}
	
	/**	Sets the AngleSensor for representing the position of this Servo.
		It is not required to set the AngleSensor when using a Servo as long as it has a ServoPeer. Otherwise the given AngleSensor is used in conjunction with a BooleanSensor to simulate the servo-behavior.
		@see #setPositionSwitch(BooleanSensor,double,double,double,int)
		@see de.jaetzold.art.platform.ServoPeer#getAngleSensor()
		*/
	public void setAngleSensor(AngleSensor sensor) {
		setCountSensor(sensor);
	}
	
	public synchronized void setCountSensor(CountSensor sensor) {
		if(sensor instanceof AngleSensor || sensor == null) {
			//if(externalAngleSensor != null) {	// we're finished initializing
												// superclass
				angleSensor = (AngleSensor)sensor;
				/*if(angleSensor != null) {
					externalAngleSensor
						.connectWith(angleSensor.getSensorPort());
				} else {
					externalAngleSensor.disconnect();
				}*/
				super.setCountSensor(angleSensor);
			//}
		} else {
			throw new IllegalArgumentException("A Servo needs an AngleSensor as CountSensor.");
		}
	}
	
	private ServoPeer servoPeer;
	public synchronized void connectWith(ActuatorPort port) {
		super.connectWith(port);
		if(getPeer() instanceof ServoPeer) {
			servoPeer = (ServoPeer)getPeer();
			setAngleSensor(servoPeer.getAngleSensor());
		} else {
			servoPeer = null;
		}
	}
	
	public synchronized void disconnect() {
		if(servoPeer != null) {
			setAngleSensor(null);
		}
		servoPeer = null;
		super.disconnect();
	}
}
