package de.jaetzold.art;

import de.jaetzold.art.platform.StateSensorPeer;

import de.jaetzold.util.StateDecider;
import de.jaetzold.util.ValueToStateStateDecider;
import de.jaetzold.util.Debug;

// TBD: maybe StateSensor as an Actuator which interprets setValue() as a setState()

/**	This class interprets the values as a state. That means somehow discrete values. For interpreting values as a state by itself (without a <code>StateSensorPeer</code>) it can have a <code>StateDecider</code>.
	<p>
	Be aware that - stricly speaking - a statesenor without a special <code>StateSensorPeer</code> requires <code>setChangeEventsEnabled(true)</code>, if the same values can produce different states when received in different states. That is for example the case if we have a <code>TransitiontablestateDecider</code>. Maybe fix that one day by setting <code>changeEventsEnabled</code> depending on our own statedeciders personality?
	<p>
	States are not ints because other subclasses do need doubles right now.
	@see de.jaetzold.util.StateDecider
	@see de.jaetzold.art.platform.StateSensorPeer
	*/
public class StateSensor extends RawSensor {
	private final Debug debug = new Debug("de.jaetzold.art.StateSensor");
	{ if(debug.debug) debug.objectIdentifier = debug.getIdentifierFor(this); }
	
	/** This is the StateDecider which gets used when we have no StateSensorPeer. If it is null the defaultStateDecider is used instead which is an instance of <code>ValueToStateStateDecider</code>.
		@see StateSensorPeer
		@see de.jaetzold.util.ValueToStateStateDecider
	*/
	protected StateDecider stateDecider;
	/** This is the default StateDecider. It gets used when we have no StateSensorPeer and stateDecider is null.
		@see #stateDecider
	*/
	private StateDecider defaultStateDecider = new ValueToStateStateDecider();
	
	/**	Returns the <code>StateDecider</code> used to interpret values as states by ourselves.
		*/
	public StateDecider getStateDecider() {
		StateSensorPeer localPeer = stateSensorPeer;
		if(localPeer != null) {
			return null;
		} else {
			return stateDecider;
		}
	}
	/**	Sets the <code>StateDecider</code> used to interpret values as states by ourselves.
		If there is a new value from the peer the given decider is used to calculate the next state with the last state as the first argument to the method <code>StateDecider.nextStateValue(double,double)</code> and the new value as the second.
		<p>
		Beware, as long as the peer is a <code>StateSensorPeer</code> the decider isn't called anytime.
		@see de.jaetzold.art.platform.StateSensorPeer
		*/
	public synchronized void setStateDecider(StateDecider decider)
	{
		stateDecider = decider;
	}

	private double lastState;
	/**	This hook from <code>RawSensor</code> is used to convert the value. Here the <code>StateDecider</code> get's called.
		@see #setStateDecider(StateDecider)
		*/
	protected double convertToIncoming(double value)
	{
		double result = super.convertToIncoming(value);
		
		// compute the state by ourselves if the peer doesn't
		if(stateSensorPeer == null) {
			StateDecider localDecider = stateDecider;
			if(localDecider == null) {
				localDecider = defaultStateDecider;
			}

			result = localDecider.nextStateValue(lastState, value);
			if(debug.debug) {
				debug.printInfo("computed " +result
								+" out of " +lastState +", " +value
								+" with " +localDecider);
			}
		}
		lastState = result;
		
		return result;
	}
	
	private StateSensorPeer stateSensorPeer;
	public synchronized void connectWith(SensorPort port) {
		super.connectWith(port);
		if(getPeer() instanceof StateSensorPeer) {
			stateSensorPeer = (StateSensorPeer)getPeer();
		} else {
			// should'nt be necessary
			stateSensorPeer = null;
		}
	}
	
	public synchronized void disconnect() {
		stateSensorPeer = null;
		super.disconnect();
	}
		
	// this sensor only has int as value, but let subclasses
	// and outgoing delegates do what they want
	// obsolete (true) as long as StateDecider can't be asked for it
	public boolean getDeliversFloatingPoint() {
		if(stateSensorPeer != null) {
			return super.getDeliversFloatingPoint();
		} else {
			return true;
		}
	}
	
	public boolean isIncomingConversionCacheable() {
		StateDecider localDecider = stateDecider;
		if(stateSensorPeer != null) {
			return super.isIncomingConversionCacheable();
		} else if(localDecider != null) {
			return localDecider.isCacheable()
					&& super.isIncomingConversionCacheable();
		} else {
			return defaultStateDecider.isCacheable()
					&& super.isIncomingConversionCacheable();
		}
	}
}
