package de.jaetzold.art;

import de.jaetzold.art.platform.StepperMotorPeer;
import de.jaetzold.art.platform.SensorPeer;
import de.jaetzold.art.platform.BaseSensorPeer;
import de.jaetzold.art.platform.BaseSensorPort;

import de.jaetzold.util.BinaryOperator;
import de.jaetzold.util.BooleanConversion;
import de.jaetzold.util.NotConnectedException;
import de.jaetzold.util.Debug;

// TBD: own Event-Type for reaching a desiredPosition (see Motor for own Event-Types)

// TBD: what to do when the countSensor is null?

// TBD: setSpeed() because a stepperMotor knows and therefore can control how fast the wheel is spinning.

// is it correct that a stepperMotor tries to keep its position?
// or should it not bother about being moved by others once the desired positon is reached?
/**	This class is an abstraction of a stepper motor.
	A stepper motor is able to control the amount of movement it does, it can "make one step" at a time. How big such a step is is not defined, but they are equal in size. That way a StepperMotor is positionable in a relative sort of fashion.<p>
	It should be noted that the getValue()-method of StepperMotor doesn't return the position of the motor, but it's state (exactly as Motor does). So keep especially in mind, that getValue() usually doesn't return what has been previously set with setValue(). To retrieve the position use the CountSensor delivered by getCountSensor().<p>
	Calls to methods from Motor like <code>on()</code> or <code>forward()</code> stop any positioning of the StepperMotor and let it move freely. It still keeps track of the current position then.<p>
	If a Port can't deliver a real StepperMotorPeer this class is able to simulate the required behavior by using a CountSensor and the capabilities of its superclass Motor.
	@see #getCountSensor()
	@see StepperMotorPeer
	@see CountSensor
	*/
public class StepperMotor extends Motor {
	private final Debug debug = new Debug("de.jaetzold.art.StepperMotor");
	{ if(debug.debug) debug.objectIdentifier = debug.getIdentifierFor(this); }

	/** Constructs a new StepperMotor.
		It is not specified whether it will have a StepperMotorPeer or an explicit CountSensor to use in simulating the step-behavior.
		*/
	public StepperMotor() {
	}
	/** Constructs a new StepperMotor.
		The given CountSensor is used to simulate the step-behavior.
		*/
	public StepperMotor(CountSensor sensor) {
		setCountSensor(sensor);
	}
	
	/**	The CountSensor representing the position of this StepperMotor. It may be the one used for simulating the step-behavior or the one directly from the StepperMotorPeer.
		*/
	protected CountSensor countSensor;
	// don't return the real countSensor because it can
	// change over time
	// it is no CountSensor because it would do a conversion to int which could break an outgoing-conversion done by countSensor
	//private Sensor externalCountSensor = new RawSensor();
	/**	Returns the CountSensor representing the position of this StepperMotor. It may be the one used for simulating the step-behavior or the one directly from the StepperMotorPeer.
		Keep in mind that it is possible that this StepperMotor gets a new Peer in which case the CountSensor returned by this method may not be valid anymore for this StepperMotor. It is possible that future versions don't have this limitation.
		@see #setCountSensor(CountSensor)
		*/
	public CountSensor getCountSensor() {
		return countSensor;
	}
	
	private SensorListener countListener =
		new SensorListener() {
			public void processEvent(SensorEvent se) {
				if(debug.debug) {
					debug.printInfo(" got event " +se
									+" from " +se.getSource());
				}
				newCountValue(se.getValue());
			}
		};
	
	/**	Sets the CountSensor for representing the position of this StepperMotor.
		It is not required to set the CountSensor when using a StepperMotor as long as it has a StepperMotorPeer. Otherwise the given CountSensor is used to simulate the step-behavior.
		@see de.jaetzold.art.platform.StepperMotorPeer#getCountSensor()
		*/
	public synchronized void setCountSensor(CountSensor sensor) {
		if(countSensor != null) {
			countSensor.removeSensorListener(countListener);
		}
		if(sensor == null) {
			positioning = false;
			positionDistanceSensor.disconnect();
		} else {
			positionDistanceSensor.connectWith(sensor.getSensorPort());
		}
		countSensor = sensor;
		if(stepperMotorPeer == null && sensor != null) {
			sensor.addSensorListener(countListener);
		}
		if(stepperMotorPeer != null) {
			stepperMotorPeer.setCountSensor(sensor);
		}
		/*if(sensor != null) {
			externalCountSensor.connectWith(sensor.getSensorPort());
		} else {
			externalCountSensor.disconnect();
		}*/
		try {
			calibrate();
		} catch(NotConnectedException nce) {
		}
	}
	
	private BinaryOperator differenceOperator =
		new BinaryOperator() {
			public double combine(double firstOperand, double secondOperand) {
				return firstOperand-secondOperand;
			}
		};
	private CombinedSensor positionDistanceSensor;
	private SensorPeer desiredCountSensorPeer =
		new BaseSensorPeer() {
			public double getValue() {
				return desiredCount;
			}
		};
	{
		SensorPort desiredCountPort =
			new BaseSensorPort() {					
				public SensorPeer getPeerFor(Sensor sensor) {
					return desiredCountSensorPeer;
				}
			};
		Sensor desiredCountSensor = new RawSensor();
		desiredCountSensor.connectWith(desiredCountPort);
		positionDistanceSensor = new CombinedSensor(desiredCountSensor,
													differenceOperator);

	}

	private BooleanSensor positionReachedSensor =
									new BooleanSensor();
	{
		positionReachedSensor.setBooleanConversionDelegate(
			new BooleanConversion() {
				public boolean isCacheable() {
					return false;
				}
				public boolean convertToBoolean(double value) {
					return Math.abs(value)<=precision;
				}
			}
		);
		positionReachedSensor
				.connectWith(positionDistanceSensor.getSensorPort());
	}
	
	/**	Returns a BooleanSensor that represents the positioning-state of this StepperMotor.
		If the StepperMotor is correctly positioned (within its precision) the BooleanSensor is true, otherwise false. Listeners may be added to the BooleanSensor which then get informed whenever that state changes.
		@see #getPrecision()
		@see #isPositionedAt(double)
		*/
	public BooleanSensor getPositionReachedSensor() {
		return positionReachedSensor;
	}
	
	private double precision = 0.5;
	public void setPrecision(double precision) {
		this.precision = Math.abs(precision);
		positionReachedSensor.update();
		// give a chance to correct our position
		calibrate();
	}
	public double getPrecision() {
		return precision;
	}
	
	/**	Tells whether this StepperMotor is currently at the position given by <code>count</code>. The precision is taken into account.
		@see #getPositionReachedSensor()
		*/
	public boolean isPositionedAt(double count) {
		CountSensor localSensor = countSensor;
		if(localSensor == null) {
			return false;
		}
		return positionReachedSensor.convertToBoolean(
						differenceOperator.combine(localSensor.getValue(),
													count));
	}
	
	private boolean positioning = true;
	public void setBackward(boolean backward) {
		positioning = false;
		super.setBackward(backward);
	}
	
	private boolean explicitOff;
	public void setOn(boolean on) {
		explicitOff = !on;
		if(on && positioning) {
			newCountValue(countSensor.getValue());
		} else {
			super.setOn(on);
		}
	}
	
	/**	Sets the current position as the desired position. This way the StepperMotor stops and holds the current position.
		*/
	public void keepCurrentPosition() {
		CountSensor localSensor = countSensor;
		if(localSensor != null) {
			try {
				setValue(localSensor.getValue());
			} catch(NotConnectedException nce) {
			}
		}
	}
	
	// "things" have changed, maybe repositioning is required
	/**	This method is called when (re)calibration is required. This may take a while. It is called implicitly when initializing with a new Peer.
		*/
	public void calibrate() {
		if(debug.debug) {
			debug.printInfo("calibrating to " +desiredCount);
		}
		try {
			setValue(desiredCount);
		} catch(NotConnectedException nce) {
		}
	}

	// TBD: move slower when approaching desiredPosition
	// this should be done through setPower/setSpeed only
	// and leave the real slowdown e.g. in Fischertechnik to the peer
	private void newCountValue(double value) {
		if(debug.debug) {
			debug.printInfo("got new value " +value
							+", desired=" +desiredCount
							+", precision=" +precision
							+", positioning=" +positioning);
		}
		if(positioning) {
			if(positionReachedSensor.convertToBoolean(
						differenceOperator.combine(desiredCount, value))) {
				if(debug.debug) {
					debug.printInfo("position reached, stopping ...");
				}
				// super required because "this" would mean explicit off
				super.setOn(false);
			} else if(!explicitOff) {
				if(value > desiredCount) {
					if(debug.debug) {
						debug.printInfo("position is backward, going to ...");
					}
					super.setBackward(true);
				} else {
					if(debug.debug) {
						debug.printInfo("position is forward, going to ...");
					}
					super.setBackward(false);
				}
				super.setOn(true);
			}
			if(debug.debug) {
				debug.printInfo("... the new direction.");
			}
		}
	}
	
	/**	Move relatively from the currently desired position a number of steps given by <code>value</code>.
		@see #setValue(double)
		*/
	public void step(double value) {
		setValue(desiredCount+value);
	}
	
	/**	The position the StepperMotor currently wants to be at.
		@see #step(double)
		@see #setValue(double)
		*/
	protected double desiredCount;
	/**	Sets the position the StepperMotor currently wants to be at. That means it repositions itself as long as the value of it's CountSensor does not match value in terms of precision. A call to Motor-specific methods like <code>forward()</code> stops the positioning until the next call to <code>setValue(double)</code>, <code>step(double)</code> or <code>calibrate()</code>.
		@see #step(double)
		@see #desiredCount
		*/
	public void setValue(double value) {
		StepperMotorPeer localPeer = stepperMotorPeer;
		if(localPeer != null) {
			// for this to work it is crucial that Motor.setValue()
			// also doesn't modify the value.
			// that is the case because StepperMotorPeer is also a MotorPeer
			super.setValue(value);
		} else {
			CountSensor localSensor = countSensor;
			if(debug.debug) {
				Double current = null;
				if(localSensor != null) {
					current = new Double(localSensor.getValue());
				}
				debug.printInfo("Are at " +current
								+", setting to " +value);
			}
			if(countSensor == null) {
				if(isConnected()) {
					throw new NotConnectedException("Need a CountSensor because Peer is'nt capable of delivering one.");
				} else {
					// setting values to not connected Actuators does nothing
					return;
				}
			}
			desiredCount = value;
			if(desiredCount != value) {	// update is costly, so try to avoid
				desiredCountSensorPeer.update();
			}
			positioning = true;
			// non-super to trigger newCountValue()
			// and "remove" explicitOff
			setOn(true);
		}
	}
	
	private StepperMotorPeer stepperMotorPeer;
	public synchronized void connectWith(ActuatorPort port) {
		super.connectWith(port);
		if(getPeer() instanceof StepperMotorPeer) {
			stepperMotorPeer = (StepperMotorPeer)getPeer();
			setCountSensor(stepperMotorPeer.getCountSensor());
		} else {
			stepperMotorPeer = null;
		}
	}
	
	public synchronized void disconnect() {
		if(stepperMotorPeer != null) {
			setCountSensor(null);
		}
		stepperMotorPeer = null;
		super.disconnect();
	}
	
	/**	Returns the speed of this StepperMotor.
		This implementation of StepperMotor does not really control the Speed by itself. If it is not connected to a StepperMotorPeer speed has the same effect as power.
		@see #getPower()
		*/
	public double getSpeed() {
		StepperMotorPeer localPeer = stepperMotorPeer;
		if(localPeer != null) {
			return localPeer.getSpeed();
		} else {
			return getPower();
		}
	}
	/**	Sets the speed of this StepperMotor to the given value.
		This often has an effect to the torque of the motor. In fact, this implementation of StepperMotor does not control the speed by itself. If it is not connected to a StepperMotorPeer it just forwards any call to <code>setSpeed(double)</code> to <code>setPower(double)</code>.
		Dependig on the value of <code>waitForCompletion</code> this Method returns immediately after it has been registerd that this is now the state the motor should have, or it returns after the state has been set.
		@see #setPower(double)
		@see #waitForCompletion
		*/
	public synchronized void setSpeed(double speed) {
		if(stepperMotorPeer != null) {
			stepperMotorPeer.setSpeed(speed);
		} else {
			setPower(speed);
		}
	}
}
