package de.jaetzold.art.examples;

import de.jaetzold.art.Servo;
import de.jaetzold.art.BooleanSensor;
import de.jaetzold.art.AngleSensor;
import de.jaetzold.art.SensorListener;
import de.jaetzold.art.SensorEvent;

import de.jaetzold.util.Debug;

import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;

public class LiSeSimpleAlgorithm implements LiSe {
	private final Debug debug =
				new Debug("de.jaetzold.art.examples.LiSeSimpleAlgorithm");
	protected Robot robot;
	protected Servo servo;
	public LiSeSimpleAlgorithm(Robot robot) {
		this.robot = robot;
		servo = robot.getTurnServo();
		stop();
		servo.calibrate();
		servo.setValue(0);
// BEGIN lise:impl:border
		final BooleanSensor borderSensor = servo.getPositionReachedSensor();
		borderSensor.addSensorListener(
			new SensorListener() {
				public void processEvent(SensorEvent se) {
					synchronized(LiSeSimpleAlgorithm.this) {
						boolean reached =
								borderSensor.convertToBoolean(se.getValue());
						if(debug.debug) {
							debug.printInfo("reached=" +reached
											+", value=" +se.getValue()
											+", state=" +state);
						}
						
						if(reached && (state == SEARCH)) {
							search();
						}
					}
				}
			}
		);
// END lise:impl:border
	}
// BEGIN lise:robot
	public interface Robot {
		public double getMaxRightAngle();
		public double getMaxLeftAngle();
		public double getSearchSpeed();
		public double getFollowSpeed();
		public Servo getTurnServo();
	}
// END lise:robot
// BEGIN lise:impl
	private boolean searchRight;
	public synchronized void seeLeft() {
		servo.setSpeed(robot.getFollowSpeed());
		servo.setValue(robot.getMaxLeftAngle());
		searchRight = false;
		setState(SEE_LEFT);
	}
	public synchronized void seeRight() {
		servo.setSpeed(robot.getFollowSpeed());
		servo.setValue(robot.getMaxRightAngle());
		searchRight = true;
		setState(SEE_RIGHT);
	}
	public synchronized void search() {
		// important for direction-change when reaching MaxAngle
		setState(SEARCH);
		
		if(servo.isPositionedAt(robot.getMaxRightAngle())) {
			searchRight = false;
		} else if(servo.isPositionedAt(robot.getMaxLeftAngle())) {
			searchRight = true;
		}
		
		servo.setSpeed(robot.getSearchSpeed());
		if(searchRight) {
			servo.setValue(robot.getMaxRightAngle());
		} else {
			servo.setValue(robot.getMaxLeftAngle());
		}
	}
	public synchronized void seeCenter() {
		servo.keepCurrentPosition();
		setState(SEE_CENTER);
	}
	public synchronized void stop() {
		servo.keepCurrentPosition();
		setState(STOP);
	}
	public synchronized boolean isSeeing() {
		switch(getState()) {
			case STOP:
			case SEARCH:
				return false;
			default:
				return true;
		}
	}
	public AngleSensor getAngleSensor() {
		return servo.getAngleSensor();
	}
// END lise:impl
// ...	
	protected int state;
	protected synchronized void setState(int state) {
		int oldState = this.state;
		this.state = state;
		changes.firePropertyChange("state", oldState, state);
	}
	
	protected PropertyChangeSupport changes =
				new PropertyChangeSupport(this);
	public void addPropertyChangeListener(PropertyChangeListener l) {
		changes.addPropertyChangeListener(l);
	}
	public void removePropertyChangeListener(PropertyChangeListener l) {
		changes.removePropertyChangeListener(l);
	}
		
	public int getState() {
		return state;
	}
	public LiSeSimpleAlgorithm() throws Exception {
		this((Robot)Class.forName(
				System.getProperty(
					className +".robotClass",
					"de.jaetzold.art.examples.LiSeSimpleFTDemoRobot"
				)
			).newInstance()
		);
	}
	public static void main(String[] argv) throws Exception {
		int index = 0;
		if(argv.length > index) {
			System.setProperty(className +".robotClass", argv[index++]);
		}
		LiSeSimpleAlgorithm lise = new LiSeSimpleAlgorithm();
		
		LiSeView view = new LiSeView(lise);
		
		LiSeCommands commands = new LiSeCommands(lise);
		
		java.awt.Frame f = new java.awt.Frame(className);
		f.setLayout(new java.awt.BorderLayout());
		f.add("North", view);
		f.add("South", commands);
		
		f.addWindowListener(
			new java.awt.event.WindowAdapter() {
				public void windowClosing(java.awt.event.WindowEvent we) {
					System.exit(0);
				}
			}
		);
		f.pack();
		f.show();
		
	}
	private static String className =
			"de.jaetzold.art.examples.LiSeSimpleAlgorithm";
// ...
}
