package de.jaetzold.art.examples;

import de.jaetzold.art.AngleSensor;
import de.jaetzold.art.LightSensor;
import de.jaetzold.art.CombinedSensor;
import de.jaetzold.art.BooleanSensor;
import de.jaetzold.art.ConstantSensorPort;
import de.jaetzold.art.subsumption.SchedulerTask;
import de.jaetzold.util.Task;
import de.jaetzold.util.Event;
import de.jaetzold.util.Queue;
import de.jaetzold.util.TaskPerformer;
import de.jaetzold.util.RequeueTask;
import de.jaetzold.util.BooleanConversion;
import de.jaetzold.util.BinaryOperator;
import de.jaetzold.util.BeanSupport;	// does some reflect-stuff
import de.jaetzold.util.Debug;

import java.beans.PropertyChangeListener;

/*	A good example for nested subsumtion could be constructed here. Outermost Scheduler switches between see-no-light behavior (search) and a see-light behavior (seeLeft/Right/Centered) which is a Scheduler by itself with behaviors for the three states
*/

public class SubsumptionLiSe implements LiSe {
	private final Debug debug =
				new Debug("de.jaetzold.art.examples.SubsumptionLiSe");
	protected LiSe lise;
	protected Task search;
	protected Task centeredTask;
	protected Task seeLeftTask;
	protected Task seeRightTask;
	
	protected SchedulerTask scheduler;
	public SchedulerTask getSchedulerTask() {
		return scheduler;
	}
	protected TaskPerformer schedulerPerformer;
	// to be able to start it and setDaemon
	public TaskPerformer getSchedulerPerformer() {
		return schedulerPerformer;
	}
	
	public SubsumptionLiSe(Robot robot) {
		this.lise = robot.getLiSe();
		scheduler = new SchedulerTask();
		
		// standard always-active behavior 'search'
		BooleanSensor constantTrue = new BooleanSensor();
		constantTrue.connectWith(new ConstantSensorPort(1));
		search = new Task() {
			public Event perform() {
				lise.search();
				return null;
			}
		};
		scheduler.addBehaviorFor(new Double(0), constantTrue, search);
		
// BEGIN lise:subsumption
		// behavior for seeing light on the left side
		final LightSensor leftSensor = robot.getLeftSensor();
		final LightSensor rightSensor = robot.getRightSensor();
		BinaryOperator seeLeftOperator =
			new BinaryOperator() {
				public double combine(	double leftValue,
										double rightValue) {
					// at least one of them has to 'see' something
					if(leftValue >= leftSensor.getMeasurableMin()
							|| rightValue >= rightSensor.getMeasurableMin()) {
						// Standard BooleanConversion: only 0 is false
						return leftValue>rightValue ? 1 : 0;
					}
					return 0;
				}
			};
		
		CombinedSensor seeLeftSensor = new CombinedSensor();
		seeLeftSensor.connectWith(leftSensor.getSensorPort());	
		seeLeftSensor.setOperator(seeLeftOperator);
		seeLeftSensor.setOperand(rightSensor);
		
		BooleanSensor seeLeftActivator = new BooleanSensor();
		seeLeftActivator.connectWith(seeLeftSensor.getSensorPort());

		seeLeftTask = new Task() {
			public Event perform() {
				lise.seeLeft();
				return null;
			}
		};
		scheduler.addBehaviorFor(new Double(1),
								seeLeftActivator,
								seeLeftTask);
// END lise:subsumption
		
		// behavior for seeing light on the right side
		BinaryOperator seeRightOperator =
			new BinaryOperator() {
				public double combine(	double rightValue,
										double leftValue) {
					// at least one of them has to "see" something
					if(leftValue >= leftSensor.getMeasurableMin()
							|| rightValue >= rightSensor.getMeasurableMin()) {
						// Standard BooleanConversion: only 0 is false
						return rightValue>leftValue ? 1 : 0;
					}
					return 0;
				}
			};
		
		CombinedSensor seeRightSensor = new CombinedSensor();
		seeRightSensor.connectWith(rightSensor.getSensorPort());
		seeRightSensor.setOperator(seeRightOperator);
		seeRightSensor.setOperand(leftSensor);	
		
		BooleanSensor seeRightActivator = new BooleanSensor();
		seeRightActivator.connectWith(seeRightSensor.getSensorPort());

		seeRightTask = new Task() {
			public Event perform() {
				lise.seeRight();
				return null;
			}
		};
		scheduler.addBehaviorFor(new Double(2),
								seeRightActivator,
								seeRightTask);
		

		// behavior for seeing light centered
		BinaryOperator centeredOperator =
			new BinaryOperator() {
				public double combine(	double leftValue,
										double rightValue) {
					if(debug.debug) {
						debug.printInfo("got " +leftValue
										+" from left and " +rightValue
										+" from right Sensor");
					}
					boolean result = false;
					// at least one of them has to "see" something
					if(leftValue >= leftSensor.getMeasurableMin()
							|| rightValue >= rightSensor.getMeasurableMin()) {
						double precision = leftSensor.getPrecision()
											+ rightSensor.getPrecision();
						result = Math.abs(leftValue - rightValue)
										<= precision * 2.5;
						// a factor instead of an absolute value
						// since LightSensor has no idea about a standard
						// scale by now (lux or whatever)
					}
					// Standard BooleanConversion: only 0 is false
					return result ? 1 : 0;
				}
			};
		
		CombinedSensor centeredSensor = new CombinedSensor();
		centeredSensor.connectWith(leftSensor.getSensorPort());
		centeredSensor.setOperator(centeredOperator);
		centeredSensor.setOperand(rightSensor);	
		
		BooleanSensor centeredActivator = new BooleanSensor();
		centeredActivator.connectWith(centeredSensor.getSensorPort());

		centeredTask = new Task() {
			public Event perform() {
				lise.seeCenter();
				return null;
			}
		};
		scheduler.addBehaviorFor(new Double(3),
								centeredActivator,
								centeredTask);
		
		// construct the Thread which continuosly calls scheduler.run()
		Queue taskQueue = new Queue();
		schedulerPerformer = new TaskPerformer(taskQueue);
		// schedulerPerformer.start(); // start the performer from the outside
		// halt first so that the algorithm must be enabled with "search()"
		schedulerPerformer.haltPerforming(true);
		taskQueue.post(new RequeueTask(scheduler, taskQueue));
	}
	
	public interface Robot {
		public LiSe getLiSe();
		public LightSensor getLeftSensor();
		public LightSensor getRightSensor();
	}
	
	public void search() {
		// start algorithm
		schedulerPerformer.resumePerforming();
	}
	public void stop() {
		// stop algorithm
		schedulerPerformer.haltPerforming();
		try {
			schedulerPerformer.waitForHalt();
		} catch(InterruptedException ie) {
		}
		lise.stop();
	}
	public void seeLeft() {
		stop();
		lise.seeLeft();
	}
	public void seeRight() {
		stop();
		lise.seeRight();
	}
	public void seeCenter() {
		stop();
		lise.seeCenter();
	}
	public synchronized boolean isSeeing() {
		return lise.isSeeing();
	}
	public AngleSensor getAngleSensor() {
		return lise.getAngleSensor();
	}
	
	public void addPropertyChangeListener(PropertyChangeListener l) {
		BeanSupport.addPropertyChangeListenerFor(lise, l);
	}
	public void removePropertyChangeListener(PropertyChangeListener l) {
		BeanSupport.removePropertyChangeListenerFor(lise, l);
	}
// ...	
	public int getState() {
		return lise.getState();
	}
	public SubsumptionLiSe() throws Exception {
		this((Robot)Class.forName(
				System.getProperty(
					className +".robotClass",
					"de.jaetzold.art.examples.SubsumptionLiSeFTDemoRobot"
				)
			).newInstance()
		);
	}
	public static void main(String[] argv) throws Exception {
		int index = 0;
		if(argv.length > index) {
			System.setProperty(className +".robotClass", argv[index++]);
		}
		if(argv.length > index) {
			System.setProperty(argv[index-1] +".liseClass", argv[index++]);
		}
		SubsumptionLiSe lise = new SubsumptionLiSe();
		
		LiSeView view = new LiSeView(lise);
		
		LiSeCommands commands = new LiSeCommands(lise);
		
		java.awt.Frame f = new java.awt.Frame(className);
		f.setLayout(new java.awt.BorderLayout());
		f.add("North", view);
		f.add("South", commands);
		
		f.addWindowListener(
			new java.awt.event.WindowAdapter() {
				public void windowClosing(java.awt.event.WindowEvent we) {
					System.exit(0);
				}
			}
		);
		f.pack();
		f.show();
		
		lise.getSchedulerPerformer().start();	// start subsumption
	}
	private static String className =
			"de.jaetzold.art.examples.SubsumptionLiSe";
// ...
}
