package de.jaetzold.art.platform;

import de.jaetzold.art.SensorEventDescriptor;
import de.jaetzold.art.AllSensorEventDescriptor;
import de.jaetzold.art.SensorEvent;

import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import de.jaetzold.util.Queue;
import de.jaetzold.util.Debug;

/**	Base class for easier implementation of the interface <code>SensorPeer</code>.
	The only thing really required is to implement the method <code>getValue()</code>. Maybe the properties <code>connected</code>, <code>precision</code> and <code>deliversFloatingPoint</code> need an implementation too to produce correct results.
	<p>
	For a more efficient <code>update()</code>-method it is advisable to assign a <code>de.jaetzold.Queue</code> to <code>eventQueue</code>. This queue then may hold instances of <code>SensorEvent</code>s and must be emptied regularly by some thread which passes the events in the queue to the method <code>processEvent(SensorEvent)</code>.
	<p>
	For a more efficient handling of Events it is advisable to implement <code>enable/disableEvent(SensorEventDescriptor)</code>.
	*/
public abstract class BaseSensorPeer implements SensorPeer {
	private final Debug debug = new Debug("de.jaetzold.art.BaseSensorPeer");
	{ if(debug.debug) debug.objectIdentifier = debug.getIdentifierFor(this); }
	
	/**	Calls <code>(int)getValue()</code>.
		@see #getValue()
		*/
	public int getIntValue() {
		return (int)getValue();
	}
	
	/**	Returns true.
		*/
	public boolean isConnected() {
		return true;
	}
	/**	Returns true.
		*/
	public boolean getDeliversFloatingPoint() {
		return true;
	}
	
	/**	Returns 0.
		*/
	public double getPrecision() {
		return 0;
	}

	public void enableEvent(SensorEventDescriptor eventDescriptor) {
	}
	public void disableEvent(SensorEventDescriptor eventDescriptor) {
	}

	protected PropertyChangeSupport changes =
				new PropertyChangeSupport(this);
	public void addPropertyChangeListener(PropertyChangeListener l) {
		changes.addPropertyChangeListener(l);
	}
	public void removePropertyChangeListener(PropertyChangeListener l) {
		changes.removePropertyChangeListener(l);
	}

	/**	If this is non-null events generated by <code>update()</code> are posted there. Otherwise they get delivered by some Thread controlled by the implementation of <code>update()</code>.
		*/
	protected Queue eventQueue;
	private SensorEventDescriptor updateSensorEventDescriptor =
											new AllSensorEventDescriptor();
	// post a new event to the event-queue
	// instead of delivering it by yourself
	// because otherwise any Thread calling update() would deliver the event
	
	// TBD: should update accept a propagationID as parameter? What if an Event results in an update which results in an Event which results ... asf.
	
	// Another speciality has to be considered:
	// If someone calls update on a Sensor, every Sensor connected to the peer gets the new Event.
	public synchronized void update() {
		if(debug.debug) {
			debug.printInfo("called update from:\n"
							+debug.stackTraceToString(
								new Exception().fillInStackTrace()));
		}
		try {
			double value = getValue();
			final SensorEvent se = new SensorEvent(this,
											updateSensorEventDescriptor,
											System.currentTimeMillis(),
											value,
											value,
											getDeliversFloatingPoint(),
											null);
			Queue localQueue = eventQueue;
			if(localQueue != null) {
				localQueue.post(se);
			} else {
				// generate new Thread to deliver the Event
				new Thread(
					new Runnable() {
						public void run() {
							if(debug.debug) {
								debug.printInfo("delivering " +se
											+" with a newly generated Thread");
							}
							try {
								processEvent(se);
							} catch(RuntimeException re) {
								if(debug.debug) {
									debug.printInfo("caught " +re
												+" while delivering:\n"
												+debug.stackTraceToString(re));
								}
							}
						}
					}
				).start();
			}
		} catch(Exception e) {
			if(debug.debug) {
				debug.printInfo("update resulted in Exception " +e);
			}
		}
	}
	
	private SensorEvent lastEvent;
	private Object lastEventMonitor = new Object();
	public void processEvent(SensorEvent event) {
		SensorEvent last = null;
		// make a copy to be thread-safe without too much blocking
		synchronized(lastEventMonitor) {
			last = lastEvent;
			lastEvent = event;
		}

		if(debug.debug) {
			debug.printInfo("firing PropertyChangeEvent because of "
							+event +" from " +event.getSource());
		}
		if(changes.hasListeners("value")) {
			PropertyChangeEvent pce =
					new PropertyChangeEvent(this,
											"value",
											last,
											event);
			pce.setPropagationId(event.getPropagationId());
			changes.firePropertyChange(pce);
		}
	}
}
