package de.jaetzold.art.platform;

import de.jaetzold.art.RobotInterface;
import de.jaetzold.art.SensorPort;
import de.jaetzold.art.Sensor;
import de.jaetzold.art.Port;

import de.jaetzold.art.SensorPort;
import de.jaetzold.art.Sensor;
import de.jaetzold.art.RobotInterface;

/**	Base class for easy implementation of the interface <code>SensorPort</code>.
	The only thing required is to feed the contructor with the desired parameters and implement the factory-method for the peers: <code>getPeerFor(Sensor)</code>
	*/
public abstract class BaseSensorPort implements SensorPort {
	/**	Default is 0.
		*/
	protected int portNumber;				// numeric identifier for this Port
	public int getPortNumber() {
		return portNumber;
	}
	/**	Default is a new plain Object.
		*/
	protected Object identifier;
	public Object getPortID() {
		return identifier;
	}
/*	protected int minimalPollingInterval;	// in microseconds
	protected boolean isPollingIntervalSettable;	// a configurable interval?
	protected boolean implementsHardwareEvents;
	//protected int getResistanceForValue(int value);
	//protected int getVoltageForValue(int value);
	protected int maximumVoltage;		// maximum voltage in millivolts which can be measured without damaging the interface
//	maybe configure the following with additional actuators
//	protected boolean powerable;		// whether the input can supply power
//	protected boolean isPowered;
	protected int granularity;			// in maximumVoltage/granulariy millivolt
	protected int accuracy;				// in +-millivolts
*/

	/**	Default is Double.NEGATIVE_INFINITY.
		*/
	protected double minimumValue;
	public double getMinimumValue() {
		return minimumValue;
	}
	protected double maximumValue;
	/**	Default is Double.POSITIVE_INFINITY.
		*/
	public double getMaximumValue() {
		return maximumValue;
	}
	/**	Default is -1 which means all 64 bits are used.
		*/
	protected long granularity;
	public long getGranularity() {
		return granularity;
	}
	
	/**	Default is an array of length 0.
		*/
	protected Port[] excludedPorts;	// which Ports cannot be used when use
	public Port[] getExcludedPorts() {
		// make a copy, so that no one can temper with our own array
		Port[] result = new Port[excludedPorts.length];
		System.arraycopy(excludedPorts, 0, result, 0, result.length);
		return result;
	}
	
	public boolean conformsTo(Object identifier) {
		return this.identifier.equals(identifier);
	}
	
	/**	Returns <code>identifier.toString()</code>
		@see #identifier
		*/
	public String toString() {
		return identifier.toString();
	}
	
	/**	The interface-representation we're associated with.
		*/
	protected RobotInterface iface;

	/** Constructs a new SensorPort with dumb default values for it's properties. The properties are described in more detail in the description of the interface <code>SensorPort</code>
		@see de.jaetzold.art.SensorPort
		*/
	public BaseSensorPort() {
		this( null,
				null,
				0,
				Double.NEGATIVE_INFINITY,
				Double.POSITIVE_INFINITY,
				// -1 means: all (64) bits are used
				// (parameter is interpreted unsigned)
				-1,
				null);
	}
	/** Constructs a new SensorPort with the given values for it's properties. The properties are described in more detail in the description of the interface <code>SensorPort</code>
		@see de.jaetzold.art.SensorPort
		*/
	public BaseSensorPort(RobotInterface iface,
							Object identifier,
							int portNumber,
							double minimumValue,
							double maximumValue,
							long granularity,
							Port[] excludedPorts) {
		this.iface = iface;
		if(identifier == null) {
			this.identifier = new Object();
		} else {
			this.identifier = identifier;
		}
		this.portNumber = portNumber;
		this.minimumValue = minimumValue;
		this.maximumValue = maximumValue;
		this.granularity = granularity;
		if(excludedPorts == null) {
			this.excludedPorts = new Port[0];
		} else {
			this.excludedPorts = excludedPorts;
		}
	}
	
	/**	This is the Factory-method for creating peers. It must be implemented by non-abstract subclasses.
		*/
	public abstract SensorPeer getPeerFor(Sensor sensor);
	
	private Object connectMonitor = new Object();
	private boolean connecting = false;
	private boolean disconnecting = false;
	public void connectWith(Sensor sensor) {
		synchronized(connectMonitor) {
			if(connecting == true) {
				return;
			}
			connecting = true;
			sensor.connectWith(this);
			connecting = false;
		}
	}
	
	public void disconnect(Sensor sensor) {
		synchronized(connectMonitor) {
			if(disconnecting == true) {
				return;
			}
			disconnecting = true;
			sensor.disconnect();
			disconnecting = false;
		}
	}
	
	public boolean equals(Object o) {
		boolean result = false;
		if(o != null && o.getClass() == getClass()) {
			BaseSensorPort port = (BaseSensorPort)o;
			if(((iface != null && iface.equals(port.iface))
						|| iface == port.iface)
					&& identifier.equals(port.identifier)
					&& portNumber == port.portNumber
					&& minimumValue == port.minimumValue
					&& maximumValue == port.maximumValue
					&& granularity == port.granularity
					&& java.util.Arrays.equals(	excludedPorts,
												port.excludedPorts)) {
				result = true;
			}
		}
		return result;
	}
}
