package de.jaetzold.art.platform;

import java.io.FilterInputStream;
import java.io.InputStream;
import java.io.IOException;

import de.jaetzold.util.StringObservable;	// just for debug-messages
import de.jaetzold.util.ByteArray;			// some useful methods for handling byte-arrays

/** Base Class for reading some kind of packets from an InputStream. Packets means a sequence of at least 1 bytes.
*/
public class PacketInputStream extends FilterInputStream {
	/** Skip forward in the InputStream as far as possible until no bytes are available. This method will only return if a call to available() to the associated InputStream returns 0 or a value not less than the last time (to prevent this method from running forever). In the latter case the method returns false, true otherwise.
	*/
	public boolean flush() {
		int count = 0;
		try {
			long last = Integer.MAX_VALUE+1; // always greater than result of available()
			while((count = available())>0) {
				skip(count);
				if(count >= last) {
					return false;
				}
				last = count;
			}
		} catch(IOException ioe) {
			if(debug) {
				System.err.println(getClass().getName() +": ");
				ioe.printStackTrace();
			}
		}
		return true;
	}

	public interface PacketHeaderChecker {
		public boolean beginsWithHeader(byte[] readBuffer,
										int readBOff,
										int readBLen)
				throws CorruptPacketException;
		public int getHeaderSkipLen();
	}
		
	protected PacketHeaderChecker packetHeaderChecker;
	public PacketHeaderChecker getPacketHeaderChecker() {
		return packetHeaderChecker;
	}
	public void setPacketHeaderChecker(PacketHeaderChecker packetHeaderChecker) {
		if(packetHeaderChecker==null) {
			packetHeaderChecker = new DefaultPacketHeaderChecker();
		}
		this.packetHeaderChecker = packetHeaderChecker;
	}

	public interface PacketDataFilter {
		public byte[] filterBufferForPacketData(byte[] packet,
												int packetLen,
												byte[] readBuffer,
												int readBOff,
												int readBLen)
				throws CorruptPacketException;
		public int getPacketLen();
		public boolean isDataComplete(	byte[] packet,
										int packetLen,
										byte[] readBuffer,
										int readBOff,
										int readBLen);
		public int getDataSkipLen();
	}
	
	protected PacketDataFilter packetDataFilter;
	public PacketDataFilter getPacketDataFilter() {
		return packetDataFilter;
	}
	public void setPacketDataFilter(PacketDataFilter packetDataFilter) {
		if(packetDataFilter==null) {
			packetDataFilter = new DefaultPacketDataFilter();
		}
		this.packetDataFilter = packetDataFilter;
	}

	public interface PacketTrailerChecker {
		public boolean isComplete(	byte[] packet,
									int packetLen,
									byte[] readBuffer,
									int readBOff,
									int readBLen)
				throws CorruptPacketException;
		public int getTrailerSkipLen();
		public byte[] removeTrailer(byte[] packet, int packetLen); 
	}
	
	protected PacketTrailerChecker packetTrailerChecker;
	public PacketTrailerChecker getPacketTrailerChecker() {
		return packetTrailerChecker;
	}
	public void setPacketTrailerChecker(PacketTrailerChecker packetTrailerChecker)
	{
		if(packetTrailerChecker==null) {
			packetTrailerChecker = new DefaultPacketTrailerChecker();
		}
		this.packetTrailerChecker = packetTrailerChecker;
	}

	public static class DefaultPacketHeaderChecker
			implements PacketHeaderChecker
	{
		public DefaultPacketHeaderChecker() {
			this(null);
		}
		
		public DefaultPacketHeaderChecker(byte[] header) {
			if(header == null) {
				this.header = new byte[0];
			} else {
				this.header = header;
			}
		}
		
		protected byte[] header;
		public byte[] getHeader() {
			return header;
		}
		public void setHeader(byte[] header) {
			this.header = header;
		}

		protected int lastHeaderSkipLen;
		
		public boolean beginsWithHeader(byte[] readBuffer,
										int readBOff,
										int readBLen) {
			lastHeaderSkipLen=0;
			// no header -> packet begins at the beginning of the stream
			if(header.length == 0) {
				return true;
			}
			
			while(lastHeaderSkipLen<readBLen) {
				// look for the first byte matching the first from the header
				if(readBuffer[(readBOff+lastHeaderSkipLen)
								%readBuffer.length]
						== header[0]) {
					// if there is not enough data left, there is no need to check now
					if(readBLen-lastHeaderSkipLen<header.length) {
						return false;
					}
					
					//first header-byte doesn't need to be read again
					lastHeaderSkipLen++;
					
					// look whether we can match the complete header
					for(int i=1; i<header.length; i++) {
						if(readBuffer[(readBOff+lastHeaderSkipLen+i-1)
										%readBuffer.length]
								!= header[i]) {
							return false;
						}
					}
					
					// the header has been found, skip it completely
					lastHeaderSkipLen += header.length-1;
					
					return true;
				} else {
					lastHeaderSkipLen++;
				}
			}
			
			return false;
		}
		
		public int getHeaderSkipLen() {
			return lastHeaderSkipLen;
		}
	}
	public static class DefaultPacketDataFilter implements PacketDataFilter {
		private int lastPacketLen;
		private int lastDataSkipLen;
		/** 
			@see de.jaetzold.util.ByteArray#manageBufferSize(byte[],int,int)
			*/
		public byte[] filterBufferForPacketData(byte[] packet,
												int packetLen,
												byte[] readBuffer,
												int readBOff,
												int readBLen)
				throws CorruptPacketException
		{
			lastDataSkipLen = 0;
			lastPacketLen = packetLen;
			
			byte[] result = ByteArray.manageBufferSize(	packet,
														packetLen,
														readBLen);
			while(readBLen-lastDataSkipLen>=1
					&& !isDataComplete(	result, lastPacketLen,
										readBuffer,
										(readBOff+lastDataSkipLen)
											%readBuffer.length,
										readBLen-lastDataSkipLen)) {
				
				result = ByteArray.manageBufferSize(result,
													packetLen,
													readBLen);
				
				result[lastPacketLen] = readBuffer[(readBOff+lastDataSkipLen)
													%readBuffer.length];
				lastPacketLen++;
				lastDataSkipLen++;
			}

			return result;
		}
		
		protected void setPacketLen(int packetLen) {
			lastPacketLen = packetLen;
		}
		public int getPacketLen() {
			return lastPacketLen;
		}

		public boolean isDataComplete(	byte[] packet,
										int packetLen,
										byte[] readBuffer,
										int readBOff,
										int readBLen) {
			if(packet != null && packetLen>0) {
				return true;
			} else {
				return false;
			}
		}
				
		protected void setDataSkipLen(int dataSkipLen) {
			lastDataSkipLen = dataSkipLen;
		}
		public int getDataSkipLen() {
			return lastDataSkipLen;
		}
	}
	public static class PacketDataComplementFilter
			extends DefaultPacketDataFilter
	{
		private int lastPacketLen;
		private int lastDataSkipLen;
		/** 
			@see de.jaetzold.util.ByteArray#manageBufferSize(byte[],int,int)
			*/
		public byte[] filterBufferForPacketData(byte[] packet,
												int packetLen,
												byte[] readBuffer,
												int readBOff,
												int readBLen)
				throws CorruptPacketException
		{
			lastDataSkipLen = 0;
			lastPacketLen = packetLen;
			
			try {
				byte[] result = ByteArray.manageBufferSize(packet, packetLen, readBLen);;

				while(readBLen-lastDataSkipLen>=2
						&& !isDataComplete(	result, lastPacketLen,
											readBuffer,
											(readBOff+lastDataSkipLen)
												%readBuffer.length,
											readBLen-lastDataSkipLen)) {
					result = ByteArray.manageBufferSize(result, packetLen, readBLen);
					try {
						if(readBuffer[(readBOff+lastDataSkipLen)%readBuffer.length]
								== ~ readBuffer[(readBOff+1+lastDataSkipLen)
													%readBuffer.length]) {
							result[lastPacketLen] =
								readBuffer[(readBOff+lastDataSkipLen)
												%readBuffer.length];
							lastPacketLen++;
							// skip 1, the other in "finally"
							lastDataSkipLen++;
						} else {
							throw new CorruptPacketException("Error in complementary byte number " +(lastPacketLen) +".");
						}
					} finally {
						// always skip at least 1 byte
						lastDataSkipLen++;
					}
				}

				return result;
			} finally {
				// copy into Variables of superclass
				setDataSkipLen(lastDataSkipLen);
				setPacketLen(lastPacketLen);
			}
		}
	}
	public static class DefaultPacketTrailerChecker
			implements PacketTrailerChecker
	{
		protected int lastPacketLen;
		
		public boolean isComplete(	byte[] packet,
									int packetLen,
									byte[] readBuffer,
									int readBOff,
									int readBLen) {
			lastPacketLen = packetLen;
			if(packet != null && packetLen>0) {
				return true;
			} else {
				return false;
			}
		}
		
		public int getTrailerSkipLen() {
			return 0;
		}
		
		public byte[] removeTrailer(byte[] packet,
									int packetLen) {
			return packet;
		}
	}
	
	public static class PacketTrailerChecksumByteChecker
			implements PacketTrailerChecker 
	{
		public boolean isComplete(	byte[] packet,
									int packetLen,
									byte[] readBuffer,
									int readBOff,
									int readBLen)
				throws CorruptPacketException
		{
			if(packet != null && packetLen>0) {
				byte checksum = 0;
				for(int i=0; i<packetLen-1; i++) {
					checksum += packet[i];
				}
				if(checksum == packet[packetLen-1]) {
					return true;
				} else {
					throw new CorruptPacketException("Error in checksum byte: "
							+StringObservable.toHexString(new byte[]{checksum})
							+"!="
							+StringObservable.toHexString(new byte[]{packet[packetLen-1]})
							+".");
				}
			} else {
				return false;
			}
		}
		
		public int getTrailerSkipLen() {
			return 0;
		}
		
		public byte[] removeTrailer(byte[] packet,
									int packetLen) {
			// remove checksum byte from the end
			// copy contents into a new array of the appropriate length:
			byte[] result = ByteArray.copy(packet, 0, packetLen-1);
			return result;
		}
	}
	
	public PacketInputStream(InputStream source) {
		this(	source,
				(byte[]) null,
				new DefaultPacketDataFilter(),
				new DefaultPacketTrailerChecker());
	}
	public PacketInputStream(	InputStream source,
								byte[] header,
								PacketDataFilter data,
								PacketTrailerChecker trailer) {
		this(	source,
				new DefaultPacketHeaderChecker(header),
				data,
				trailer);
	}
	public PacketInputStream(	InputStream source,
								PacketHeaderChecker header,
								PacketDataFilter data,
								PacketTrailerChecker trailer) {
		super(source);
		setPacketHeaderChecker(header);
		setPacketDataFilter(data);
		setPacketTrailerChecker(trailer);
	}

	// not implemented yet!
	protected int defaultReadTimeout = 0;	// 0 or less means disabled
	public int getDefaultReadTimeout() {
		return defaultReadTimeout;
	}
	public void setDefaultReadTimeout(int defaultReadTimeout) {
		this.defaultReadTimeout = defaultReadTimeout;
	}
	
	// don't set this to 0 because at least ibm's linux Comm does not return
	// from a read waiting for data if the port gets closed meanwhile
	protected int pollingPause = 1;	// 0 or less means disabled
	public int getPollingPause() {
		return pollingPause;
	}
	public void setPollingPause(int pollingPause) {
		this.pollingPause = pollingPause;
	}

	private byte[] readBuffer = new byte[8];
	/** Beginning of not-processed data in readBuffer. */
	private int readBOff = 0;
	/** Number of not-processed bytes in readBuffer. */
	private int readBLen = 0;

	// synchronized because of global readBuffer
	public synchronized byte[] readPacket()
			throws IOException, CorruptPacketException
	{
		return readPacket(defaultReadTimeout);
	}
	// synchronized because of global readBuffer
	// timeout not implemented yet!
	public synchronized byte[] readPacket(int timeout)
			throws IOException, CorruptPacketException
	{
		byte[] packet = null;
		int packetLen = 0;
		boolean gotHeader = false;		// have we seen the header already?
		boolean packetDataComplete = false;
		boolean packetComplete = false;
		int minBufferLen = 0;
		
		do {
			if(debug) {
				if(readBOff>=readBuffer.length) {
					// mustn't be possible:
					throw new Error(getClass().getName() +": Buffer offset out of bounds.");
				}
				if(readBLen<0) {
					// mustn't be possible:
					throw new Error(getClass().getName() +": Buffer length < 0.");
				}
			}			

			if(debug) System.out.println(getClass().getName()
										+": Filling Buffer ...");
			// fill readBuffer as much as possible:
			// but it must contain at least 1 byte more than before
			minBufferLen = Math.min(readBuffer.length, readBLen+1);
			int available = 0;
			while(((available = in.available()) > 0
						&& readBLen < readBuffer.length)
					|| readBLen < minBufferLen) {
				if(debug) System.out.println(getClass().getName()
										+": reading from " +in
										+", available==" +available);
				if(available==0 && pollingPause>0) {
					try {
						if(debug) System.out.println(getClass().getName()
										+": sleeping for "
										+pollingPause
										+" because nothing available.");
						Thread.currentThread().sleep(pollingPause);
					} catch(InterruptedException ie) {
					}
					continue;
				}
				
				int num = in.read(readBuffer,
								(readBOff+readBLen) % readBuffer.length,
								readBOff+readBLen < readBuffer.length
									// beginning of free space lies after readBOff:
									? readBuffer.length - (readBOff+readBLen)
									// beginning of free space lies before readBOff:
									: readBuffer.length - readBLen);
				if(num > 0) {
					readBLen += num;
					if(debug) {
						if(readBLen>readBuffer.length) {
							// mustn't be possible:
							throw new Error(getClass().getName() +": Buffer overflow.");
						}
						System.out.println(getClass().getName()
								+": read " +num
								+": " +StringObservable
										.toHexString(	readBuffer,
														(readBOff+readBLen-num)%readBuffer.length,
														num,
														" ")
								+" -> not processed " +readBLen
								+": " +StringObservable
										.toHexString(	readBuffer,
														readBOff,
														Math.min(readBuffer.length-readBOff, readBLen),
														" ")
								+(	readBOff+readBLen>readBuffer.length
									? " " +StringObservable
										.toHexString(	readBuffer,
														0,
														readBOff+readBLen-readBuffer.length,
														" ")
									: "")
								);
					}
				}
				// if buffer full, increase its size
				if(readBLen >= readBuffer.length) {
					byte[]tmp = readBuffer;
					readBuffer = new byte[readBuffer.length*2];
					System.arraycopy(tmp, readBOff,
									readBuffer, 0,
									tmp.length-readBOff);	
					System.arraycopy(tmp, 0,
									readBuffer, tmp.length-readBOff,
									readBOff);
					if(debug) {
						System.out.println(getClass().getName()
								+": increased buffer to " +readBuffer.length
								+": " +StringObservable
										.toHexString(tmp,
												readBOff,
												Math.min(tmp.length-readBOff,
														readBLen),
												" ")
								+(	readBOff+readBLen>tmp.length
									? " | " +StringObservable
											.toHexString(tmp,
														0,
														readBOff+readBLen
																-tmp.length,
														" ")
									: "")
								+" -> " +StringObservable
										.toHexString(readBuffer,
													0,
													readBLen,
													" ")
								);
					}

					readBOff = 0;
				}
			}
			
			// First thing is to check for the header, if we haven't got one already
			if(!gotHeader) {
				if(debug) System.out.println(getClass().getName()
											+": Searching for header ...");
				gotHeader = packetHeaderChecker.beginsWithHeader(readBuffer, readBOff, readBLen);
				int skipLen = packetHeaderChecker.getHeaderSkipLen();

				// the delegates are not allowed to skip over not-seen data
				if(skipLen>readBLen) {
					throw new IndexOutOfBoundsException(getClass().getName()
								+": beginsWithHeader() resulted in true, but headerSkipLen=" +skipLen +">readBLen=" +readBLen +". "
								+"There is an Error in the packetHeaderChecker!");
				}
				// skip over what contains no relevant data
				skip(skipLen);
			}
			
			// If there is still no header, continue from beginning
			if(!gotHeader) {
				continue;
			}
			
			if(!packetDataComplete) {	
				if(debug) {
					System.out.println(getClass().getName()
											+": Filtering PacketData ...");
				}
				// filter the remaining bytes for packet-content
				packet = packetDataFilter.filterBufferForPacketData(packet, packetLen, readBuffer, readBOff, readBLen);
				packetLen = packetDataFilter.getPacketLen();
				packetDataComplete = packetDataFilter.isDataComplete(packet, packetLen, readBuffer, readBOff, readBLen);
				
				int skipLen = packetDataFilter.getDataSkipLen();
				
				// the delegates are not allowed to skip over not-seen data
				if(skipLen>readBLen) {
					throw new IndexOutOfBoundsException(getClass().getName()
								+": beginsWithHeader() resulted in true, but headerSkipLen=" +skipLen +">readBLen=" +readBLen +". "
								+"There is an Error in the packetHeaderChecker!");
				}
				// skip over what contains no relevant data anymore
				skip(skipLen);
			}
			
			// If there is still data missing, continue filling buffer
			if(!packetDataComplete) {
				if(debug) {
					System.out.println(getClass().getName()
									+": packet data not complete, continue reading");
				}			
				// there is no trailer to check for by now, so continue always
				continue;
			}
			
			if(debug) {
				System.out.println(getClass().getName()
											+": Searching for trailer ...");
			}
			// look for trailer, checksum, ...
			packetComplete = packetTrailerChecker.isComplete(packet, packetLen, readBuffer, readBOff, readBLen);
			
			int skipLen = packetTrailerChecker.getTrailerSkipLen();

			// the delegates are not allowed to skip over not-seen data
			if(skipLen>readBLen) {
				throw new IndexOutOfBoundsException(getClass().getName()
							+": beginsWithHeader() resulted in true, but headerSkipLen=" +skipLen +">readBLen=" +readBLen +". "
							+"There is an Error in the packetHeaderChecker!");
			}
			// skip over what contains no relevant data anymore
			skip(skipLen);

			// If there is still (part of) the trailer missing, continue filling buffer
			if(!packetComplete) {
				if(debug) {
					System.out.println(getClass().getName()
									+": packet trailer not complete, continue reading");
				}				
				// there is no trailer to check for by now, so continue always
				continue;
			}
			
			// if we reach this point packetComplete must be true
			// remove any data from the packet which was only needed to identify the trailer
			packet = packetTrailerChecker.removeTrailer(packet, packetLen);
			packetLen = Math.min(packetLen, packet.length);
			 
		} while(!packetComplete);

		// packet is now complete and contained no errors
		// copy contents into a new array of the appropriate length:
		byte[] result = packet;
		if(packet.length!=packetLen) {
			result = ByteArray.copy(packet, 0, packetLen);
		}

		if(debug) {
			System.out.println(getClass().getName() +": "
								+" returning packet: " +StringObservable.toHexString(result));
		}
		return result;
	}
	
	private long skipSequenceNo;
	/**	Skips over n bytes in the input-buffer. If there are less than n bytes in the buffer, this method also calls skip(long) on the associated InputStream for the rest of the bytes. This Method is synchronized because of a global input-buffer.
		@return The total number of bytes skipped.
		@exception IOException, only when the call to skip(long) on the associated InputStream throws one.
		*/
	public synchronized long skip(long n)
			throws IOException
	{
		if(n<1) {
			return 0;
		}
		
		long skipLen = Math.min(readBLen, n);

		if(debug) {
			System.out.println(getClass().getName()
					+": skipping " +n +"->["
					+StringObservable
						.toHexString(readBuffer,
									readBOff,
									Math.min(readBuffer.length-readBOff,
											(int)skipLen),
									" ")
					+(	readBOff+skipLen>readBuffer.length
						? " " +StringObservable
								.toHexString(readBuffer,
											0,
											readBOff+(int)skipLen
													-readBuffer.length,
											" ")
						: "")
					+"], leaving "
					+(readBLen-skipLen)
					+". Call " +skipSequenceNo);
			System.out.flush();
			skipSequenceNo++;
		}

		readBOff+=skipLen;
		readBLen-=skipLen;
		readBOff = readBOff % readBuffer.length;
		
		if(skipLen!=n) {
			skipLen += super.skip(n-skipLen);
		}
		
		return skipLen;
	}

	/** Returns the number of bytes in the input-buffer added to the result of super.available().
		*/
	public int available()
    		throws IOException
	{
		return readBLen+super.available();
	}
	
	private static boolean debug;
	static {
		String v = null;
		v = System.getProperty("de.jaetzold.art.PacketInputStream"+".debug");
		if(v==null) {
			v = System.getProperty("de.jaetzold.art"+".debug");
		}
		if(v==null) {
			v = System.getProperty("de.jaetzold"+".debug");
		}
		if(v==null) {
			v = System.getProperty("de"+".debug");
		}
		if(v==null) {
			v = System.getProperty("debug");
		}
		// results to false if v!="true" (ignoring case)
		debug = Boolean.valueOf(v).booleanValue();
	}
}
