package de.jaetzold.art.platform;

import de.jaetzold.art.RobotInterface;
import de.jaetzold.art.RobotInterfaceDefinition;
import de.jaetzold.art.RobotInterfaceStringDefinition;
import de.jaetzold.art.SensorPort;
import de.jaetzold.art.ActuatorPort;
import de.jaetzold.art.Port;

import javax.comm.CommPortIdentifier;
import java.util.Enumeration;
import java.util.Vector;

/**	This base class implements methods from <code>RobotInterface</code> for retrieving <code>Port</code>-instances through specific arguments. They all rely on the Factory-Methods <code>getSensorPorts()</code> and <code>getActuatorPorts()</code> and these are therefore the only Port-related methods to implement.
<p>
Additionally it defines some useful methods for handling <code>SerialPort</code>s from the Java Communications API because it is not unusual for interface-hardware to be controllable through a serial connection.
	*/
public abstract class RobotInterfaceAdapter implements RobotInterface {
	/**	Contruct names of serial ports. This is by now only done for the usual type of <code>RobotInterfaceStringDefinition</code>.
		The parameter interfaceIds tells the method which part at the beginning of the <code>RobotInterfaceStringDefinition</code> may just be an id which has nothing to do with the serial port identifier following it. For a more detailed description of these Strings see <code>RobotInterfaceStringDefinition</code>
		@see de.jaetzold.art.RobotInterfaceStringDefinition
		*/
	protected static String[] getSerialPortNames(
								RobotInterfaceDefinition[] defs,
								String[] interfaceIds) {
		Vector portNames = new Vector();
		
		String[] spn = getSerialPortNames();
		
		// generate explicit ports out from defs like FT or FT<i>
		for(int i=0; i<defs.length; i++) {
			if(defs[i] != null
					&& defs[i] instanceof RobotInterfaceStringDefinition) {
				String id = ((RobotInterfaceStringDefinition)defs[i])
													.getInterfaceString();
				for(int j=0; j<interfaceIds.length; j++) {
					if(id.startsWith(interfaceIds[j])) {
						if(id.length() <= interfaceIds[j].length()) {
							for(int k=0; k<spn.length; k++) {
								portNames.add(spn[k]);
							}
						} else {
							try {
								int portNumber = Integer.parseInt(
										id.substring(interfaceIds[j].length()));
								String[] spnNumber =
										getSerialPortNames(portNumber);
								for(int k=0; k<spnNumber.length; k++) {
									portNames.add(spnNumber[k]);
								}
							} catch(NumberFormatException nfe) {
								portNames.add(
										id.substring(interfaceIds[j].length()));
							}
						}
					}
				}
			}
		}
		// remove all "doubles" 
		for(int i=0; i<portNames.size(); i++) {
			boolean remove = false;
			for(int j=0; j<i; j++) {
				if(portNames.elementAt(j)
						.equals(portNames.elementAt(i))) {
					remove = true;
				}
			}
			if(remove) {
				portNames.removeElementAt(i);
				i--;
			}
		}
		String[] result = new String[portNames.size()];
		portNames.copyInto(result);
		return result;
	}
	
	/**	Contructs names of serial ports out of a number.
		The names returned include COM-n-, /dev/ttyS-n- and serial-n- where -n- is replaced by the argument portNumber. The method also returns the portName in the list of available CommPorts from JavaComm which is at the position number <code>portnumber</code>.
		@see javax.comm
		*/
	protected static String[] getSerialPortNames(int portNumber) {
		String[] result;
		if(portNumber>0) {
			int i = 0;
			String[] javaCommNames = getSerialPortNames();
			if(portNumber<=javaCommNames.length) {
				// default-JavaComm with this index exists
				result = new String[4];
				result[i++] = javaCommNames[portNumber-1];
			} else {
				result = new String[3];
			}
			// windows, ibm-linux
			result[i++] = "COM" +portNumber;
			// rxtx-linux
			result[i++] = "/dev/ttyS" +(portNumber-1);
			// TINI
			result[i++] = "serial" +(portNumber-1);
		} else {
			result = new String[0];
		}
		
		return result;
	}
	
	/**	Retunrs all names of serial ports known to JavaComm.
		@see javax.comm
		*/
	protected static String[] getSerialPortNames() {
		Enumeration e = CommPortIdentifier.getPortIdentifiers();
		Vector names = new Vector();
		while(e.hasMoreElements()) {
			CommPortIdentifier cpi = (CommPortIdentifier)e.nextElement();
			if(cpi.getPortType() == CommPortIdentifier.PORT_SERIAL) {
				names.add(cpi.getName());
			}
		}
		String[] result = new String[names.size()];
		names.copyInto(result);
		return result;
	}

	public ActuatorPort getActuatorPorts(int index) {
		return getActuatorPorts()[index];
	}
	
	public SensorPort getSensorPorts(int index) {
		return getSensorPorts()[index];
	}
	
	public Port getPort(Object identifier) {
		Port[] ports = getActuatorPorts();
		for(int i=0; i<ports.length; i++) {
			if(ports[i].conformsTo(identifier)) {
				return ports[i];
			}
		}
		ports = getSensorPorts();
		for(int i=0; i<ports.length; i++) {
			if(ports[i].conformsTo(identifier)) {
				return ports[i];
			}
		}
		return null;
	}
	
	// TBD: define that for ports 
/*	/** Tells where events are generated. If this Method returns true events are
	generated in the Hardware (or actually by a CPU contained in it) and not
	inside the JVM where this Object lives in. Otherwise this Object might poll
	the Interface actively to look whether some event-condition did become true.
	May be true or false just for some events ...
	@see #addSensorListener()
	*/
/*	public boolean implementsHardwareEvents() { return false; }*/
}
