package de.jaetzold.art.platform.fischertechnik;

import de.jaetzold.util.Debug;
import de.jaetzold.util.Event;
import de.jaetzold.util.Task;
import de.jaetzold.util.TaskPerformer;
import de.jaetzold.util.Queue;
import de.jaetzold.util.RequeueTask;
import de.jaetzold.util.Range;
import de.jaetzold.util.SingleRange;
import de.jaetzold.util.TimeoutException;
import de.jaetzold.util.StringObservable;
import de.jaetzold.util.ByteArray;

import de.jaetzold.art.RobotInterface;
import de.jaetzold.art.RobotInterfaceDefinition;
import de.jaetzold.art.RobotInterfaceStringDefinition;
import de.jaetzold.art.Sensor;
import de.jaetzold.art.LightSensor;
import de.jaetzold.art.Actuator;
import de.jaetzold.art.Port;
import de.jaetzold.art.SensorPort;
import de.jaetzold.art.ActuatorPort;
import de.jaetzold.art.SensorEvent;
import de.jaetzold.art.SensorEventDescriptor;
import de.jaetzold.art.platform.RobotInterfaceAdapter;
import de.jaetzold.art.platform.PacketInputStream;
import de.jaetzold.art.platform.PacketFilter;
import de.jaetzold.art.platform.CorruptPacketException;
import de.jaetzold.art.platform.BaseSensorPeer;
import de.jaetzold.art.platform.BaseSensorPort;
import de.jaetzold.art.platform.BaseActuatorPort;
import de.jaetzold.art.platform.SensorPeer;
import de.jaetzold.art.platform.ActuatorPeer;
import de.jaetzold.art.platform.LightSensorPeer;

import javax.comm.CommPortIdentifier;
import javax.comm.SerialPort;
import javax.comm.NoSuchPortException;
import javax.comm.PortInUseException;
import javax.comm.UnsupportedCommOperationException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.util.Map;
import java.util.Hashtable;
import java.util.Vector;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyVetoException;

public class Fischertechnik extends RobotInterfaceAdapter {
	private static Debug debug = new Debug("de.jaetzold.art.Fischertechnik");
	
	protected PacketInputStream inStream;
	protected OutputStream outStream;
	
	protected Queue writerQueue = new Queue();
	protected TaskPerformer writer = new TaskPerformer(writerQueue);
	protected Queue readerQueue = new Queue();
	protected TaskPerformer reader = new TaskPerformer(readerQueue);
	protected PacketFilterTable filterTable = new PacketFilterTable();
	{
		readerQueue.post(new RequeueTask(new ReadPacketTask(filterTable),
										readerQueue));
		writerQueue.post(new RequeueTask(new PollDigitalsTask(),
										writerQueue));
	}
	
	protected Queue eventQueue = new Queue();
	protected Queue notifierQueue = new Queue();
	protected TaskPerformer notifier = new TaskPerformer(notifierQueue);
	{
		notifierQueue.post(new RequeueTask(new CondenseEventsTask(),
											notifierQueue));
	}
	
	class PacketFilterTable	implements	PacketFilter
	{
		protected PacketFilter actualFilter;
		protected Object actualFilterMonitor = new Object();
		public void setActualFilter(PacketFilter actualFilter)
		{
			synchronized(actualFilterMonitor) {
				while(this.actualFilter != null && !this.actualFilter.done()) {
					try {
						actualFilterMonitor.wait();
					} catch(InterruptedException ie) {
					}
				}
			}
			this.actualFilter = actualFilter;
		}
		
		public Event process(byte[] packet)
				throws CorruptPacketException
		{
			synchronized(actualFilterMonitor) {
				if(actualFilter == null) {
					return null;
				}
				Event result = null;
				try {
					result = actualFilter.process(packet);
				} finally {
					actualFilterMonitor.notifyAll();
				}
				return result;
			}
		}
		public boolean done() {
			return false;
		}
	}

	/**	This Task talks directly with the interface-hardware.
		It tries to read values as fast as it can and puts any
		change into a queue for further processing.
		*/
	class ReadPacketTask	implements	de.jaetzold.util.Task
	{
		protected PacketFilter filter;
		ReadPacketTask(PacketFilter filter) {
			this.filter = filter;
		}
		public Event perform()
			throws Exception
		{
			byte[] packet = inStream.readPacket();

			filter.process(packet);

			return null;
		}
	}

	protected class FTPacketDataInputFilter
			extends PacketInputStream.DefaultPacketDataFilter
	{
		protected int packetLength;
		/** The programmer has to take care, that the packetLength gets set at the right time.
		*/
		public void setPacketLength(int packetLength) {
			this.packetLength = packetLength;
		}
		public int getPacketLength() {
			return packetLength;
		}
		
		public byte[] filterBufferForPacketData(byte[] packet,
												int packetLen,
												byte[] readBuffer,
												int readBOff,
												int readBLen)
				throws CorruptPacketException
		{
			// clear anything in the buffer what is probably too old
			// if we start reading a _new_ package only
			// that is important, because there are no packet-id's and so
			// we don't know when which packet begins
			// and therefore only take the _last one_
			int oldStuffLen = 0;
			if(packet == null || packetLen == 0) {
				oldStuffLen = readBLen-this.packetLength;
			}
			try {
				byte[] result =
					super.filterBufferForPacketData(packet,
													packetLen,
													readBuffer,
													(readBOff+oldStuffLen)
														%readBuffer.length,
													readBLen-oldStuffLen);
				return result;
			} finally {
				setDataSkipLen(getDataSkipLen()+oldStuffLen);
			}
		}


		public boolean isDataComplete(	byte[] packet,
										int packetLen,
										byte[] readBuffer,
										int readBOff,
										int readBLen) {
			if(packet != null
					&& packetLen > 0
					&& packetLen >= this.packetLength) {
				return true;
			} else {
				return false;
			}
		}
	}
	
	FTPacketDataInputFilter ftPacketDataInputFilter;
	
	/**	This Task talks directly with the interface-hardware.
		It tries to write values as fast as it can but waits up to 1000 milliseconds
		for an answer to each write.
		*/
	class PollDigitalsTask	implements	de.jaetzold.util.Task,
										PacketFilter
	{
		private boolean undocumented = false;
		private boolean bothAnalog = false;
		private int everyEXCycles = 2;
		private int everyEYCycles = 2;
		// that should be dynamic
		private boolean extension = true;
		
		private static final byte query = (byte)0xC0;
		private static final byte mainModule = (byte)0x01;
		private static final byte extModule = (byte)0x02;
		private static final byte analogEX = (byte)0x04;
		private static final byte analogEY = (byte)0x08;
		
		private int answerPacketLength = 1;
		private int cycleNumberEX = 0;
		private int cycleNumberEY = 0;
		private byte lastQueryByte = 0;
		private boolean doAnalogEX = false;
		private boolean doAnalogEY = false;
		private boolean doExtModule = false;
		
		// to measure the average polling-freq
		private long cycleCount;
		// how long should we look in the past for the long-term average
		private long maxCycleCount = 1<<16;
		
		private Object performingMonitor = new Object();
		private int sendMotorState;
		private int wasMotorState;

		private long[] sendOnMillis = new long[8];		
		private long[] sendOffMillis = new long[8];		
		private long[] wasOnMillis = new long[8];		
		private long[] wasOffMillis = new long[8];		
		private int[] sendPulseDir = new int[8];		
		private int[] wasPulseDir = new int[8];
		private boolean motorSendPulseUpdated;
		private boolean motorUpdateWasPulse;
		
		private ActuatorValueDefinition[] actuatorDefs =
					new ActuatorValueDefinition[8];
		{
			for(int i=0; i<actuatorDefs.length; i++) {
				actuatorDefs[i] = new ActuatorValueDefinition(i+1);
			}
		}

		public Event perform()
				throws	InterruptedException,
						TimeoutException,
						IOException
						
		{
			// this class is otherwise really not thread-safe
			// because of so many used instance variables (do ...)
			synchronized(performingMonitor) {
				// compute querypacket-id and length of corresponding answer
				doAnalogEX = false;
				doAnalogEY = false;
				doExtModule = false;
				cycleNumberEX++;
				cycleNumberEY++;
				if(cycleCount<maxCycleCount) {
					cycleCount++;
				}

				byte queryByte = query;
				answerPacketLength = 1;
				// both is possible, but undocumented and only leads to another
				// byte in the answer for which i don't know what it stands for
				if(extension) {
					queryByte |= extModule;
					doExtModule = true;
					answerPacketLength++;
					if(undocumented) {
						queryByte |= mainModule;
						answerPacketLength++;
					}
				} else {
					queryByte |= mainModule;
				}

				if(everyEXCycles>0 && cycleNumberEX>=everyEXCycles) {
					doAnalogEX = true;
				}
				if(everyEYCycles>0 && cycleNumberEY>=everyEYCycles) {
					doAnalogEY = true;
				}
				// if both want, but undocumented command is not available
				if(doAnalogEX && doAnalogEY && !bothAnalog) {
					// if EX has already last time
					if((lastQueryByte & analogEX) == analogEX) {
						// kick EX out
						doAnalogEX=false;
					} else {
						// kick EY out
						doAnalogEY=false;
					}
				}
				if(doAnalogEX) {	
					queryByte |= analogEX;
					answerPacketLength+=2;
					cycleNumberEX = 0;
				}
				if(doAnalogEY) {	
					queryByte |= analogEY;
					answerPacketLength+=2;
					cycleNumberEY = 0;
				}

				done = true;
				filterTable.setActualFilter(null);
				done = false;

				ftPacketDataInputFilter.setPacketLength(answerPacketLength);

				filterTable.setActualFilter(this);

				synchronized(motorStateMonitor) {
					sendMotorState = shouldMotorState;
					wasMotorState = motorState;
					if(motorUpdatePulse) {
						// onMillis
						System.arraycopy(shouldMotorOnMillis, 0,
										sendOnMillis, 0,
										sendOnMillis.length);
						// offMillis
						System.arraycopy(shouldMotorOffMillis, 0,
										sendOffMillis, 0,
										sendOffMillis.length);
						// pulseDir
						System.arraycopy(shouldMotorPulseDir, 0,
										sendPulseDir, 0,
										sendPulseDir.length);
										
						motorUpdatePulse = false;
						motorSendPulseUpdated = true;
					}
				
					// remember pulses for the next time
					if(motorUpdateWasPulse) {
						// onMillis
						System.arraycopy(motorOnMillis, 0,
										wasOnMillis, 0,
										wasOnMillis.length);
						// offMillis
						System.arraycopy(motorOffMillis, 0,
										wasOffMillis, 0,
										wasOffMillis.length);
						// pulseDir
						System.arraycopy(motorPulseDir, 0,
										wasPulseDir, 0,
										wasPulseDir.length);
						motorUpdateWasPulse = false;
					}
				}
				
				// manipulate sendMotorState according to pulse-lengths
				int nOutputs = extension ? 8 : 4;
				long time = System.currentTimeMillis();
				for(int i=0; i<nOutputs; i++) {
					// both 0 means just use motorState, no pulsing by us
					if(sendOnMillis[i] + sendOffMillis[i] != 0) {
						// left is < 0
						int sendMotorBit = (sendPulseDir[i]<0 ? 1 : 2)<<(i*2);
						int otherBit = (sendPulseDir[i]<0 ? 2 : 1)<<(i*2);
						long currentPulseLength = time - motorLastMillis[i];
						boolean on = false;
						if((wasMotorState & sendMotorBit) > 0) {// was on
							if(currentPulseLength < sendOnMillis[i]
									|| sendOffMillis[i] <= 0) {
								on = true;
							} 
						} else {								// was off
							if(currentPulseLength >= sendOffMillis[i]
									&& sendOnMillis[i]>0) {
								on = true;
							}
						}
						if(on) {
							sendMotorState =
								(sendMotorState | sendMotorBit) & (~ otherBit);
						} else {
							sendMotorState =
								sendMotorState & (~(sendMotorBit | otherBit));
						}
						if(debug.debug) {
							debug.printInfo("switched pulsed motor " +(i+1)
											+" " +(on?"on":"off")
											+" - sendBit " +sendMotorBit
											+", other=" +otherBit
											+" - wasMotorState=" +wasMotorState
											+", sendMotorState="
											+sendMotorState);
						}
					}
				}
				// set last edge-time of all changed motors
				int diff = sendMotorState ^ wasMotorState;
				for(int i=0; i<nOutputs; i++) {
					int bits = 3<<(i*2);
					if((diff & bits) > 0) {
						motorLastMillis[i] = time;
					}
				}

				// construct packet according to query-type (main, ext, undoc)
				byte[] packet = null;
				if(doExtModule) {
					if(undocumented) {
						packet = new byte[]{queryByte,
											(byte)sendMotorState,
											(byte)(sendMotorState>>8),
											(byte)0x00};
					} else {
						packet = new byte[]{queryByte,
											(byte)sendMotorState,
											(byte)(sendMotorState>>8)};
					}
				} else {
					packet = new byte[]{queryByte,
										(byte)sendMotorState};
				}

				lastQueryByte = queryByte;
				try {
					outStream.write(packet);
				} catch(IOException ioe) {
					alive = false;
					connected = false;
					throw ioe;
				}
				
				if(debug.debug) {
					debug.printInfo("written "
								+StringObservable.toHexString(packet));
				}
				synchronized(answerMonitor) {
					if(!done) {
						answerMonitor.wait(1000);
					}
				}
				if(!done) {
					alive = false;
					throw new TimeoutException("Wait for answer timed out.");
				}
				if(debug.debug) {
					debug.printInfo("got answer to "
								+StringObservable.toHexString(packet));
				}
				connected = true;
				alive = true;
				
				// we don't need that anymore
				filterTable.setActualFilter(null);
			}
			
			Thread.currentThread().yield();
			return null;
		}

		protected Object answerMonitor = new Object();
		protected boolean done = false;
		private long lastTime;
		private double longAveragePreciseMillis;
		public Event process(byte[] packet)
				throws CorruptPacketException
		{
			synchronized(answerMonitor) {
				if(packet.length!=answerPacketLength) {
					if(answerPacketLength
							==ftPacketDataInputFilter.getPacketLength()) {
						throw new CorruptPacketException("packet.length="
												+packet.length
												+", expected "
												+answerPacketLength
												+". There is probably an error in the PacketInputStream (or in its delegates)!");
					} else {
						throw new CorruptPacketException("packet.length="
												+packet.length
												+", expected "
												+answerPacketLength
												+". Someone else is interfering here!");
					}
				}
				if(debug.debug) {
					debug.printInfo("processing answer-packet "
									+StringObservable.toHexString(packet));
				}
				
				// this will be the time of all generated events in this cycle
				long time = System.currentTimeMillis();
				if(lastTime != 0) {
					longAveragePreciseMillis =
							((longAveragePreciseMillis*(cycleCount-1))
								+(time-lastTime))
							/cycleCount;
					longAveragePollingMillis = (long)longAveragePreciseMillis+1;
					if(debug.debug) {
						debug.printInfo("long average pollingMillis are "
										+longAveragePollingMillis
										+", cycleCount is " +cycleCount
										+", last interval was "
										+(time-lastTime) +"ms");
					}			
				}
				lastTime = time;
				
				// Generate Event for each changed _output_
				// is it a good idea to generate this events first?
				// it depends when the actual digitals-reading and motor-setting
				// happes in the Hardware
				// reading sensors after setting the motors or before
				// or at the same time?
				// this is important to do before events get delivered!!
				synchronized(motorStateMonitor) {
					motorState = sendMotorState;
					if(motorSendPulseUpdated) {
						// onMillis
						System.arraycopy(sendOnMillis, 0,
										motorOnMillis, 0,
										sendOnMillis.length);
						// offMillis
						System.arraycopy(sendOffMillis, 0,
										motorOffMillis, 0,
										sendOffMillis.length);
						// pulseDir
						System.arraycopy(sendPulseDir, 0,
										motorPulseDir, 0,
										sendPulseDir.length);
						motorSendPulseUpdated = false;
						motorUpdateWasPulse = true;
					}		
					motorStateMonitor.notifyAll();
				}
				
				// manipulate motorValue according to pulse-lengths
				int nOutputs = extension ? 8 : 4;
				SensorEvent[] events = new SensorEvent[nOutputs];
				for(int i=0; i<nOutputs; i++) {
					double wasValue;
					double sendValue;
					
					wasValue = actuatorDefs[i].convertStateToValue(
												wasMotorState,
												wasOnMillis[i],
												wasOffMillis[i],
												wasPulseDir[i]);
					sendValue = actuatorDefs[i].convertStateToValue(
												sendMotorState,
												sendOnMillis[i],
												sendOffMillis[i],
												sendPulseDir[i]);
					if(sendValue != wasValue) {
						events[i] = new SensorEvent(actuatorDefs[i],
													null,
													time,
													wasValue,
													sendValue,
													true,
													null);
					}
				}
				// deliver Events
				for(int i=0; i<nOutputs; i++) {
					if(events[i] != null) {
						if(debug.debug) {
							debug.printInfo("Posting new Event" +events[i]
											+" for MotorInput " +(i+1));
						}
						eventQueue.post(events[i]);
					}
				}
				
				int skipLen = 0;
				
				int newDigitalsState = 0;
				byte newUndocumentedValue = undocumentedValue;
				if(doExtModule) {
					newDigitalsState = ByteArray.toShort(packet[skipLen+1],
															packet[skipLen]);
					//newDigitalsState = (packet[skipLen+1]<<8) | packet[skipLen];
					skipLen+=2;
					if(undocumented
							&& (lastQueryByte & mainModule) == mainModule) {
						newUndocumentedValue = packet[skipLen];
						skipLen++;
					} 
				} else {
					newDigitalsState = packet[skipLen];
					skipLen++;
				}
				
				int newAnalogEX = analog[EX];
				if(doAnalogEX) {
					newAnalogEX = ByteArray.toShort(packet[skipLen],
													packet[skipLen+1]);
					skipLen+=2;
				}
				int newAnalogEY = analog[EY];
				if(doAnalogEY) {
					newAnalogEY = ByteArray.toShort(packet[skipLen],
													packet[skipLen+1]);
					skipLen+=2;
				}
				
				// Generate Event for each changed input
				
				// digitals
				if(newDigitalsState != digitalsState) {
					int oldDigitalsState = digitalsState;
					int diff = newDigitalsState ^ digitalsState;

					// this is important to do before events get delivered!!
					digitalsState = newDigitalsState;
					int nInputs = doExtModule ? 16 : 8;
					for(int i=0; i<nInputs; i++) {
						int bit = 1<<i;
						if((diff & bit) > 0) {
							if(debug.debug) {
								debug.printInfo("Posting new Event for Input " +(i+1));
							}
							eventQueue.post(
									new SensorEvent(
											new DigitalValueDefinition(i+1),
							// include an instance of AllSensorEventsDescriptor
							// here ?
											null,
											time,
											(oldDigitalsState & bit)>>>i,
											(newDigitalsState & bit)>>>i,
											false,
											null));
						}
					}
				}
				
				// analogs
				if(newAnalogEX != analog[EX]) {
					AnalogValueDefinition avd = new AnalogValueDefinition(EX);
					int oldAnalog = analog[EX];
					analog[EX] = newAnalogEX;
					eventQueue.post(
							new SensorEvent(avd,
											null,
											time,
											oldAnalog,
											newAnalogEX,
											false,
											null));
				}
				if(newAnalogEY != analog[EY]) {
					AnalogValueDefinition avd = new AnalogValueDefinition(EY);
					int oldAnalog = analog[EY];
					analog[EY] = newAnalogEY;
					eventQueue.post(
							new SensorEvent(avd,
											null,
											time,
											oldAnalog,
											newAnalogEY,
											false,
											null));
				}
				
				// undocumented
				if(undocumented && newUndocumentedValue != undocumentedValue) {
					UndocumentedValueDefinition uvd =
							new UndocumentedValueDefinition();
					int oldUndoc = undocumentedValue;
					undocumentedValue = newUndocumentedValue;
					eventQueue.post(
						new SensorEvent(uvd,
										null,
										time,
										oldUndoc,
										newUndocumentedValue,
										false,
										null));
				}
								
				done = true;
				answerMonitor.notifyAll();
			}
			//Thread.currentThread().yield();
			return null;
		}
		public boolean done() {
			return done;
		}
	}	
	
		/*Every time a new change is added it looks whether there is any event to generate for it.
		If so, it also looks at the event-queue to see if there is any event which can be merged with the new one.
		Maybe it will generate the event and then ask every other event already in the queue if they will merge with it or not.
		The event-delivering thread has to deliver "the whole queue at once", or it could
		happen that an event willing to merge often will never be delivered when there is
		another event which comes equally often and won't merge. if the event-thread is slow takes only the first
		one in the queue to deliver it will always see the non-merger and never the merger.
		what if the queue gets filled more quickly than it gets emptied?
		*/
	/** This Task tries to reduce the number of changes in the queue.
		It also generates DeliveryTasks for any remaining Events.
		Since this Task is in the same Queue as the Tasks which really deliver the events and only gets added to the Queue again after it added a DeliveryTask for each Event it always processes the whole Queue and then waits until all the DeliveryTasks are finished.
		*/
	public class CondenseEventsTask implements de.jaetzold.util.Task {
		// TBD: collect statistics on the time it takes until this task can process again
		// TBD: implement some mechanism to prevent others from adding new events to eventQueue
		// if the come too fast. look from time to time whether the queue really gets emptied.
		// TBD: the condensing ... ;-)
		public Event perform() 
			throws InterruptedException	// in peekNext()
		{
			if(debug.debug) {
				debug.printInfo("starting ...");
			}
			eventQueue.peekNext();	// wait until at least one element is in the queue
			// the problem is, that this way any other Tasks in the notifierQueue
			// may be blocked. This is no problem as long as this task is the only one
			// which puts tasks in the queue
			while(eventQueue.hasMoreElements()) {
				final SensorEvent se = (SensorEvent)eventQueue.nextElement(); // does not block
				if(debug.debug) {
					debug.printInfo("scheduling " +se +" for delivery.");
				}
				Task deliver = new Task() {
						public Event perform() {
							SensorValueDefinition svd = (SensorValueDefinition)se.getSource();
							SensorPeer sp = (SensorPeer)sensorValueListeners.get(svd);
							if(sp != null) {
								if(debug.debug) {
									debug.printInfo("delivering " +se
												+" to " +sp);
								}
								sp.processEvent(se);
							} else {
								if(debug.debug) {
									debug.printInfo("discarding " +se
												+" because of missing peer for "
												+svd);
								}
							}
							return null;
						}
				};
				notifierQueue.post(deliver);
			}
			return null;
		}
	}
	
	interface SensorValueDefinition {
		public abstract boolean equals(Object o);
		public abstract int hashCode();
	}
	

	class ActuatorValueDefinition implements SensorValueDefinition {
		// TBD: conversion in the other direction also in here
		public double convertStateToValue(	int state,
											long onMillis,
											long offMillis,
											int pulseDir) {
			// both 0 means just use motorState, no pulsing by us
			if(onMillis + offMillis != 0) {
				int direction = (pulseDir<0 ? -1 : 1);
				double power = (double)onMillis / (double)(onMillis+offMillis);
				return direction*power;
			} else {
				return convertStateToValue(state);
			}
		}
		private double convertStateToValue(int state) {
			double result = 0;
			int shiftedState = (state>>>((inputID-1)*2)) & 3;
			// TBD: is that also correct if both are on?
			if((shiftedState & 1) > 0
					&& (shiftedState ^ 2) > 0) {
				result = -1;
			} else if((shiftedState ^ 1) > 0
					&& (shiftedState & 2) > 0) {
				result = 1;
			}
			return result;
		}
		
		int inputID;
		ActuatorValueDefinition(int inputID) {
			this.inputID = inputID;
		}
		public boolean equals(Object o) {
			boolean result = false;
			if(o!=null && this.getClass() == o.getClass()
					&& ((ActuatorValueDefinition)o).inputID == inputID) {
				result = true;
			}
			
			if(debug.debug) {
				debug.printInfo("tested for equality:"
							+this
							+(result?"":"!")
							+"=" +o);
			}
			return result;
		}
		public String toString() {
			return ""+this.getClass()+" for motorInput number " +inputID;
		}
		public int hashCode() {
			// the lowest 8 bits are for digital inputs
			return inputID<<8;
		}
	}
	
	class DigitalValueDefinition implements SensorValueDefinition {
		int inputID;
		DigitalValueDefinition(int inputID) {
			this.inputID = inputID;
		}
		public boolean equals(Object o) {
			boolean result = false;
			if(o!=null && this.getClass() == o.getClass()
					&& ((DigitalValueDefinition)o).inputID == inputID) {
				result = true;
			}
			
			if(debug.debug) {
				debug.printInfo("tested for equality:"
							+this
							+(result?"":"!")
							+"=" +o);
			}
			return result;
		}
		public String toString() {
			return ""+this.getClass()+" for input number " +inputID;
		}
		public int hashCode() {
			return inputID;
		}
	}
	
	class AnalogValueDefinition implements SensorValueDefinition {
		int inputID;
		AnalogValueDefinition(int inputID) {
			this.inputID = inputID;
		}
		
		public boolean isTriggered(int value) {
			return value != 0;
		}
		
		public boolean equals(Object o) {
			boolean result = false;
			if(o!=null && this.getClass() == o.getClass()
					&& ((AnalogValueDefinition)o).inputID == inputID) {
				result = true;
			}
			
			if(debug.debug) {
				debug.printInfo("tested for equality:"
							+this
							+(result?"":"!")
							+"=" +o);
			}
			return result;
		}
		public String toString() {
			return ""+this.getClass()
					+" for analog input " 
					+(inputID==EX?"EX":(inputID==EY?"EY":"unknown!"));
		}
		public int hashCode() {
			// the lowest 8 bits are for digital inputs
			// the next 16 bits are for outputs
			return inputID<<24;
		}
	}
	
	class UndocumentedValueDefinition implements SensorValueDefinition {
		public boolean isTriggered(int value) {
			return value != 0x02;
		}
		
		public boolean equals(Object o) {
			return this.getClass() == o.getClass();
		}
		
		public String toString() {
			return ""+this.getClass()
					+" for unknown protocol value.";
		}
		public int hashCode() {
			return 0;
		}
	}
	
	protected int motorState = 0;
	protected int shouldMotorState = 0;
	protected Object motorStateMonitor = new Object();
	protected int digitalsState;
	protected int[] analog = new int[2];
	protected byte undocumentedValue;
	
	// pulse-lengths
	protected long[] motorOffMillis = new long[8];
	protected long[] motorOnMillis = new long[8];
	protected long[] shouldMotorOffMillis = new long[8];
	protected long[] shouldMotorOnMillis = new long[8];
	// last edge
	protected long[] motorLastMillis = new long[8];
	// direction
	protected int[] motorPulseDir = new int[8];
	protected int[] shouldMotorPulseDir = new int[8];
	// should the local copies of the pulse-lengths in the polling-thread
	// be updated?
	protected boolean motorUpdatePulse;
	protected long longAveragePollingMillis = 10;	// just a guess
	
	public static final int EX = 0;
	public static final int EY = 1;
	
	protected int analogNumberingStart = 16;
	
	protected Hashtable sensorValueListeners = new Hashtable();
	
	// TBD: maybe public and then an Exception if detecting goes wrong?
	private Fischertechnik(String portName) 
			throws	NoSuchPortException,
					PortInUseException,
					UnsupportedCommOperationException,
					IOException
	{
		CommPortIdentifier cpi = CommPortIdentifier.getPortIdentifier(portName);
		SerialPort port = (SerialPort)cpi.open(getClass().getName(), 2000);

		try {
			port.setSerialPortParams(9600, port.DATABITS_8, port.STOPBITS_1, port.PARITY_NONE);
			InputStream in = null;
			OutputStream out = null;
			try {
				in = port.getInputStream();
			} catch(IOException ioe) {
				debug.printError(0, debug.className +": Port " +portName +" could not deliver InputStream.");
				if(debug.debug) {
					ioe.printStackTrace();
				}
				out = null;
				port.close();
				return;
			}
			try {
				out = port.getOutputStream();
			} catch(IOException ioe) {
				debug.printError(0, debug.className +": Port " +portName +" could not deliver OutputStream.");
				if(debug.debug) {
					ioe.printStackTrace();
				}
				in = null;
				port.close();
				return;
			}
			
			ftPacketDataInputFilter = new FTPacketDataInputFilter();
			inStream = new PacketInputStream(in,
											(byte[])null,
											ftPacketDataInputFilter,
											null);
			outStream = out;
			// flush inStream
			try {
				inStream.flush();
			} catch(Exception ioe) {
				debug.printError(0, debug.className +": Port " +portName +" could not be flushed.");
				if(debug.debug) {
					ioe.printStackTrace();
				}
				outStream = null;
				port.close();
				return;
			}
			
			// autodetect + init if successful
			int waitSum = 0;
			int waitCur = 8;
			
			debug.printInfo(2, debug.className +": Starting autodetect on port " +portName +" ...");

			try {
				outStream.write(new byte[]{(byte)0xC1, (byte)motorState});
			} catch(IOException ioe) {
				if(debug.debug) {
					debug.printError("Port " +portName +" didn't allow write.");
					ioe.printStackTrace();
				}
				outStream = null;
				port.close();
				return;
			}
/*			outStream.write(new byte[]{(byte)0xC1, motorState});
			outStream.write(new byte[]{(byte)0xC1, motorState});
			outStream.write(new byte[]{(byte)0xC1, motorState});
			outStream.write(new byte[]{(byte)0xC1, motorState});*/
			try {
				while(inStream.available() == 0 && waitSum<513) {
					if(debug.debug) {
						debug.printInfo("waiting for reply ...");
					}
					waitSum += waitCur;
					try {
						Thread.currentThread().sleep(waitCur);
					} catch(InterruptedException ie) {
					}
					waitCur *= 2;
				}
				try {
					Thread.currentThread().sleep(50);
				} catch(InterruptedException ie) {
				}
				if(inStream.available() != 1) {
					// no Fischertechnik-Interface could be detected
					if(debug.debug) {
						debug.printInfo("wrong number of bytes available: "
									+inStream.available());
					}
					outStream = null;
					//debug.printInfo(inStream.read());
					port.close();
					return;
				}
			} catch(IOException ioe) {
				if(debug.debug) {
					debug.printError("Port " +portName +" didn't allow available.");
					ioe.printStackTrace();
				}
				outStream = null;
				port.close();
				return;
			}
			
			try {
				digitalsState = (byte) inStream.read();
			} catch(IOException ioe) {
				if(debug.debug) {
					debug.printError("Port " +portName +" didn't allow read.");
					ioe.printStackTrace();
				}
				outStream = null;
				port.close();
				return;
			}
			
			// TBD: change after PacketInputStream has got an own toString()
			//debug.printInfo(10, debug.className +": Using " +inStream +", " +outStream);
			debug.printInfo(10, debug.className +": Using " +in +", " +out);
						
			notifier.setDaemon(true);
			reader.setDaemon(true);
			writer.setDaemon(true);
			
			notifier.start();
			reader.start();
			writer.start();
		
			debug.printInfo(0, "Successfully initialized Fischertechnik-Interface, found on serialport " +portName);
			interfaceDefinition = new
					RobotInterfaceStringDefinition("FT"+portName);
			connected = true;
		} catch(UnsupportedCommOperationException ucoe) {
			port.close();
			throw ucoe;
		}
	}
	
	private static Map portNamesToInterfaces = new Hashtable();
	private static final String[] STRING_IDS = new String[]{"ALL", "FT"};
	// maybe move this stuff into the adapter by using reflect or a Factory there for the constructor
	// TBD: null-reference should always be accepted
	/** Scans the given Ports.
		*/
	public static synchronized RobotInterface[] getInterfaces(
										RobotInterfaceDefinition[] ports)
	{
		String[] portNames = getSerialPortNames(ports, STRING_IDS);
		
		// try all names in portNames
		Vector interfaces = new Vector();
		for(int i=0; i<portNames.length; i++) {
			RobotInterface iface;
			iface = (RobotInterface)portNamesToInterfaces.get(portNames[i]);
			if(iface == null) {
				try {
					iface = new Fischertechnik(portNames[i]);
					// was initialization successful ?
					if(iface.isConnected()) {
						interfaces.add(iface);
						portNamesToInterfaces.put(portNames[i], iface);
					} else {
						if(debug.debug) {
							debug.printInfo(iface +" is not connected, not adding.");
						}
					}
				} catch(Exception e) {
					if(debug.debug) {
						debug.printError("Got Exception " +e
										+" while trying to open connection to port "
										+portNames[i]);
						e.printStackTrace();
					}
				}
			} else {
				interfaces.add(iface);
			}		
		}
		
		RobotInterface[] result = new RobotInterface[interfaces.size()];
		interfaces.copyInto(result);
		return result;
	}
	
	protected Hashtable portsToPeers = new Hashtable();
	// Factory-Methods for Actuator- and Sensor-Peers:
	// TBD: check that the port belongs to us?
	// not necessary since the only one who can call this is our own Port
	synchronized ActuatorPeer getActuatorPeer(	ActuatorPort port,
												Sensor sensor)
	{
		ActuatorPeer result = (ActuatorPeer)portsToPeers.get(port);
		if(result == null) {
			result = new FTActuatorPeer(port.getPortNumber());
			portsToPeers.put(port, result);
		}
		
		return result;
	}
	
	synchronized SensorPeer getSensorPeer(SensorPort port, Sensor sensor) {
		SensorPeer result = (SensorPeer)portsToPeers.get(port);
		if(result == null) {
			if(port.getPortNumber() >= analogNumberingStart) {
				if(sensor instanceof LightSensor) {
					result =
						new FTAnalogLightSensorPeer(port.getPortNumber()
													-analogNumberingStart);
				} else {
					result = new FTAnalogSensorPeer(port.getPortNumber()
													-analogNumberingStart);
				}
			} else {
				result = new FTSensorPeer(port.getPortNumber());
			}
			portsToPeers.put(port, result);
		}
		
		return result;
	}
	
	class FTActuatorPeer extends FTSensorPeer implements ActuatorPeer
	{
		private int left;
		private int right;
		private int portNumber;
		FTActuatorPeer(int portID) {
			super(portID);
			left = (1 << ((portID-1)*2));
			right = (1 << (((portID-1)*2)+1));
			portNumber = portID;
			valueDefinition = new ActuatorValueDefinition(portNumber);
		}
		
		public void propertyChange(PropertyChangeEvent evt) {
			if(debug.debug) {
				debug.printInfo("got event: " +evt);
			}
			
			debug.printError("Why is that used? Stacktrace is:"
							+debug.getCurrentStackTraceString());
			System.exit(1);
		}
		public void setValue(double newValue) {
			setValue(newValue, true);
		}
		// synchronized, so that we can wait until _our_ state has been set
		// and no one modifies it in that time
		public synchronized void setValue(double newValue, boolean waitToComplete) {
			double wantedValue = newValue;
			synchronized(motorStateMonitor) {
				// the first three "cases" need them to be set to 0
				shouldMotorOnMillis[portNumber-1] = 0;
				shouldMotorOffMillis[portNumber-1] = 0;
				
				if(newValue == 0) {
					shouldMotorState &= ~(left|right);
				} else if(newValue <= -1) {
					shouldMotorState &= ~(left|right);
					shouldMotorState |= left;
					wantedValue = -1;
				} else if(newValue >= 1) {
					shouldMotorState &= ~(left|right);
					shouldMotorState |= right;
					wantedValue = 1;
				} else {
					// implement 8 Power-Levels 0.125, 0.25, ... , 0.875, 1.0
					// in both directions
					// TBD: if both pulseLength are significantly greater than longAveragePollingMillis, divide them to get a better granularity of the pulsing
					long pulseLengthBase = longAveragePollingMillis*8;
					shouldMotorPulseDir[portNumber-1] = (newValue<0 ? -1 : 1);
					double power = Math.abs(newValue);
					shouldMotorOnMillis[portNumber-1] =
										Math.round(pulseLengthBase*power);
					shouldMotorOffMillis[portNumber-1] =
										Math.round(pulseLengthBase*(1-power));
					wantedValue = ((ActuatorValueDefinition)valueDefinition)
						.convertStateToValue(shouldMotorState,
											shouldMotorOnMillis[portNumber-1],
											shouldMotorOffMillis[portNumber-1],
											shouldMotorPulseDir[portNumber-1]);

					if(debug.debug) {
						debug.printInfo("base is " +pulseLengthBase
										+" and newValue is " +newValue
										+", setting onMillis="
										+shouldMotorOnMillis[portNumber-1]
										+", offMillis="
										+shouldMotorOffMillis[portNumber-1]);
					}
				}
				motorUpdatePulse = true;
			}
			if(waitToComplete) {
				synchronized(motorStateMonitor) {
					// wait until our desired state is reached
					try {
						double value;
						while(Math.abs((value = getValue()) - wantedValue)
								> getPrecision()) {
							if(debug.debug) {
								debug.printInfo("value is " +value
												+" and wanted is " +wantedValue
												+", precision="
												+getPrecision());
							}
							motorStateMonitor.wait();
						}
					} catch(InterruptedException ie) {
					}
				}
			}
		}
		public double getValue() {
			double result;
			synchronized(motorStateMonitor) {
				result = ((ActuatorValueDefinition)valueDefinition)
						.convertStateToValue(motorState,
											motorOnMillis[portNumber-1],
											motorOffMillis[portNumber-1],
											motorPulseDir[portNumber-1]);
			}
			if(debug.debug) {
				debug.printInfo("returning value " +result
								+" for motorState " +motorState
								+" of motor number " +portNumber);
			}
			return result;
		}

		public boolean getDeliversFloatingPoint() {
			return true;
		}
		
		public double getPrecision() {
			return 1.0/16.0;
		}
		
		public void vetoableChange(PropertyChangeEvent evt)
				throws PropertyVetoException
		{
		}
	}
	
	// This is really just a fake to test the feature because I don't know what "lux" the values mean.
	// TBD: It might be better to provide an FTLightSensorScalingConversion
	// which users of Fischertechnik can set to their LightSensor
	class FTAnalogLightSensorPeer	extends		FTAnalogSensorPeer
									implements	LightSensorPeer
	{
		private final Debug debug =
			new Debug("de.jaetzold.art.Fischertechnik.FTAnalogLightSensorPeer");
		{ if(debug.debug) debug.objectIdentifier =
												debug.getIdentifierFor(this); }
		FTAnalogLightSensorPeer(int inputID) {
			super(inputID);
		}
		public double getValue() {
			return convertAnalogToLightValue(super.getValue());
		}
		
		public double getMeasurableMin() {
			return 1;	// a value of 0 is'nt really "measured",
						// it's more out of the range the sensor can measure
		}
		public double getMeasurableMax() {
			return 1023;// even very bright light will never produce a value higher than
		}
		
		public double convertAnalogToLightValue(double value) {
			// 1024 is upper border
			double result = value;
			if(value > 0) {
				result = Math.max(1, 1025-value);
				if(debug.debug) {
					if(result!=(1025-value)) {
						debug.printError("interface delivered analog greater 1024, value was " +value);
					}
				}
			}
			if(debug.debug) {
				debug.printInfo("converted " +value +" to " +result);
			}
			return result;
		}
		
		public void processEvent(SensorEvent event) {
			// manipulate value in event if it comes from our
			// valueDefinition
			SensorEvent newEvent = event;
			if(event.getSource().equals(valueDefinition)) {
				newEvent = new SensorEvent(event.getSource(),
											event.getEventDescriptor(),
											event.getTimestamp(),
								convertAnalogToLightValue(event.getOldValue()),
								convertAnalogToLightValue(event.getNewValue()),
											event.isFloatingPoint(),
											event.getPropagationId());
			}
			super.processEvent(newEvent);
		}
	}
	
	class FTAnalogSensorPeer extends FTSensorPeer {
		private int inputNumber;
		FTAnalogSensorPeer(int inputID) {
			super(inputID);
			if(inputID >= analog.length) {
				throw new IllegalArgumentException("No analog input with number " +inputID +" available.");
			}
			inputNumber = inputID;
			valueDefinition =
					new AnalogValueDefinition(inputID);
		}
		public double getValue() {
			return analog[inputNumber];
		}
		
		public double getPrecision() {
			return 31;	// TBD: is this somewhat conservative?
		}
	}
	
	class FTSensorPeer extends BaseSensorPeer {
		private int inputID;
		FTSensorPeer(int inputID) {
			this.inputID = (1<<(inputID-1));
			valueDefinition =
					new DigitalValueDefinition(inputID);
		}
		public double getValue() {
			return ((digitalsState & inputID) > 0) ? 1 : 0;
		}
		
		public boolean getDeliversFloatingPoint() {
			return false;
		}
		
		SensorValueDefinition valueDefinition;
		SensorEventDescriptor eventDescriptor;
		
		public void enableEvent(SensorEventDescriptor eventDescriptor) {
			if(this.eventDescriptor == null && eventDescriptor != null) {
				if(debug.debug) {
					debug.printInfo("enabling " +eventDescriptor +" for " +this +" (" +valueDefinition +")");
				}
				this.eventDescriptor = eventDescriptor;
				sensorValueListeners.put(valueDefinition, this);
			}
		}
		public void disableEvent(SensorEventDescriptor eventDescriptor) {
			if(eventDescriptor != null) {
				if(debug.debug) {
					debug.printInfo("disabling " +eventDescriptor +" for " +this);
				}
				this.eventDescriptor = null;
			}
		}
		
		public boolean isConnected() {
			return Fischertechnik.this.isConnected();
		}
	}
	/** Returns a description of all ActuatorPorts this Interface provides. The
	description consists of the portnumber and a minimal pulsewidth this
	actuatorport can generate. (also: voltage, polarity, AC/DC, maximum number
	of automated pulses, granularity of voltage, excludes use of other
	actuators?, asf.)
	*/
	public ActuatorPort[] getActuatorPorts() {
		//TBD: think about whether there should only be one Port/Peer Object for every Port
		boolean extension = Boolean.getBoolean(debug.className +".extensionModule");
		ActuatorPort[] result;
		if(extension) {
			result = new ActuatorPort[8];
		} else {
			result = new ActuatorPort[4];
		}
		for(int i=0; i<result.length; i++) {
			result[i] = new FTActuatorPort(	this,
											"M" +(i+1), i+1, "DC",
											-1, 1, 17,
											new Port[0]);
		}
		
		return result;
	}
	
	public SensorPort[] getSensorPorts() {
		boolean extension = Boolean.getBoolean(debug.className +".extensionModule");
		SensorPort[] result;
		int digitalsSize = 0;
		if(extension) {
			digitalsSize=16;
		} else {
			digitalsSize=8;
		}
		result = new SensorPort[digitalsSize+analog.length];
		
		for(int i=0; i<digitalsSize; i++) {
			result[i] = new FTSensorPort(	this,
											"E" +(i+1), i+1,
											0, 1, 2,
											new Port[0]);
		}
		for(int i=0; i<analog.length; i++) {
			String id = (i == EX ? "X"
									: (i==EY ? "Y"
												: "?"));
			result[digitalsSize+i] =
						new FTSensorPort(	this,
											"E" +id,
											analogNumberingStart+i,
											0, 1024, 1025,
											new Port[0]);
			if(debug.debug) {
				debug.printInfo(" created analog Port " +result[digitalsSize+i]
								+" with number " +(analogNumberingStart+i));
			}
		}
		
		return result;
	}
	
	class FTSensorPort extends BaseSensorPort {
		FTSensorPort(RobotInterface iface,
					String portName,
					int portNumber,
					int minimumValue,
					int maximumValue,
					int granularity,
					Port[] excludedPorts) {
			super(iface, portName, portNumber, minimumValue, maximumValue, granularity, excludedPorts);
		}
		public SensorPeer getPeerFor(Sensor s) {
			SensorPeer peer = getSensorPeer(this, s);
			return peer;
		}
	}
	class FTActuatorPort extends BaseActuatorPort {
		FTActuatorPort(RobotInterface iface,
						String portName,
						int portNumber,
						String type,
						int minimumValue,
						int maximumValue,
						int granularity,
						Port[] excludedPorts) {
			super(iface, portName, portNumber, type, minimumValue, maximumValue, granularity, excludedPorts);
		}
		public SensorPeer getPeerFor(Sensor s) {
			ActuatorPeer peer = getActuatorPeer(this, s);
			return peer;
		}
		public ActuatorPeer getPeerFor(Actuator a) {
			ActuatorPeer peer = getActuatorPeer(this, a);
			return peer;
		}
	}
	/** Returns a description of all SensorPorts this Interface provides. The
	description consists of the portnumber and some other information describing
	the capabilities of the sensorport. (pollingfreq., settable pollingfreq.?,
	hw-events?, scaling (which voltage/resistance for which value),
	maximum-value/voltage?, Powered?, counting?, excludes use of other
	sensors?, asf.).
	*/ 
	private boolean alive;
	public boolean isAlive() {
		return alive;
	}
	
	private boolean connected;
	public boolean isConnected() {
		return connected;
	}
	
	private RobotInterfaceStringDefinition interfaceDefinition;
	public RobotInterfaceDefinition getInterfaceDefinition() {
		return interfaceDefinition;
	}
	// TBD: rewrite this to use STRING_IDS and maybe
	// getSerialPortNames(RobotInterfaceDefinition[],String[])
	public boolean conformsTo(RobotInterfaceDefinition definition) {
		if(interfaceDefinition != null
				&& definition instanceof RobotInterfaceStringDefinition) {
			String otherId = ((RobotInterfaceStringDefinition)definition)
									.getInterfaceString();
			if(otherId.startsWith("FT")) {
				if(otherId.length() <= 2) {
					// FT is our main id
/* exit 1 */		return true;
				}
				String otherPort = otherId.substring(2);
				try {
					int portNumber = Integer.parseInt(otherPort);
					// a portnumber had been specified
					// test all known serial port names
					String[] names = getSerialPortNames(portNumber);
					String myPort =
							interfaceDefinition
									.getInterfaceString().substring(2);
					for(int i=0; i<names.length; i++) {
						if(myPort.equals(names[i])) {
/* exit 2 */				return true;
						}
					}
				} catch(NumberFormatException nfe) {
					// a port name had been specified
/* exit 3 */		return otherId.equals(interfaceDefinition
												.getInterfaceString());
				}				
			}
		}
/* exit 4 */	
		return false;
			
	}
}
