package de.jaetzold.art.platform.lego;

import de.jaetzold.art.RobotInterface;
import de.jaetzold.art.RobotInterfaceDefinition;
import de.jaetzold.art.RobotInterfaceStringDefinition;
import de.jaetzold.art.Sensor;
import de.jaetzold.art.SensorEvent;
import de.jaetzold.art.SensorEventDescriptor;
import de.jaetzold.art.BooleanSensor;
import de.jaetzold.art.StateSensor;
import de.jaetzold.art.CountSensor;
import de.jaetzold.art.LightSensor;
import de.jaetzold.art.ActuatorPort;
import de.jaetzold.art.SensorPort;
import de.jaetzold.art.Port;
import de.jaetzold.art.platform.ActuatorPeer;
import de.jaetzold.art.platform.SensorPeer;
import de.jaetzold.art.platform.LightSensorPeer;
import de.jaetzold.art.platform.CountSensorPeer;
import de.jaetzold.art.platform.StateSensorPeer;
import de.jaetzold.art.platform.BaseSensorPeer;
import de.jaetzold.art.platform.PacketChecker;

import de.jaetzold.util.Debug;
import de.jaetzold.util.Task;
import de.jaetzold.util.Event;
import de.jaetzold.util.EqualSourceEvent;
import de.jaetzold.util.Range;
import de.jaetzold.util.StringObservable;
import de.jaetzold.util.StateDecider;

import javax.comm.CommPortIdentifier;
import javax.comm.SerialPort;
import javax.comm.NoSuchPortException;
import javax.comm.PortInUseException;
import javax.comm.UnsupportedCommOperationException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.Map;
import java.util.Hashtable;
import java.util.Vector;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyVetoException;

// TBD: Events are sometimes generated faster than they are delivered ...
// TBD: Think about a way to add writePacketTasks to the _beginning_ of the writerQueue so they don't have to wait for all the polling tasks

public class Mindstorms extends RCX {
	private static Debug debug = new Debug("de.jaetzold.art.Mindstorms");

	public static final int CONNECTION_ALIVE = 1;
	public static final int CONNECTION_DEAD = 0;
	private int connectionState = CONNECTION_DEAD;
	public int getConnectionState() {
		return connectionState;
	}
	
	public boolean isConnected() {
		boolean result = super.isConnected();
		if(result) {
			result = testForTower();
		}
		return result;
	}
	
	protected boolean testForTower() {
		if(serialPort != null) {
			boolean rts = true;
			for(int i=0; i<10; i++) {
				rts = !rts;
				serialPort.setRTS(rts);
				if(serialPort.isCTS() != rts) {
					return false;
				}
				// wait for a random amount of time
				try {
					Thread.currentThread().sleep((int)(Math.random()*10));
				} catch(InterruptedException ie) {
				}
				// test again
				if(serialPort.isCTS() != rts) {
					return false;
				}
			}
		} else {
			return false;
		}
		return true;
	}
	
	private SerialPort serialPort;
	
	private Mindstorms(String portName) 
			throws	NoSuchPortException,
					PortInUseException,
					UnsupportedCommOperationException,
					IOException
	{
		CommPortIdentifier cpi = CommPortIdentifier.getPortIdentifier(portName);
		SerialPort port = (SerialPort)cpi.open(getClass().getName(), 2000);

		try {
			port.setSerialPortParams(2400, port.DATABITS_8, port.STOPBITS_1, port.PARITY_ODD);

			// Check for Tower first
			serialPort = port;
			if(!testForTower()) {
				if(debug.debug) {
					debug.printInfo("No tower found on port " +portName);
				}
				port.close();
				return;
			}
			debug.printInfo(2, "Mindstorms tower found on port " +portName);

			InputStream in = port.getInputStream();
			OutputStream out = port.getOutputStream();
			// All packets from the MS should start with 55 ff 00
			// but some do with ee ff 00, therefore look only for ff 00?
			inStream = new RCXPacketInputStream(in, new byte[]{(byte)0x55, (byte)0xff, (byte)0x00});
			// The header of outgoing MS-Packets is: 55 ff 00
			outStream = new RCXPacketOutputStream(out, new byte[]{	(byte) 0x55,
																	(byte) 0xff,
																	(byte) 0x00});

			inStream.setPacketChecker(new PacketChecker()
					{
						private Hashtable lengthTable = new Hashtable();
						{
							Number one = new Byte((byte)1);
							Number two = new Byte((byte)2);
							Number three = new Byte((byte)3);
							Number four = new Byte((byte)4);
							
/***** BEGIN Packet-list *****/

// trailing "r" stands for REQUEST/COMMAND
// trailing "a" stands for ANSWER/REPLY
// trailing "m" stands for MESSAGE (ANSWER without REQUEST)
// the values put in here are with bit 8 cleared

lengthTable.put(new Byte((byte)0xe1), two);		// set motor direction r
lengthTable.put(new Byte((byte)0x16), one);		// set motor direction a

lengthTable.put(new Byte((byte)0xd1), two);		// clear sensor value r
lengthTable.put(new Byte((byte)0x26), one);		// clear sensor value a

lengthTable.put(new Byte((byte)0xb1), two);		// set power down delay r
lengthTable.put(new Byte((byte)0x46), one);		// set power down delay a

lengthTable.put(new Byte((byte)0x70), one);		// delete all subroutines r
lengthTable.put(new Byte((byte)0x87), one);		// delete all subroutines a

lengthTable.put(new Byte((byte)0x50), one);		// stop all tasks r
lengthTable.put(new Byte((byte)0xa7), one);		// stop all tasks a

lengthTable.put(new Byte((byte)0x42), three);	// set sensor mode r
lengthTable.put(new Byte((byte)0xb5), one);		// set sensor mode a

lengthTable.put(new Byte((byte)0x32), three);	// set sensor type r
lengthTable.put(new Byte((byte)0xc5), one);		// set sensor type a

lengthTable.put(new Byte((byte)0x40), one);		// delete all tasks r
lengthTable.put(new Byte((byte)0xb7), one);		// delete all tasks a

lengthTable.put(new Byte((byte)0x30), one);		// get battery power r
lengthTable.put(new Byte((byte)0xc7), three);	// get battery power a

lengthTable.put(new Byte((byte)0x21), two);		// set motor on/off r
lengthTable.put(new Byte((byte)0xd6), one);		// set motor on/off a

lengthTable.put(new Byte((byte)0x14), four);	// set variable r
lengthTable.put(new Byte((byte)0xe3), one);		// set variable a

lengthTable.put(new Byte((byte)0x13), three);	// set motor power r
lengthTable.put(new Byte((byte)0xe4), one);		// set motor power a

lengthTable.put(new Byte((byte)0x12), three);	// get value (poll) r
lengthTable.put(new Byte((byte)0xe5), three);	// get value (poll) a

lengthTable.put(new Byte((byte)0x10), one);		// alive r
lengthTable.put(new Byte((byte)0xe7), one);		// alive a

lengthTable.put(new Byte((byte)0xa5), new Byte((byte)6));	// unlock firmware r
lengthTable.put(new Byte((byte)0x52), new Byte((byte)26));	// unlock firmware a

/***** END Packet-list *****/
						
						}
						public boolean isComplete(byte[] packet, int len) {
							if(len <= 0) {		// no empty packets
								return false;
							}
							if(packet==null) {	// no empty packets
								if(debug.debug) {
									debug.printInfo("Caution! been asked to check completeness of null-packet (len was " +len +").");
								}
								return false;
							}
							// they are only checked for length, which only
							// depends on the first byte
							// so the length can be retrieved from a table ...
							Number length = (Number)lengthTable.get(
										new Byte((byte)(packet[0] & ~ 0x08)));

							if(debug.debug && length == null) {
								debug.printInfo("Caution! been asked to check completeness of packet for which i don't know the length (id was "
										+StringObservable.toHexString(packet,
																	0,
																	1,
																	" ")
													+").");
							}

							// Only if we know the length of the packet and it isn't len
							// return false
							// add length of checksum to length of packet
							if(length != null
									&& length.intValue()+1 != len) {
								return false;
							}
							return true;
						}
					});

			//debug.printInfo(10, debug.className +": Using " +inStream +", " +outStream);
			debug.printInfo(10, debug.className +": Using " +in +", " +out);

			inStream.flush();

			notifier.start();
			reader.start();
			writer.start();

			if(debug.debug) {
				System.out.println(getClass().getName() +": "
									+"checking alive ...");
			}
			if(alive()) {
				connectionState = CONNECTION_ALIVE;
			} else {
				in.close();
				out.close();
				port.close();
				writer.endPerforming();
				reader.endPerforming();
				notifier.endPerforming();
				return;
			}
			if(debug.debug) {
				System.out.println(getClass().getName() +": "
									+"Mindstorms alive.");
			}
			debug.printInfo(0, "Successfully initialized LEGO-Mindstorms Interface, found on serialport " +portName);
			interfaceDefinition = new
					RobotInterfaceStringDefinition("MS"+portName);
			connected = true;
			//unlockFirmware();

			// should check result of that call
		} catch(UnsupportedCommOperationException ucoe) {
			port.close();
			throw ucoe;
		}
	}

	static {
		byte[] tmp = "Do you byte, when I knock?".getBytes();
		UNLOCK_FIRMWARE = new byte[tmp.length+1];
		System.arraycopy(tmp, 0, UNLOCK_FIRMWARE, 1, tmp.length);
		UNLOCK_FIRMWARE[0] = (byte)0xa5;
	}
	
	private static Map portNamesToInterfaces = new Hashtable();
	private static final String[] STRING_IDS = new String[]{"ALL","RCX","MS"};
	// maybe move this stuff into the adapter by using reflect there for the constructor
	/** Scans the given Ports.
		*/
	public static synchronized RobotInterface[] getInterfaces(
										RobotInterfaceDefinition[] ports)
	{
		String[] portNames = getSerialPortNames(ports, STRING_IDS);
		
		// try all names in portNames
		Vector interfaces = new Vector();
		for(int i=0; i<portNames.length; i++) {
			RobotInterface iface;
			iface = (RobotInterface)portNamesToInterfaces.get(portNames[i]);
			if(iface == null) {
				try {
					iface = new Mindstorms(portNames[i]);
					// was initialization successful ?
					if(iface.isConnected()) {
						interfaces.add(iface);
						portNamesToInterfaces.put(portNames[i], iface);
					} else {
						if(debug.debug) {
							debug.printInfo(iface +" is not connected, not adding.");
						}
					}
				} catch(Exception e) {
					if(debug.debug) {
						debug.printError("Got Exception " +e
										+" while trying to open connection to port "
										+portNames[i]);
						e.printStackTrace();
					}
				}
			} else {
				interfaces.add(iface);
			}		
		}
		
		RobotInterface[] result = new RobotInterface[interfaces.size()];
		interfaces.copyInto(result);
		return result;
	}
	
	/** every Sensor gets its own Peer */
	protected Map portToPeerCollection = new Hashtable();
	// Factory-Methods for Actuator- and Sensor-Peers:
	// TBD: Better check whether this Port really belongs to me?
	// not necessary since the only one who can call this is our own Port
	protected synchronized ActuatorPeer getActuatorPeer(ActuatorPort port,
														Sensor sensor)
	{
		Collection peers = (Collection)portToPeerCollection.get(port);
		if(peers == null) {
			peers = new Vector();
			portToPeerCollection.put(port, peers);
		}
		ActuatorPeer result;
		
		result = new RCXActuatorPeer((byte)port.getPortNumber());
		
		// don' know why still to remember them this way ...
		// peers.add(result);
		if(debug.debug) {
			debug.printInfo("created new peer " +result
							+" for " +sensor +" on port " +port);
		}
		return result;
	}
	
	protected synchronized SensorPeer getSensorPeer(SensorPort port,
													Sensor sensor)
	{
		Collection peers = (Collection)portToPeerCollection.get(port);
		if(peers == null) {
			peers = new Vector();
			portToPeerCollection.put(port, peers);
		}
		
		SensorPeer result;
		// by always creating a peer, a ports sensor is always configured
		if(sensor instanceof CountSensor) {
			result = new MSCountSensorPeer((byte)port.getPortNumber());
		} else if(sensor instanceof BooleanSensor) {
			result = new RCXBooleanSensorPeer((byte)port.getPortNumber());
		} else if(sensor instanceof StateSensor) {
			result = new RCXStateSensorPeer((byte)port.getPortNumber());
		} else if(sensor instanceof LightSensor) {
			result = new MSLightSensorPeer((byte)port.getPortNumber());
		} else {
			result = new RCXRawSensorPeer((byte)port.getPortNumber());
		}
		
		// don' know why still to remember them this way ...
		// peers.add(result);
		if(debug.debug) {
			debug.printInfo("created new peer " +result
							+" for " +sensor +" on port " +port);
		}
		return result;
	}
	/** Returns a description of all ActuatorPorts this Interface provides. The
	description consists of the portnumber and a minimal pulsewidth this
	actuatorport can generate. (also: voltage, polarity, AC/DC, maximum number
	of automated pulses, granularity of voltage, excludes use of other
	actuators?, asf.)
	*/
	public ActuatorPort[] getActuatorPorts() {
		return new ActuatorPort[]{
			new RCXActuatorPort(this, "A", 0, "DC", -1, 1, 17, new Port[0]),
			new RCXActuatorPort(this, "B", 1, "DC", -1, 1, 17, new Port[0]),
			new RCXActuatorPort(this, "C", 2, "DC", -1, 1, 17, new Port[0])};
	}
	
	/** Returns a description of all SensorPorts this Interface provides. The
	description consists of the portnumber and some other information describing
	the capabilities of the sensorport. (pollingfreq., settable pollingfreq.?,
	hw-events?, scaling (which voltage/resistance for which value),
	maximum-value/voltage?, Powered?, counting?, excludes use of other
	sensors?, asf.).
	*/ 
	public SensorPort[] getSensorPorts() {
		return new SensorPort[]{
				new RCXSensorPort(this, "1", 0, 0, 1023, 1024, new Port[0]),
				new RCXSensorPort(this, "2", 1, 0, 1023, 1024, new Port[0]),
				new RCXSensorPort(this, "3", 2, 0, 1023, 1024, new Port[0])};
	}
		
	protected class LightSensorValueDefinition
					extends PollSensorValueDefinition
	{
		public LightSensorValueDefinition(byte argument) {
			super((byte)0x09, argument);	// sets sensorNumber in superclass
		}
		public byte[][] getConfigurePackets() {
			// set sensor type: 0x32 <sensor> <type>
			// configure to standard light-type
			// set sensor mode: 0x42 <sensor> <mode+slope>
			// configure to raw mode
			return new byte[][]	{	{ (byte)0x32, sensorNumber, (byte)0x03},
									{ (byte)0x42, sensorNumber, (byte)(0x00<<5)}
								};	// order is important
			// intead of configuring to raw mode, the raw value (source 0x12)
			// could be polled. But the display would show the percentage
			// value then and that might be too confusing
			
			// nevertheless, the raw value is inverted, to resemble
			// more light with greater values
		}
		public Object parseResult(byte[] packet) {
			Number result = (Number)super.parseResult(packet);
			if(result != null) {
				return new Integer(1023-result.intValue());
			} else {
				return null;
			}
		}
	}
	protected class MSLightSensorPeer	extends		RCXRawSensorPeer
										implements	LightSensorPeer
	{
		protected MSLightSensorPeer(byte inputID,
									SensorValueDefinition valueDefinition) {
			super(inputID, valueDefinition);
		}
		protected MSLightSensorPeer(byte inputID) {
			this(inputID, new LightSensorValueDefinition(inputID));
		}
		// are these good values? 
		public double getMeasurableMin() {
			return 320; // (tested with flashlight)
		}
		public double getMeasurableMax() {
			return 830; // (tested with black, non-reflecting fabric)
		}
		public double getPrecision() {
			return 7;	// that's around the usual deviation
		}
	}
	
	protected class RotationSensorValueDefinition
					extends PollSensorValueDefinition
	{
		// needed to test for equality because of localCount
		protected Object peerID;
		public RotationSensorValueDefinition(byte argument, Object peerID) {
			super((byte)0x09, argument);	// sets sensorNumber in superclass
			this.peerID = peerID;
		}
		public byte[][] getConfigurePackets() {
			// set sensor type: 0x32 <sensor> <type>
			// configure to standard rotation-sensor type (and mode)
			return new byte[][] { { (byte)0x32, sensorNumber, (byte)0x04} };
		}
		protected double localCount;
		public Object parseResult(byte[] packet) {
			Number result = (Number)super.parseResult(packet);
			if(result != null) {
				return new Double(result.doubleValue() +localCount);
			} else {
				return null;
			}
		}
		void setCount(double count) {
			localCount = count;
		}
		public boolean equals(Object o) {
			if(super.equals(o) &&  o != null
					&& o instanceof RotationSensorValueDefinition
					&& peerID.equals(((RotationSensorValueDefinition)o).peerID))
			{
				return true;
			}
			return false;
		}
	}
	protected class MSCountSensorPeer	extends		RCXStateSensorPeer
										implements	CountSensorPeer
	{
		RotationSensorValueDefinition rotationDefinition;
		protected MSCountSensorPeer(byte inputID,
								RotationSensorValueDefinition valueDefinition) {
			super(inputID, valueDefinition);
			rotationDefinition = valueDefinition;
		}
		protected MSCountSensorPeer(byte inputID) {
			this(inputID, new RotationSensorValueDefinition(inputID,
															new Object()));
		}
		
		public synchronized void setCount(double count) {
			if(debug.debug) {
				debug.printInfo("clearing rotation-sensor to " +count);
			}
			// packet for clearing the sensor value (setting it to 0);
			try {
				call(new byte[] { (byte)0xd1, inputID});
			} catch(Exception e) {
				if(debug.debug) {
					debug.printInfo("Exception while clearing sensor value:"
									+debug.stackTraceToString(e));
				}
			}
			rotationDefinition.setCount(count);
		}
		public boolean getDeliversFloatingPoint() {
			return true; // (because setCount accepts double)
		}
	}
	
	// But a BooleanSensorPeer configures the Sensor to be a Switch
	protected class RCXBooleanSensorPeer	extends		RCXStateSensorPeer
											implements	StateSensorPeer
	{
		protected RCXBooleanSensorPeer(byte inputID,
									SensorValueDefinition valueDefinition) {
			super(inputID, valueDefinition);
		}
		protected RCXBooleanSensorPeer(byte inputID) {
			this(inputID, new BooleanConversionValueDefinition(inputID));
		}
	}

	// for the standard-RCX boolean conversion a StateSensorPeer is used
	// because the RCX really doesn't know what's "true" and "false"
	// just 0 and 1
	protected class RCXStateSensorPeer	extends		RCXRawSensorPeer
										implements	StateSensorPeer
	{
		protected RCXStateSensorPeer(byte inputID,
									SensorValueDefinition valueDefinition) {
			super(inputID, valueDefinition);
		}
		protected RCXStateSensorPeer(byte inputID) {
			this(inputID, new StateConversionValueDefinition(inputID));
		}
	}

	protected class RCXRawSensorPeer extends BaseSensorPeer {
		protected byte inputID;
		protected RCXRawSensorPeer(byte inputID,
									SensorValueDefinition valueDefinition) {
			this.inputID = inputID;
			this.valueDefinition = valueDefinition;
			
			// standard-sensor needs configuration
			byte[][] packets = valueDefinition.getConfigurePackets();
			if(packets != null) {
				for(int i=0; i<packets.length; i++) {
					if(packets[i] != null) {
						try {
							call(packets[i]);	// send configure-packet
						} catch(Exception e) {
							if(debug.debug) {
								debug.printInfo("Exception while"
												+"configuring sensor:"
												+debug.stackTraceToString(e));
							}
						}
					}
				}
			}
			
			// addSensorValueListener(valueDefinition, this);
		}
		public RCXRawSensorPeer(byte inputID) {
			this(inputID, new RawValueDefinition(inputID));
		}
		public boolean getDeliversFloatingPoint() {
			return false;
		}
		public double getValue() {
			Number result = (Number)getSensorValue(valueDefinition);
			if(result == null) {
				if(debug.debug) {
					debug.printError("getSensorValue returned " +result
									+"delivering 0, isAlive=" +isAlive());
				}
				// could there be done anything better?
				// maybe a NotConnectedException?
				return 0;
			}
			// MS only has int as values
			return result.intValue();
		}
		
		SensorValueDefinition valueDefinition;
		SensorEventDescriptor eventDescriptor;
		
		public synchronized void enableEvent(
									SensorEventDescriptor eventDescriptor)
		{
			if(this.eventDescriptor == null && eventDescriptor != null) {
				if(debug.debug) {
					debug.printInfo("enabling " +eventDescriptor +" for " +this);
				}
				this.eventDescriptor = eventDescriptor;
				startPolling(valueDefinition, this);
			}
		}
		public synchronized void disableEvent(
									SensorEventDescriptor eventDescriptor)
		{
			if(eventDescriptor != null) {
				if(debug.debug) {
					debug.printInfo("disabling " +eventDescriptor +" for " +this);
				}
				this.eventDescriptor = null;
				stopPolling(valueDefinition, this);
			}
		}
		
		public boolean isConnected() {
			return Mindstorms.this.isConnected();
		}
	}
	
	protected class RCXActuatorPeer	extends		RCXRawSensorPeer
									implements	ActuatorPeer
	{
		// values are polled if necessary, that's more consistent (but slower!)
		//boolean forward = true;
		boolean shouldForward = true;
		//boolean off = true;
		boolean shouldOff = true;
		//int power = 1;
		int shouldPower = 0;
		boolean gotValueSet = false;
		byte outputBit;
		
		protected RCXActuatorPeer(byte outputID,
								SensorValueDefinition valueDefinition)
		{
			super(outputID, valueDefinition);
			
			this.outputBit = (byte)(1<<outputID);
			getValue();	// set proper state of should-values
		}
		
		protected RCXActuatorPeer(byte outputID) {
			this(outputID, new MotorSensorValueDefinition(outputID));
		}
		
		public void setValue(double newValue) {
			setValue(newValue, true);
		}
		public void setValue(final double newValue,
							final boolean waitToComplete) {
			Task t = null;
			// collect parameters for packets
			
			// all these synchronized are to ensure, that
			// the possible three writePacketTasks don't mix with others
			// for this Actuator. Sets the MotorState more consistent
			// also the should-Values don't get mixed up that way
			synchronized(this) {
				// motor direction
				if((shouldForward && newValue<0)
						|| (!shouldForward && newValue>0)
						|| shouldOff && newValue != 0) {// we don't know the
												// direction if state was off
					if(debug.debug) {
						debug.printInfo("changing motor-direction from "
										+(shouldForward?"forward":"backward")
										+" to "
										+(newValue>0?"forward":"backward"));
					}

					byte state = 0;
					shouldForward = newValue>0;
					state = shouldForward ? (byte)(outputBit|0x80) : outputBit;
					t = new WritePacketTask(
							new byte[]{	(byte)0xe1,
										state});
					writerQueue.post(t);
				}
				// motor power
				final int absValue = (int)Math.round(Math.max(1,
														Math.abs(newValue*8)));
				if((Math.abs(absValue-shouldPower) > getPrecision()
						&& newValue!=0)
						|| shouldOff && newValue != 0) {// we don't know the
												// direction if state was off
					if(debug.debug) {
						debug.printInfo("changing motor-power from "
										+shouldPower +" to " +absValue
										+" (setValue was " +newValue +")");
					}
					shouldPower = absValue;
					t = new WritePacketTask(
							new byte[]{	(byte)0x13,
										outputBit,
										(byte)0x02,
										(byte)Math.min(7, shouldPower-1)});
					writerQueue.post(t);
				}
				// on/off
				if((newValue==0 && !shouldOff)
						|| (newValue != 0 && shouldOff)) {
					if(debug.debug) {
						debug.printInfo("switching motor from "
										+(shouldOff?"off":"on")
										+" to "
										+(newValue==0?"off":"on"));
					}
					shouldOff = newValue==0;
					byte state = (byte)(outputBit | (shouldOff ? 0x40 : 0x80));
					t = new WritePacketTask(
							new byte[]{	(byte)0x21,
										state});
					writerQueue.post(t);
				}
			} // end synchronized(this)
			
			// TBD: if !waitToComplete who empties the writerResultQueue?
			if(waitToComplete && t!=null) {
				if(debug.debug) {
					debug.printInfo("waiting for completion of Task " +t);
				}
				Object result = null;
				while(result == null) {
					try {
						result = writerResultQueue
									.getNextEqual(new EqualSourceEvent(t));
					} catch(InterruptedException ie) {
					}
				}
				if(debug.debug) {
					debug.printInfo("Task " +t +" completed with result "
									+result);
				}
			}
		}
		
		// let this happen per MotorSensorValueDefinition
		// it is possibly unecessarily polled that way, but is more consistent
		/*public double getValue() {
			return !off ? ((power)*(!forward?-1:1))/8.0 : 0;
		}*/
		public void processEvent(SensorEvent se) {
			// update local (should) on/forward/power
			// TBD: it would be better if valueDefinition provides the
			// real values since power/forward are undefined if off
			double newValue = se.getValue();
			synchronized(this) {
				shouldOff = newValue == 0;
				if(!shouldOff) {
					shouldForward = newValue > 0;
					// round and max are not necessary here
					shouldPower = (int)Math.abs(newValue*8);
				}
			}
			super.processEvent(se);
		}
		
		public boolean getDeliversFloatingPoint() {
			return true;
		}
		public double getPrecision() {
			return 1.0/16.0;
		}
	}
	
	// TBD: what is to gain from a special MotorPeer over an ActuatorPeer?
	// a peer side information about e.g. the power-level even if the motor is off
	
	private RobotInterfaceStringDefinition interfaceDefinition;
	public RobotInterfaceDefinition getInterfaceDefinition() {
		return interfaceDefinition;
	}
	// TBD: rewrite this to use STRING_IDS and maybe
	// getSerialPortNames(RobotInterfaceDefinition[],String[])
	public boolean conformsTo(RobotInterfaceDefinition definition) {
		if(interfaceDefinition != null
				&& definition instanceof RobotInterfaceStringDefinition) {
			String otherId = ((RobotInterfaceStringDefinition)definition)
									.getInterfaceString();
			if(otherId.startsWith("MS") || otherId.startsWith("RCX")) {
				if(otherId.length() <= 2 && otherId.startsWith("MS")) {
					// MS is our main id
/* exit 1 */		return true;
				}
				if(otherId.length() <= 3 && otherId.startsWith("RCX")) {
					// RCX is our parent id
/* exit 2 */		return true;
				}
				int mainLength = otherId.startsWith("MS") ? 2 : 3;
				String otherPort = otherId.substring(mainLength);
				try {
					int portNumber = Integer.parseInt(otherPort);
					// a portnumber had been specified
					// test all known serial port names
					String[] names = getSerialPortNames(portNumber);
					String myPort =
							interfaceDefinition
									.getInterfaceString().substring(2);
					for(int i=0; i<names.length; i++) {
						if(myPort.equals(names[i])) {
/* exit 3 */				return true;
						}
					}
				} catch(NumberFormatException nfe) {
					// a port name had been specified
					String myPort =
							interfaceDefinition
									.getInterfaceString().substring(2);
/* exit 4 */		return otherPort.equals(myPort);
				}				
			}
		}
/* exit 5 */	
		return false;
			
	}
}
