package de.jaetzold.art.platform.lego;

import de.jaetzold.art.Port;
import de.jaetzold.art.Sensor;
import de.jaetzold.art.SensorPort;
import de.jaetzold.art.SensorEvent;
import de.jaetzold.art.Actuator;
import de.jaetzold.art.ActuatorPort;
import de.jaetzold.art.RobotInterface;
import de.jaetzold.art.platform.RobotInterfaceAdapter;
import de.jaetzold.art.platform.SensorPeer;
import de.jaetzold.art.platform.ActuatorPeer;
import de.jaetzold.art.platform.BaseSensorPort;
import de.jaetzold.art.platform.BaseActuatorPort;
import de.jaetzold.art.platform.PacketFilter;
import de.jaetzold.art.platform.CorruptPacketException;

import de.jaetzold.util.Debug;
import de.jaetzold.util.Event;
import de.jaetzold.util.Task;
import de.jaetzold.util.TaskPerformer;
import de.jaetzold.util.Queue;
import de.jaetzold.util.ByteArray;
import de.jaetzold.util.StringObservable;
import de.jaetzold.util.RequeueTask;
import de.jaetzold.util.EqualSourceEvent;
import de.jaetzold.util.ExceptionEvent;
import de.jaetzold.util.StateDecider;
import de.jaetzold.util.ValueToStateStateDecider;

import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Hashtable;
import java.io.IOException;

// TBD: fix possible memory-leak in resultQueues from TaskPerformers. How to
// decide which result can be removed?
// One option is to store them with java.lang.ref.SoftReference-Objects
// this way they get removed automatically when memory becomes too full
public abstract class RCX extends RobotInterfaceAdapter {
	private static Debug debug = new Debug("de.jaetzold.art.RCX");

	protected boolean connected;
	public boolean isConnected() {
		return connected;
	}
	
	protected boolean alive;
	public boolean isAlive() {
		if(polledSensorValues.isEmpty()) {
			return alive();
		} else {
			return alive;
		}
	}

	// this one is just for the case when nothing else is polled regularly
	protected boolean alive() {
		SensorValueDefinition alive = new AliveValueDefinition();
		return ((Boolean)getSensorValue(alive)).booleanValue();
	}
	
	// 9.4.01: to satisfy the ibm-compiler
	public RCX() {}
	
	public Object getSensorValue(SensorValueDefinition valueDefinition) {
		// is it not polled?
		if(!polledSensorValues.containsKey(valueDefinition)
				// or is no value already there?
				|| !sensorValues.containsKey(valueDefinition)) {
			// TBD: don't use a new task in teh second case
			// but get it from polledSensorValues
			// and wait until it has successfully polled the value
			Task t = new GetSensorValueTask(valueDefinition);
			try {
				if(debug.debug) {
					debug.printInfo("Getting value explicitly for "
									+valueDefinition);
				}
				call(t);
			} catch(Exception e) {
				if(debug.debug) {
					debug.printError(debug.stackTraceToString(e));
				}
				// should mark the sensorvalue as invalid or so
			}
		}
		Object result = sensorValues.get(valueDefinition);
		if(debug.debug) {
			debug.printInfo("returning " +result +" from getSensorValue("
							+valueDefinition +")");
		}
		return result;
	}
	
	public synchronized void startPolling(SensorValueDefinition valueDefinition,
											SensorPeer recipient)
	{
		Set peerSet = (Set)sensorValueListeners.get(valueDefinition);
		if(peerSet == null) {
			peerSet = new HashSet();
			sensorValueListeners.put(valueDefinition, peerSet);
		}
		synchronized(peerSet) {
			if(!peerSet.contains(recipient)) {
				peerSet.add(recipient);
			}
			if(!polledSensorValues.containsKey(valueDefinition)) {
				Set definitions = (Set)packetToValueDefinitionSet
										.get(valueDefinition.getMapPacket());
				if(definitions == null) {
					definitions = new HashSet();
					packetToValueDefinitionSet
							.put(valueDefinition.getMapPacket(), definitions);
				}
				synchronized(definitions) {
					Task t;
					Iterator definitionIterator = definitions.iterator();
					if(!definitionIterator.hasNext()) {	// definitions.isEmpty()
						// no polling task already started
						t = new RequeueTask(new GetSensorValueTask(
												valueDefinition.getPacket()),
											writerQueue);
						writerQueue.post(t);
					} else {
						t = (Task)polledSensorValues
									.get(definitionIterator.next());
					}
					definitions.add(valueDefinition);
					polledSensorValues.put(valueDefinition, t);
				}
			}
		}
	}
	
	public synchronized void stopPolling(SensorValueDefinition valueDefinition,
											SensorPeer recipient)
	{
		Set peerSet = (Set)sensorValueListeners.get(valueDefinition);
		if(peerSet == null) {
			peerSet = new HashSet();
			sensorValueListeners.put(valueDefinition, peerSet);
			return;	// nothing to stop
		}
		synchronized(peerSet) {
			if(peerSet.contains(recipient)) {
				peerSet.remove(recipient);
			}
			if(polledSensorValues.containsKey(valueDefinition)
					&& peerSet.isEmpty()) {
				Set definitions = (Set)packetToValueDefinitionSet
										.get(valueDefinition.getMapPacket());
				if(definitions == null) {
					definitions = new HashSet();
					packetToValueDefinitionSet
							.put(valueDefinition.getMapPacket(), definitions);
				}
				synchronized(definitions) {
					definitions.remove(valueDefinition);
					RequeueTask rt = (RequeueTask)polledSensorValues
													.remove(valueDefinition);
					if(definitions.isEmpty()) {
						// no one listens anymore, stop polling task
						rt.stop();
					}
				}
			}
		}
	}
	
	public byte[] call(byte[] packet)
			throws Exception
	{
		Task t = new WritePacketTask(packet);
		Object result = call(t);
		if(result instanceof PacketEvent) {
			return ((PacketEvent)result).getPacket();
		}
		return null;
	}
	
	public Object call(Task t)
			throws Exception
	{
		writerQueue.post(t);
		Object result = null;
		while(result == null) {
			try {
				result = writerResultQueue.getNextEqual(new EqualSourceEvent(t));
			} catch(InterruptedException ie) {
			}
		}
		if(result instanceof ExceptionEvent) {
			throw ((ExceptionEvent)result).getException();
		}
		if(debug.debug) {
			System.out.println(getClass().getName() +": "
								+"returning " +result +" from call(" +t +")");
		}
		return result;
	}
	
	protected RCXPacketInputStream inStream;
	protected RCXPacketOutputStream outStream;
	/**	This Thread talks directly with the interface-hardware.
		It tries to read (and write) values as fast as it can and puts any
		change into a queue for further processing.
		*/
	protected Queue writerQueue = new Queue();
	protected TaskPerformer writer = new TaskPerformer(writerQueue);
	protected Queue writerResultQueue = new Queue();
	{
		writer.setResultEventQueue(writerResultQueue);
		writer.setDaemon(true);
		writer.setName(debug.className +".writer");
	}
	protected Queue readerQueue = new Queue();
	protected TaskPerformer reader = new TaskPerformer(readerQueue);
	protected RCXPacketFilterTable filterTable = new RCXPacketFilterTable();
	{
		readerQueue.post(new RequeueTask(new ReadPacketTask(filterTable),
										readerQueue));
		reader.setDaemon(true);
		reader.setName(debug.className +".reader");
	}
	
	protected Queue eventQueue = new Queue();
	protected Queue notifierQueue = new Queue();
	protected TaskPerformer notifier = new TaskPerformer(notifierQueue);
	{
		notifierQueue.post(new RequeueTask(new CondenseEventsTask(),
											notifierQueue));
		notifier.setDaemon(true);
		notifier.setName(debug.className +".notifier");
	}

	public class CondenseEventsTask implements de.jaetzold.util.Task {
		public Event perform() 
			throws InterruptedException	// in peekNext()
		{
			if(debug.debug) {
				debug.printInfo("starting ...");
			}
			// wait until at least one element is in the queue
			eventQueue.peekNext();
			// the problem is, that this way any other Tasks in the
			// notifierQueue may be blocked. This is no problem as long as
			// this task is the only one which puts tasks in the queue
			while(eventQueue.hasMoreElements()) {
				SensorEvent se = (SensorEvent)eventQueue.nextElement(); // does not block
				class DeliveryTask implements Task {
					SensorEvent event;
					SensorPeer recipient;
					DeliveryTask(SensorEvent event, SensorPeer recipient) {
						this.event = event;
						this.recipient = recipient;
					}
					public Event perform() {
						if(recipient != null) {
							recipient.processEvent(event);
						}
						return null;
					}
				}
				SensorValueDefinition svd =
							(SensorValueDefinition)se.getSource();
				Set peerSet = (Set)sensorValueListeners.get(svd);
				if(peerSet != null) {
					synchronized(peerSet) {
						Iterator peerIterator = peerSet.iterator();
						while(peerIterator.hasNext()) {
							SensorPeer peer = (SensorPeer)peerIterator.next();
							Task deliver = new DeliveryTask(se, peer);
							if(debug.debug) {
								debug.printInfo("scheduling " +se
												+" for delivery to " +peer
												+" with " +deliver);
							}
							notifierQueue.post(deliver);
						}
					}
				}
			}
			return null;
		}
	}
	
	protected Hashtable sensorValues = new Hashtable();		
	protected Hashtable polledSensorValues = new Hashtable();
	// table SensorValueDefinition -> Peer(s)
	// Peer -> PropertyChangeListener
	protected Hashtable sensorValueListeners = new Hashtable();
	protected Hashtable packetToValueDefinitionSet = new Hashtable();
	
	public class GetSensorValueTask	extends	WritePacketTask
	{
		SensorValueDefinition localDefinition;
		ByteArray mapPacket;
		public GetSensorValueTask(SensorValueDefinition defintion) {
			this(defintion.getPacket());
			localDefinition = defintion;
		}
		public GetSensorValueTask(byte[] packet) {
			super(packet);
			mapPacket = new ByteArray(packet);
		}
		
		public Event perform()
				throws Exception
		{
			Event result = super.perform();
			Set valueDefinitionSet = (Set)packetToValueDefinitionSet
															.get(mapPacket);
			if(localDefinition != null) {
				// add it to the Set
				if(valueDefinitionSet == null) {
					valueDefinitionSet = new HashSet();
				}
				valueDefinitionSet.add(localDefinition);
			}
			if(debug.debug) {
				if(valueDefinitionSet == null || valueDefinitionSet.isEmpty()) {
					debug.printError("\n\n\n why is "
									+valueDefinitionSet +" empty?\n\n\n");
				}
			}
			if(valueDefinitionSet != null) {
				synchronized(valueDefinitionSet) {
					Iterator definitionIterator = valueDefinitionSet.iterator();
					while(definitionIterator.hasNext()) {
						SensorValueDefinition svd =
							(SensorValueDefinition)definitionIterator.next();
						
						Object oldValue = sensorValues.get(svd);
						Object newValue = svd.parseResult(resultPacket);
						if(debug.debug) {
							debug.printInfo("read new Value " +newValue
										+" (old: " +oldValue +")"
										+", packet "
										+StringObservable
											.toHexString(resultPacket)
										+", def: "
										+svd);
						}
						if(!newValue.equals(oldValue)) {
							SensorEvent se = null;
							if(polledSensorValues.containsKey(svd)) {
								// create Event (including timestamp, oldval, newval, source)
								if(debug.debug) {
									debug.printInfo("-> creating event");
								}
								se = svd.createEvent(oldValue, newValue);
							}
							// update value
							sensorValues.put(svd, newValue);
							// queue Event
							if(se != null) {
								eventQueue.post(se);
							}
						}
					} // iteration
				} // synchronized
			}
			return result;
		}
	}
	
	private byte lastWrittenPacketID;
	public class WritePacketTask	implements	de.jaetzold.util.Task,
												PacketFilter
	{
		// this values should be tuned, maybe dynamically
		protected int retryTimeout = 300;
		protected int maxRetry = 5;
		
		protected PacketFilter customFilter = null;
		protected byte[] packet;
		
		public WritePacketTask(byte[] packet) {
			this.packet = packet;
		}
		public WritePacketTask(byte[] packet, PacketFilter customFilter) {
			this(packet, customFilter, 200);
		}
		public WritePacketTask(byte[] packet, PacketFilter customFilter, int timeout) {
			this(packet, timeout);
			this.customFilter = customFilter;
		}
		public WritePacketTask(byte[] packet, int timeout) {
			this(packet);
			this.retryTimeout = timeout;
		}
		
		public Event perform() 
				throws Exception
		{
			// check sequence bit
			if(lastWrittenPacketID == packet[0]) {
				packet = ByteArray.copy(packet);
				packet[0] ^= 0x08;
			}
			lastWrittenPacketID = packet[0];
			int retryCount = 0;
			synchronized(filterTable) {
				// clear resultPacket, in case this task had been requeued
				resultPacket = null;
				filterTable.setFilterFor((byte) ~ packet[0], this);
				// clear resultPacket, in case there came another packet too early
				resultPacket = null;
				try {
					while(!done() && retryCount < maxRetry) {
						outStream.writePacket(packet);
						retryCount++;
						filterTable.wait(retryTimeout);
						//if(inStream.available()>0 && no timeout) wait rest of time
					}
				} catch(IOException ioe) {
					alive = false;
					connected = false;
					throw ioe;
				}
				connected = true;
				filterTable.removeFilterFor((byte) ~ packet[0]);
				
				if(resultPacket != null) {
					alive = true;
				} else {
					alive = false;
				}
			}
			return new PacketEvent(this, resultPacket);
		}
		
		protected byte[] resultPacket;
		public Event process(byte[] packet)
				throws CorruptPacketException
		{
			if(debug.debug) {
				debug.printInfo("filtering "
								+StringObservable.toHexString(packet)
								+" as answer to "
								+StringObservable.toHexString(this.packet));
			}
			// for the right id is taken care by RCXPacketFilterTable
			// after that we accept every packet
			resultPacket = packet;
			Event result;
			if(customFilter == null) {
				result =  null;
			} else {
				result =  customFilter.process(packet);
			}
			synchronized(filterTable) {
				filterTable.notifyAll();
			}
			return result;
		}
		public boolean done() {
			return resultPacket != null;
		}
	}
	
	public class PacketEvent	extends		java.util.EventObject
								implements	de.jaetzold.util.Event
	{
		protected byte[] packet;
		public byte[] getPacket() {
			return packet;
		}
		public PacketEvent(Object source, byte[] packet) {
			super(source);
			this.packet = packet;
		}
		public Event process() { return null; }
		public String toString() {
			return super.toString() +"[" +StringObservable.toHexString(packet) +"]";
		}
	}
	
	public class Null {
		public boolean equals(Object o) {
			return o == null || o.getClass() == this.getClass();
		}
	}
	
	public class ReadPacketTask		implements	de.jaetzold.util.Task
	{
		protected PacketFilter filter;
		public ReadPacketTask(PacketFilter filter) {
			// filter has to be non-null
			this.filter = filter;
		}

		public Event perform()
				throws	CorruptPacketException,
						IOException
		{
			byte[] packet = inStream.readPacket();

			// check for validity:
			if(packet.length == 0) {
				if(debug.debug) {
					debug.printInfo("read packet with length 0 ***");
				} else {
					return null;
				}
			}

			// let the filter generate an Event if it wants to
			Event result = null;
			synchronized(filter) {
				result = filter.process(packet);
				//filter.notifyAll();
			}
				
			return result;
		}
	}
			
	public interface SensorValueDefinition {
		public byte[] getPacket();
		public Object parseResult(byte[] packet);
		public SensorEvent createEvent(Object oldValue, Object newValue);
		public byte[][] getConfigurePackets();
		// TBD: let Defs create their GetSensorValueTask instead and give the Task an equals depending on the packet
		public ByteArray getMapPacket();
	}
	
	protected static final byte ALIVE		= (byte)0x10;
	public class AliveValueDefinition	implements	SensorValueDefinition {
		protected byte[] packet;
		protected ByteArray mapPacket;
		public AliveValueDefinition() {
			packet = new byte[]{ALIVE};
			mapPacket = new ByteArray(packet);
		}
		public byte[] getPacket() {
			return packet;
		}
		public boolean equals(Object o) {
			if(o.getClass() == this.getClass()) {
				return true;
			}
			return false;
		}
		public int hashCode() {
			return packet[0];
		}
		public Object parseResult(byte[] packet) {
			if(packet != null) {
				return new Boolean(true);
			} else {
				return new Boolean(false);
			}
		}
		public SensorEvent createEvent(Object oldValue, Object newValue) {
			return null;
		}
		public byte[][] getConfigurePackets() {
			return null;
		}
		public ByteArray getMapPacket() {
			return mapPacket;
		}
	}
	public class MotorSensorValueDefinition extends PollSensorValueDefinition {
		public MotorSensorValueDefinition(byte number) {
			super((byte)0x03, number);
		}
// 3   motor state <0,1,2> <0x07=power 0x08=fwd 0x40=off 0x80=on 0x00=float	
		private double motorStateToValue(short s) {
			boolean on = (s & 0x80) > 0;
			double result = 0;
			if(on) {
				boolean forward = (s & 0x08) > 0;
				int power = (s & 0x07)+1;
				result = (power*(!forward?-1:1))/8.0;
			}
			return result;
		}
		public Object parseResult(byte[] packet) {
			if(packet != null) {
				short newValue = ByteArray.toShort(packet[2], packet[1]);
				double result = motorStateToValue(newValue);
				if(debug.debug) {
					debug.printInfo(this +" parsing " +newValue +" to " +result);
				}
				return new Double(result);
			} else {
				if(debug.debug) {
					debug.printInfo(this +" parsing nothing!");
				}
				return null;
			}
		}
		public byte[][] getConfigurePackets() {
			return null;
		}
	}
	
	/** Definition for the standard booleanconversion of the RCX. Instead of configuring the sensor to boolean and then polling the regular value, the source for the standard boolean converion, which exists for every sensor type, is used.
		However, the sensor <emph>is</emph> configured to be a touch-sensor, but that should cause no problems if this is changed later.
		*/
	protected class BooleanConversionValueDefinition
					extends PollSensorValueDefinition
	{
		public BooleanConversionValueDefinition(byte argument) {
			super((byte)0x0d, argument);
		}
		public byte[][] getConfigurePackets() {
			// set sensor type: 0x32 <sensor> <type>
			// configure to standard boolean-mode/touch-type
			return new byte[][] { { 0x32, sensorNumber, 0x01} };
		}
	}

	/** Definition for the standard raw value of the RCX. Instead of configuring the sensor to raw and then polling the regular value, the source for the raw value, which exists for every sensor type, is used.
		Note: this is the only normal SensorValueDefinition which does not configure the sensor.
		*/
	protected class StateConversionValueDefinition
					extends PollSensorValueDefinition
	{
		public StateConversionValueDefinition(byte argument) {
			super((byte)0x0d, argument);
		}
		public byte[][] getConfigurePackets() {
			return null;
		}
	}

// example for closures to put into the RobotInterface

	/** Definition for the standard raw value of the RCX. Instead of configuring the sensor to raw and then polling the regular value, the source for the raw value, which exists for every sensor type, is used.
		However, the sensor <emph>is</emph> configured to a raw, but that should cause no problems if this is changed later.
		*/
	protected class RawValueDefinition
					extends PollSensorValueDefinition
	{
		public RawValueDefinition(byte argument) {
			super((byte)0x0c, argument);
		}
		public byte[][] getConfigurePackets() {
			// set sensor type: 0x32 <sensor> <type>
			// configure to standard raw-mode/type
			// TBD: would it be better to just configure the mode and leave the type?
			return new byte[][] { { 0x32, sensorNumber, 0x00} };
			// set sensor mode: 0x42 <sensor> <mode+slope>
			// e.g. new byte[] { 0x42, sensorNumber, 0x07<<<5}
		}
	}

// example for closures to put into the RobotInterface
// it turned out to be too confusing in this case
/*	public class StateSensorValueDefinition	extends PollSensorValueDefinition {
		public StateSensorValueDefinition(byte source, byte argument) {
			super(source, argument);
		}
		
		StateDecider defaultDecider = new ValueToStateStateDecider();
		StateDecider decider = defaultDecider;
		public StateDecider getStateDecider() {
			return decider;
		}
		public void setStateDecider(StateDecider decider) {
			if(decider != null) {
				this.decider = decider;
			} else {
				this.decider = this.defaultDecider;
			}
		}
		
		private double lastState;
		public Object parseResult(byte[] packet) {
			if(packet != null) {
				int newValue = ByteArray.toShort(packet[2], packet[1]);
				double newState = decider.nextStateValue(lastState, newValue);
				lastState = newState;
				if(debug.debug) {
					debug.printInfo(this +" parsing " +newValue +" to " +newState);
				}
				return new Double(newState);
			} else {
				if(debug.debug) {
					debug.printInfo(this +" parsing nothing!");
				}
				return null;
			}
		}
	}*/

	protected static final byte POLL		= (byte)0x12;
	public class PollSensorValueDefinition	implements	SensorValueDefinition {
		protected byte[] packet;
		protected byte sensorNumber;
		protected ByteArray mapPacket;
		public PollSensorValueDefinition(byte source, byte argument) {
			packet = new byte[]{POLL, source, argument};
			sensorNumber = argument;
			mapPacket = new ByteArray(packet);
		}
		public byte[] getPacket() {
			return packet;
		}
		public boolean equals(Object o) {
			/*	A try to write a clever equals().
				The following should also work with subclasses, but
				not with subclasses from two different branches
			if(o instanceof SensorValueDefinition)
				if(o.getClass() != this.getClass()
						&& this.getClass().isInstance(o)) {
					return o.equals();
				} else if(o.getClass().isInstance(this)
							&& ((SensorValueDefinition)o).packet[1] == packet[1]
							&& ((SensorValueDefinition)o).packet[2] == packet[2]) {
					return true;
				}
			}
			*/
			// stick to the conservative way, should be evaluated faster
			if(o!= null && o.getClass() == this.getClass()
						&& ((PollSensorValueDefinition)o).packet[1] == packet[1]
						&& ((PollSensorValueDefinition)o).packet[2] == packet[2]) {
				return true;
			}
			return false;
		}
		public int hashCode() {
			return (packet[2]<<16)+(packet[1]<<8)+packet[0];
		}
		public Object parseResult(byte[] packet) {
			if(packet != null) {
				return new Integer(ByteArray.toShort(packet[2], packet[1]));
			} else {
				return null;
			}
		}
		public SensorEvent createEvent(Object oldValue, Object newValue) {
			if(oldValue == null) {
				oldValue = new Integer(0);
			}
			return new SensorEvent(this,
									null,
									System.currentTimeMillis(),
									((Number)oldValue).doubleValue(),
									((Number)newValue).doubleValue(),
								// TBD: an own getDelFP-Method would be better
									newValue instanceof Double,
									null);
		}
		public byte[][] getConfigurePackets() {
			return null;
		}
		public ByteArray getMapPacket() {
			return mapPacket;
		}
	}
	// will be set by subclasses
	protected static  byte[] UNLOCK_FIRMWARE;
	public boolean unlockFirmware() {
		try {
			call(new WritePacketTask(UNLOCK_FIRMWARE, 2000));
		} catch(Exception e) {}
		// should check result of that call
		return true;
	}

	class RCXSensorPort extends BaseSensorPort {
		RCXSensorPort(RobotInterface iface,
					String portName,
					int portNumber,
					int minimumValue,
					int maximumValue,
					int granularity,
					Port[] excludedPorts) {
			super(iface, portName, portNumber, minimumValue, maximumValue, granularity, excludedPorts);
		}
		public SensorPeer getPeerFor(Sensor s) {
			SensorPeer peer = getSensorPeer(this, s);
			return peer;
		}
	}
	class RCXActuatorPort extends BaseActuatorPort {
		RCXActuatorPort(RobotInterface iface,
						String portName,
						int portNumber,
						String type,
						int minimumValue,
						int maximumValue,
						int granularity,
						Port[] excludedPorts) {
			super(iface, portName, portNumber, type, minimumValue, maximumValue, granularity, excludedPorts);
		}
		public SensorPeer getPeerFor(Sensor s) {
			ActuatorPeer peer = getActuatorPeer(this, s);
			return peer;
		}
		public ActuatorPeer getPeerFor(Actuator a) {
			ActuatorPeer peer = getActuatorPeer(this, a);
			return peer;
		}
	}
	
	protected abstract SensorPeer getSensorPeer(SensorPort p, Sensor s);
	protected abstract ActuatorPeer getActuatorPeer(ActuatorPort p, Sensor s);
}
