package de.jaetzold.art.subsumption;

import de.jaetzold.art.BooleanSensor;
import de.jaetzold.art.SensorListener;
import de.jaetzold.art.SensorEvent;
import de.jaetzold.util.Task;
import de.jaetzold.util.Event;
import de.jaetzold.util.Debug;

import java.util.Map;
import java.util.HashMap;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Iterator;

// a behavior doesn't know it's priority!
// thats what some kind of builder does set

// a scheduler which adds listeners
// another scheduler-version which always asks the sensor directly
// might also be of interest

// better a SchedulerTask and a Scheduler which manages the performer
public class SchedulerTask implements Task {
	private final Debug debug =
						new Debug("de.jaetzold.art.subsumption.SchedulerTask");
	{ if(debug.debug) debug.objectIdentifier = debug.getIdentifierFor(this); }
	// by using double a corresponding insert-method is
	// (theoretically) not necessary
	// but maybe not that easy to understand and program?
	private Map priorityToBehaviorEntry = new HashMap();
	private SortedMap behaviorMap = new TreeMap();
	public synchronized void add(Entry entry) {
		addBehaviorFor(entry.getPriority(),
						entry.getBooleanSensor(),
						entry.getBehavior());
	}
	public synchronized void addBehaviorFor(Comparable priority,
											BooleanSensor sensor,
											Task behavior)
	{
		if(priorityToBehaviorEntry.containsKey(priority)) {
			throw new IllegalArgumentException("Priority " +priority
												+" already assigned.");
		}

		LocalEntry entry = new ListeningEntry(	priority,
												sensor,
												behavior);

		priorityToBehaviorEntry.put(priority, entry);
		if(debug.debug) {
			debug.printInfo("added behavior " +behavior
							+" for sensor " +sensor
							+" with priority " +priority);
		}
	}

	public synchronized void removeBehaviorFor(Comparable priority) {
		LocalEntry entry = 
				(LocalEntry)priorityToBehaviorEntry.remove(priority);
		if(entry != null) {
			entry.removed();
			if(debug.debug) {
				debug.printInfo("removed behavior " +entry.getBehavior()
								+" for sensor " +entry.getBooleanSensor()
								+" with priority " +priority);
			}
		}
	}
	
	public Iterator getEntryIterator() {
		return new Iterator() {
			Iterator local = priorityToBehaviorEntry.values().iterator();
			public boolean hasNext() {
				return local.hasNext();
			}
			private Comparable lastEntryPriority;
			public Object next() {
				Entry nextEntry = (Entry)local.next();
				lastEntryPriority = nextEntry.getPriority();
				return nextEntry;
			}
			public void remove() {
				Comparable localPriority = lastEntryPriority;
				lastEntryPriority = null;
				removeBehaviorFor(localPriority);
			}
		};
	}

	public interface Entry {
		public BooleanSensor getBooleanSensor();
		public Task getBehavior();
		public Comparable getPriority();
	}
	
	private class LocalEntry implements Entry {
		protected BooleanSensor sensor;
		protected Task behavior;
		protected Comparable priority;
		protected boolean removed;

		public LocalEntry(Comparable priority,
							BooleanSensor sensor,
							Task behavior)
		{
			this.sensor = sensor;
			this.behavior = behavior;
			this.priority = priority;
		}

		public void activate() {
			synchronized(behaviorMap) {
				if(!removed) {
					behaviorMap.put(priority, behavior);
				}
			}
		}

		public void deactivate() {
			synchronized(behaviorMap) {
				behaviorMap.remove(priority);
			}
		}

		public void removed() {
			synchronized(behaviorMap) {
				removed = true;
			}
			deactivate();
		}
		
		public BooleanSensor getBooleanSensor() {
			return sensor;
		}
		public Task getBehavior() {
			return behavior;
		}
		public Comparable getPriority() {
			return priority;
		}
	}
	
	private class ListeningEntry extends LocalEntry {
		private SensorListener listener;
		private boolean removed;

		public ListeningEntry(Comparable priority,
							BooleanSensor sensor,
							Task behavior)
		{
			super(priority, sensor, behavior);

			listener = new SensorListener() {
				public void processEvent(SensorEvent se) {
					boolean newState = ListeningEntry.this.sensor
											.convertToBoolean(se.getValue());
					if(newState) {
						activate();
					} else {
						deactivate();
					}
				}
			};
			synchronized(behaviorMap) {
				sensor.addSensorListener(listener);
				// TBD: catch the NotConnectedException?
				// why would one want to use a not-connected BooleanSensor?
				
				// this is necessary because the listener
				// only gets informed when something _changes_
				if(sensor.getBooleanValue()) {
					activate();
				}
			}
		}

		public void removed() {
			super.removed();
			sensor.removeSensorListener(listener);
		}
	}

	private Object lastPriority;
	public Event perform() throws Exception {
		Task behavior;
		Object highestPriority;
		synchronized(behaviorMap) {
			highestPriority = behaviorMap.lastKey();
			behavior = (Task)behaviorMap.get(highestPriority);
		}
		if(debug.debug) {
			if(highestPriority != lastPriority) {
				debug.printInfo("switching from priority " +lastPriority
								+" to " +highestPriority
								+" with behavior " +behavior);
				lastPriority = highestPriority;
			}
		}
		return behavior.perform();
		// in a loop
		// or better with SubsumptionListeners
		// select from all behaviors whose condition is true
		// the one which has the highest priority
		// and tell it to 'behave'
	}
}
