package de.jaetzold.util;

import java.util.Date;
import java.util.Observable;
import java.util.Observer;

/**	Ein Observable fuer Zeit. Immer dann wenn getDate() aufgerufen wird und sich die Zeit aufgrund der mit <code>precision</code> festgelegten Genauigkeit geaendert hat, wird dieses Observable als geaendert markiert. Da getDate() bei jedem Aufruf von notifyObservers() auch aufgerufen wird, braucht man sich um den Aufruf nicht explizit zu kuemmern.
Also z.B. bei precision==MINUTES jede neue Minute.
*/
public class DateObservable extends Observable {
    /** Erzeugt eine neues DateObservable mit precision = MINUTES */
    public DateObservable() {
        this(MINUTES);
    }
    /** Erzeugt eine neues DateObservable mit der angegebenen precision. */
    public DateObservable(int precision) {
        date = new Date(System.currentTimeMillis());
        setPrecision(precision);
    }

    /** Die Konstante fuer precision = MILLISECONDS. */
    public static final int MILLISECONDS = 0;
    /** Die Konstante fuer precision = SECONDS. */
    public static final int SECONDS = 1;
    /** Die Konstante fuer precision = MINUTES. */
    public static final int MINUTES = SECONDS*60;
    /** Die Konstante fuer precision = HOURS. */
    public static final int HOURS = MINUTES*60;
    /** Die Konstante fuer precision = DAYS. */
    public static final int DAYS = HOURS*24;
//    public static final long WEEKS = DAYS*7;
    // Die vielfachen haben keine wirkliche Bedeutung
    /** Die Konstante fuer precision = MONTHS. */
    public static final int MONTHS = DAYS*28;
    /** Die Konstante fuer precision = YEARS. */
    public static final int YEARS = DAYS*365;

    /** Die precision dieses Objektes. */
    protected int precision;
    /** Liefert die aktuelle precision fuer dieses Objekt. */
    public int getPrecision() {
        return precision;
    }
    /** Setzt die neue precision fuer dieses Objekt. */
    public void setPrecision(int precision) {
        this.precision = precision;
    }

    /** Vergleicht die uebergebenen Date-Objekte mit der precision dieses DateObservable.
        Ist der eingestellte Wert fuer Precision unbekannt, so sind d1 und d2 immer unterschiedlich.
        @param d1 das mit d2 zu vergleichende Date-Objekt.
        @param d2 das mit d1 zu vergleichende Date-Objekt.
        @return true, wenn d1 und d2 sich mit der precision dieses DateObservable unterscheiden.
        */
    public boolean differWithPrecision(Date d1, Date d2) {
        boolean result = false;
        switch(precision) {
            case MILLISECONDS:
                result = result || d1.getTime() != d2.getTime();
            case SECONDS:
                result = result || d1.getSeconds() != d2.getSeconds();
            case MINUTES:
                result = result || d1.getMinutes() != d2.getMinutes();
            case HOURS:
                result = result || d1.getHours() != d2.getHours();
            case DAYS:
                result = result || d1.getDate() != d2.getDate();
            case MONTHS:
                result = result || d1.getMonth() != d2.getMonth();
            case YEARS:
                result = result || d1.getYear() != d2.getYear();
                break;
            default:
                result = true;
        }
        return result;
    }

    /** Das letzte mit getDate() gelieferte Datum. */
    protected Date date;
    /** Liefert das aktuelle Datum mit Genauigkeit precision. Hat sich das Datum seit dem letzten Aufruf geaendert wird dieses Observable als geaendert markiert.
        */
    public Date getDate() {
        Date date = new Date(System.currentTimeMillis());
        if(differWithPrecision(date, this.date)) {
            this.date = date;
            setChanged();
        }
        return this.date;
    }

    /** Fuegt einen neuen Observer hinzu. Das Observable wird dafuer als geandert markiert. */
    public void addObserver(Observer o) {
        super.addObserver(o);
        setChanged();
    }

    /** Benachrichtigt alle Observer wenn eine Aenderung im Datum stattgefunden hat. Um das festzustellen wird als erstes getDate() aufgerufen.
        @see #getDate()
        */
    public void notifyObservers() {
        getDate();
        super.notifyObservers();
    }
}
