package de.jaetzold.util;

import java.util.Map;
import java.util.Hashtable;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.math.BigInteger;

/**	This class encapsulates common code for debugging */
public class Debug {
	// there could be more Objects (I guess), but long should be enough
	// for debugging ... :-)
	private static long numberOfCreatedObjectIdentifiers = 0;
	public String createObjectIdentifier() {
		synchronized(Debug.class) {
			String result;
			result = numberOfCreatedObjectIdentifiers +"";
			numberOfCreatedObjectIdentifiers++;
			return result;
		}
	}
	private static Map referenceToIdentifier = new Hashtable();
	// beware that equality of references is determinded by equals()
	public String getIdentifierFor(Object o) {
		synchronized(Debug.class) {
			String result = (String)referenceToIdentifier.get(o);
			if(result == null) {
				String unqualifiedClassname = o +"";
				if(o != null) {
						String qualifiedClassName = o.getClass().getName();
						int dotIndex = qualifiedClassName.lastIndexOf(".");
						// if "." is not found the whole name is used
						// (because dotIndex is -1 then)
						unqualifiedClassname =
								qualifiedClassName.substring(dotIndex+1);
				}
				String id = createObjectIdentifier();
				result = unqualifiedClassname +"-" +id;
				referenceToIdentifier.put(o, result);
				if(debug) {
					printInfo("created new identifier " +result +" for " +o);
				}
			}
			return result;
		}
	}
	/** Name of the class, including package-name. Useful for knowing the classname even in static-methods. Used as the header for any properties.
		@see #debug
	*/
	public String className;
	public String objectIdentifier = "";
	
	/** Constructs a new debug-object for the Class named <code>className</code>.
		The given String should be full qualified classname (including package-name) for the normal funtion of this Debug-Object but any String is possible as the name isn't used to access a class.
		*/
	public Debug(String className) {
		this.className = className;
		
		// initialize debug to default-value
		String v = getProperty("debug");		
		// results to false if v!="true" (ignoring case)
		debug = Boolean.valueOf(v).booleanValue();
		
		// initialize debug to default-value
		v = getProperty("errorLevel");		
		try {
			errorLevel = Integer.parseInt(v);
		} catch(NumberFormatException nfe) {
			if(v != null) {
				printError("Property errorLevel has a bad format: " +v);
			}
			errorLevel = 0;
		}
		
		// initialize debug to default-value
		v = getProperty("infoLevel");		
		try {
			infoLevel = Integer.parseInt(v);
		} catch(NumberFormatException nfe) {
			if(errorLevel>=0 && v != null) {
				printError("Property infoLevel has a bad format: " +v);
			}
			infoLevel = 0;
		}
	}
	
	public String getProperty(String name) {
		String v = null;
		v = System.getProperty(className +"." +name);
		int index = className.length();
		while(v==null && (index=className.lastIndexOf('.', index-1)) != -1) {
			v = System.getProperty(className.substring(0,index) +"." +name);
		}
		if(v==null) {
			v = System.getProperty(name);
		}
		
		return v;
	}
	
	public String getCurrentStackTraceString() {
		Exception e = new Exception();
		e.fillInStackTrace();
		return stackTraceToString(e);
	}
	public String stackTraceToString(Throwable t) {
		StringWriter result = new StringWriter();
		t.printStackTrace(new PrintWriter(result));
		return result.toString();
	}
	
	/** Print className and message on stderr
		@see #className
		@see #printInfo(String)
	*/
	public void printError(String message) {
		System.err.println(className +": (" +objectIdentifier +") " +message);
	}
	/** Print className and message on stdout
		@see #className
		@see #printError(String)
	*/
	public void printInfo(String message) {
		System.out.println(className +": (" +objectIdentifier +") " +message);
	}
	/** Print message on stdout if level<=infoLevel or debug==true.
		@see #infoLevel
		@see #debug
	*/
	public void printInfo(int level, String message) {
		if(level <= infoLevel || debug) {
			System.out.println(message);
		}
	}
	/** Print message on stdout if level<=errorLevel or debug==true.
		@see #errorLevel
		@see #debug
	*/
	public void printError(int level, String message) {
		if(level <= errorLevel || debug) {
			System.err.println(message);
		}
	}
	
	/** Reflects the value of the property className.debug at the time this class had been initialized. If such a property does not exist enclosing packages are tried for this property.
		<p>Example:
		If the package name is a.b and the class is called C, debug will be true if the property
		<ul>
		<li>a.b.C.debug exists and is true</li>
		<li>a.b.debug exists and is true and a.b.C.debug doesn't exist</li>
		<li>a.debug exists and is true and a.b.debug as well as a.b.C.debug don't exist</li>
		<li>debug exists and is true and a.debug as well as a.b.debug and a.b.C.debug don't exist</li>
		</ul>
		In any other case debug will be set to false.</p>
		@see #className
		@see #printError(String)
		@see #printInfo(String)
	*/
	public boolean debug = false;

	/** Set verbosity of error messages.
		The higher the value, the more and/or longer messages. A value of -1 switches (nearly) all error messages off. This value can be set as a System-Property similar to debug.
		@see #debug
		@see #printError(int,String)
		*/
	public int errorLevel;

	/** Set verbosity of informational messages.
		The higher the value, the more and/or longer messages. A value of -1 switches (nearly) all informational messages off. This value can be set as a System-Property similar to debug.
		@see #debug
		@see #printInfo(int,String)
		*/
	public int infoLevel;
}
