package de.jaetzold.util;

import java.util.Vector;

/**	A class for passing events from one Thread to another.
	Usually there is only one thread on one end, which retrieves the posted events
	by calling <code>getNextEvent()</code>. However there can be more than one Thread
	calling <code>getNextEvent()</code> but then, each event will always be returned
	only to one of them.
	*/
public class EventQueue {
    /** Vector in which all the events get stored before they get processed. */
    private Vector eventQueue = new Vector();

	/** Tells whether there are any events in the queue.
		This only ensures that the next call to <code>getNextEvent()</code> does not block
		if both calls are in the same block, synchronized to this queue.
		#see #getNextEvent()
		*/
	public boolean isEmpty() {
		return eventQueue.isEmpty();
	}
	/** Return but not delete. */
	public synchronized Event peekNextEvent(int timeout)
        	throws InterruptedException
	{
		if(timeout>0) {
			long startMillis = System.currentTimeMillis();
			long waitTime = 0;
        	while(eventQueue.isEmpty()) {
				waitTime = System.currentTimeMillis()-startMillis;
				if(waitTime>=timeout) { return null; }
            	this.wait(timeout-waitTime);
        	}
		} else {
        	while(eventQueue.isEmpty()) {
            	this.wait();
        	}
		}
        Event e = (Event)eventQueue.firstElement();
        return e;
	}
    /** Returns and deletes the next event in the queue.
		If the queue is already empty, this method blocks until a new event
		gets added with <code>post(de.jaetzold.util.Event)</code>.
		@see #post(de.jaetzold.util.Event)
		@see #getNextEvent(int)
        */
    public synchronized Event getNextEvent()
        	throws InterruptedException
    {
		return getNextEvent(0);
	}
    /** Returns and deletes the next event in the queue if there is any before
		the specified timeout. If the queue is empty, this method blocks until
		a new event gets added with <code>post(de.jaetzold.util.Event)</code> or
		until after <code>timeout</code> milliseconds have elapsed. In this case
		it is possible that this method returns <code>null</code>.
		@return the next Event in the Queue if there is any before the timeout,
				null otherwise.
		@see #post(de.jaetzold.util.Event)
		@see #getNextEvent()
        */
    public synchronized Event getNextEvent(int timeout)
        	throws InterruptedException
    {
        Event e = peekNextEvent(timeout);
        eventQueue.removeElementAt(0);
        return e;
    }

    /** Adds the given event-object to the end of the queue.
		This method notifies all Threads currently waiting on this queues monitor.
		*/
    public synchronized void post(Event theEvent) {
        eventQueue.addElement(theEvent);
        this.notifyAll();
    }
	
	/**	Removes the first occurence of an element equal to the given Event from
		this queue if there is any. It calls <code>theEvent.equals()</code> to
		determine equality.
		Implementation note: It relies on the underlying List to do it that way.
		@return	false if there didn't exit such an element in this Queue,
				true otherwise
		*/
	public synchronized boolean remove(Event theEvent) {
		return eventQueue.removeElement(theEvent);
	}
	
	/**	Removes the all elements equal to the given Event from this queue.
		All occurrences of such elements are removed by this method.
		It calls <code>theEvent.equals()</code> to determine equality.
		Implementation note: It relies on the underlying List to do it that way.
		@returns the number of removed occurrences (0 if there weren't any)
		*/
	public synchronized int removeAll(Event theEvent) {
		int result = 0;
		while(eventQueue.removeElement(theEvent)) {
			result++;
		}
		return result;
	}
	
	public boolean contains(Event event) {
		return eventQueue.indexOf(event) != -1;
	}
	
	public synchronized Event getNextEventEqual(Event event, int timeout) {
		if(timeout>0) {
			long startMillis = System.currentTimeMillis();
			long waitTime = 0;
        	while(!contains(event)) {
				waitTime = System.currentTimeMillis()-startMillis;
				if(waitTime>=timeout) { return null; }
            	this.wait(timeout-waitTime);
        	}
		} else {
        	while(!contains(event)) {
            	this.wait();
        	}
		}
		// get the Object equal to event from queue:
        Event e = (Event)eventQueue.elementAt(eventQueue.indexOf(event));
        return e;
	}
}
