package de.jaetzold.util;

import java.util.Vector;
import java.util.NoSuchElementException;

/**	A class for passing Objects from one Thread to another. The Threads have to
	take care for themselves that the Objects passed are from the correct type.
	Usually there is only one thread on the reading end, which retrieves the posted Objects
	by calling <code>getNext()</code>. However there can be more than one Thread
	calling <code>getNext()</code> but then, each object will always be returned
	only to one of them.
	*/
public class Queue implements java.util.Enumeration {
    /** Vector in which all the objects get stored before they are returned. */
    private Vector objectQueue = new Vector();

	/** Tells whether there are any objects in the queue.
		This only ensures that the next call to <code>getNext()</code> does not
		block if both calls are in the same block, synchronized to this queue.
		@see #getNext()
		*/
	public boolean isEmpty() {
		return objectQueue.isEmpty();
	}
	public boolean hasMoreElements() {
		return !isEmpty();
	}
	public Object nextElement() {
		Object result = null;
		while(true) {
			try {
				result = getNext(1);	// If there is no next element this call
										// returns after 1 millisec.
			} catch(InterruptedException ie) {
				// clear interrupt-status
				Thread.currentThread().interrupted();
				continue;
			}
			break;
		}
		if(result == null) {
			throw new NoSuchElementException();
		}
		return result;
	}
	/** Return but not delete. */
	public synchronized Object peekNext()
        	throws InterruptedException
	{
		return peekNext(0);
	}
	/** Return but not delete. */
	public synchronized Object peekNext(int timeout)
        	throws InterruptedException
	{
		if(timeout>0) {
			long startMillis = System.currentTimeMillis();
			long waitTime = 0;
        	while(objectQueue.isEmpty()) {
				waitTime = System.currentTimeMillis()-startMillis;
				if(waitTime>=timeout) { return null; }
				try {
            		this.wait(timeout-waitTime);
				} catch(InterruptedException ie) {
					// clear interrupt-status
					Thread.currentThread().interrupted();
				}
        	}
		} else {
        	while(objectQueue.isEmpty()) {
				// do not catch this InterruptException
				// so that a call without timeout can be interrupted
            	this.wait();
        	}
		}
        Object result = objectQueue.firstElement();
        return result;
	}
    /** Returns and deletes the next object in the queue.
		If the queue is already empty, this method blocks until a new event
		gets added with <code>post(java.lang.Object)</code>.
		@see #post(java.lang.Object)
		@see #getNext(int)
        */
    public synchronized Object getNext()
        	throws InterruptedException
    {
		return getNext(0);
	}
    /** Returns and deletes the next event in the queue if there is any before
		the specified timeout. If the queue is empty, this method blocks until
		a new event gets added with <code>post(de.jaetzold.util.Event)</code> or
		until after <code>timeout</code> milliseconds have elapsed. In this case
		it is possible that this method returns <code>null</code>.
		@return the next Event in the Queue if there is any before the timeout,
				null otherwise.
		@see #post(Object)
		@see #getNext()
        */
    public synchronized Object getNext(int timeout)
        	throws InterruptedException
    {
        Object result = peekNext(timeout);
        objectQueue.removeElementAt(0);
        return result;
    }

    /** Adds the given object to the end of the queue.
		This method notifies all Threads currently waiting on this queues monitor.
		*/
    public synchronized void post(Object object) {
        objectQueue.addElement(object);
        this.notifyAll();
    }
	
	/**	Removes the first occurence of an element equal to the given object from
		this queue if there is any. It calls <code>object.equals()</code> to
		determine equality.
		Implementation note: It relies on the underlying List to do it that way.
		@return	false if there didn't exit such an element in this Queue,
				true otherwise
		*/
	public synchronized boolean remove(Object object) {
		return objectQueue.removeElement(object);
	}
	
	/**	Removes the all elements equal to the given object from this queue.
		All occurrences of such elements are removed by this method.
		It calls <code>object.equals()</code> to determine equality.
		Implementation note: It relies on the underlying List to do it that way.
		@return the number of removed occurrences (0 if there weren't any)
		*/
	public synchronized int removeAll(Object object) {
		int result = 0;
		while(objectQueue.removeElement(object)) {
			result++;
		}
		return result;
	}
	
	public boolean contains(Object object) {
		return objectQueue.indexOf(object) != -1;
	}
	
	public synchronized Object getNextEqual(Object object)
        	throws InterruptedException
	{
		return getNextEqual(object, 0);
	}
	
	public synchronized Object getNextEqual(Object object, int timeout)
        	throws InterruptedException
	{
		if(timeout>0) {
			long startMillis = System.currentTimeMillis();
			long waitTime = 0;
        	while(!contains(object)) {
				waitTime = System.currentTimeMillis()-startMillis;
				if(waitTime>=timeout) { return null; }
				try {
            		this.wait(timeout-waitTime);
				} catch(InterruptedException ie) {
					// clear interrupt-status
					Thread.currentThread().interrupted();
				}
        	}
		} else {
        	while(!contains(object)) {
				// do not catch this InterruptException
				// so that a call without timeout can be interrupted
            	this.wait();
        	}
		}
		// get the Object equal to the given object from queue:
        Object result = null;
		int index = objectQueue.indexOf(object);
		if(index >= 0) {
			result = objectQueue.elementAt(index);
		}
        return result;
	}
}
