package de.jaetzold.util;

import java.beans.PropertyVetoException;
import java.beans.PropertyChangeEvent;

public class RangeStateDecider implements StateDecider {
	public double nextStateValue(double actualState, double value) {
		// needs no sync/copy to be Thread-safe
		// because ranges is not visible to the outside
		// anyone who wants to access ranges via getRanges() gets a copy.
		// if this method gets called while ranges is being initialized
		// it should do no harm too
		// (if System.arracopy() doesn't set any arrayelement first to
		//  a non-null value and then to null)
		
		if(ranges!=null) {
			for(int i=0; i<ranges.length; i++) {
				if(ranges[i] != null) {
					if(ranges[i].contains(value)) {
						return i;
					}
				}
			}
		}
		
		return 0;
	}
	
	private boolean isCacheableCache = false;
	public boolean isCacheable() {
		if(!isCacheableCache) {
			if(ranges!=null) {
				for(int i=0; i<ranges.length; i++) {
					if(ranges[i] == null
							|| !ranges[i].isCacheable()) {
						return false;
					}
				}
				isCacheableCache = true;
			}
		}
		return isCacheableCache;
	}
	
	private Range[] ranges;
	
	public Range getRanges(int index) {
		return ranges[index];
	}
	public Range[] getRanges() {
		Range[] result = null;
		if(ranges != null) {
			result = new Range[ranges.length];
			System.arraycopy(ranges, 0, result, 0, ranges.length);
		}
		return result;
	}
	public synchronized void setRanges(int index, Range range)
			throws PropertyVetoException
	{
		if(this.ranges[index] == null) {
			this.ranges[index] = range;
		} else {
			throw new PropertyVetoException(
						"A Range can not be set again after there had been assigned a non-null value to it.",
						new PropertyChangeEvent(this,
												"ranges",
												this.ranges,
												ranges));
		}
	}
	public synchronized void setRanges(Range[] ranges)
			throws PropertyVetoException
	{
		if(this.ranges == null) {
			this.ranges = new Range[ranges.length];
		// if System.arracopy() sets any arrayelement first to a
		// non-null value and then to null, nextState() has to be sync'ed too
			System.arraycopy(ranges, 0, this.ranges, 0, ranges.length);
		} else {
			throw new PropertyVetoException(
						"Ranges can not be set again after there had been assigned a non-null value to it.",
						new PropertyChangeEvent(this,
												"ranges",
												this.ranges,
												ranges));
		}
	}
	
	public RangeStateDecider() {
		// don't initialize ranges!
	}
	public RangeStateDecider(Range range) {
		this(new Range[] {range});
	}
	public RangeStateDecider(Range[] ranges) {
		try {
			setRanges(ranges);
		} catch(PropertyVetoException pve) {
			// At least in this implementation this point will be never reached
		}
	}
}
