package de.jaetzold.util;

/**	A Thread which processes Tasks. It continuously tries to get Tasks from
	its associated TaskQueue and calls their <code>perform()</code>-method.
	This call can return an Event or null.
	Any results from that call or any Exceptions are posted to the resultEventQueue
	of this TaskPerformer if it has any.
	In parts a similar functionality could've been aquired by using <code>java.util.Timer</code>
	and <code>java.util.TimerTask</code>. But this code here therefor works in Java 1.1.x
	and higher, and has some extra-features like the resultEventQueue and the associated
	possibility of passing exceptions from this thread to another.
	*/
public class TaskPerformer extends Thread {
	protected Object resultEventQueueMonitor = new Object();
	protected Queue resultEventQueue = null;
	public Queue getResultEventQueue() {
		return resultEventQueue;
	}
	public void setResultEventQueue(Queue resultEventQueue) {
		synchronized(resultEventQueueMonitor) {
			this.resultEventQueue = resultEventQueue;
		}
	}

	protected Queue taskQueue;
	protected Object taskQueueMonitor = new Object();
	public Queue getTaskQueue() {
		return taskQueue;
	}
	public void setTaskQueue(Queue taskQueue) {
		if(taskQueue == null) {
			throw new IllegalArgumentException("No null-pointer allowed.");
		}
	
		synchronized(taskQueueMonitor) {
			if(!ending) {
				this.taskQueue = taskQueue;
			} else {
				throw new IllegalStateException("This Performer has already been ended.");
			}
		}
	}
	
	public TaskPerformer(Queue taskQueue) {
		setTaskQueue(taskQueue);
	}

	private Object interruptMonitor = new Object();
	/**	Interrupts this Thread if it is currently processing the given Event.
		If the given Event is <code>null</code> then this Thread is interrupted
		if it is currently processing any Event.
		This method should be used instead of <code>interrupt()</code>, because
		there should be no reason to interrupt this Thread while doing other things
		apart from processing an Event.
		*/
	public void interrupt(Task task) {
		synchronized(interruptMonitor) {
			if(isProcessing && (this.task == task || task == null)) {
				interrupt();
			}
		}
	}
	
	private boolean ending;
	public void endPerforming() {
		if(!ending) {
			ending = true;
			interrupt();
			if(debug) {
				printInfo("Noted Request to end performing by setting ending=true.");
			}
			if(!isAlive()) {
				return;
			}
			synchronized(taskQueueMonitor) {
				while(taskQueue != null) {
					if(debug) {
						printInfo("Waiting for taskQueue to become null.");
					}
					interrupt();
					try {
						taskQueueMonitor.wait();
					} catch(InterruptedException ie) {
					}
				}
			}
		}
	}
	
	private boolean halt;
	private boolean halted;
	private Object haltMonitor = new Object();
	public void haltPerforming() {
		haltPerforming(false);
	}
	public void haltPerforming(boolean interruptAndWait) {
		synchronized(haltMonitor) {
			halt = true;
			if(!isAlive()) {
				return;
			}
			if(interruptAndWait) {
				do {
					interrupt();
					try {
						waitForHalt();
					} catch(InterruptedException ie) {
					}
				// clear and try again
				} while(Thread.interrupted());
			}
		}
	}
	public void waitForHalt() throws InterruptedException {
		synchronized(haltMonitor) {
			while(!halted && isAlive()) {
				haltMonitor.wait();
			}
		}
	}
	public void resumePerforming() {
		synchronized(haltMonitor) {
			halt = false;
			haltMonitor.notifyAll();
		}
	}
	
	private Task task;
	private boolean isProcessing = false;	
	public void run() {
		while(true) {
			task = null;
			if(debug) {
				printInfo("getting next Task ...");
			}
			try {
				synchronized(haltMonitor) {
					while(halt) {
							// signals this Thread to suspend execution
							halted = true;
							haltMonitor.notifyAll();
							haltMonitor.wait();
					}
					halted = false;
				}
				synchronized(taskQueueMonitor) {
					if(ending) {
						// signals this Thread to stop ...
						if(debug) {
							printInfo("End performing because ending==true.");
						}
						taskQueue = null;
						synchronized(haltMonitor) {
							halted = true;
							haltMonitor.notifyAll();
						}
						taskQueueMonitor.notifyAll();
						
						return;
					}
					// if there comes a ClassCastException here it will be
					// reported in the resultEventQueue
					task = (Task)taskQueue.getNext();
				}

				synchronized(interruptMonitor) {
					isProcessing = true;
				}
				if(debug) {
					printInfo("performing task " +task);
				}
				Event result = task.perform();
				if(debug) {
					printInfo("got result " +result);
				}
				synchronized(task) {
					// so one can wait to be notified when this task completes its processing
					task.notifyAll();
				}
				synchronized(interruptMonitor) {
					isProcessing = false;
				}

				if(result != null) {
					synchronized(resultEventQueueMonitor) {
						if(resultEventQueue != null) {
							if(debug) {
								printInfo("posting result " +result);
							}
							resultEventQueue.post(result);
						}
					}
				}
				
				if(interrupted()) {
					throw new InterruptedException("Interrupted during event-processing.");
				}
				// how much speedup would we gain from leaving the following out?
				yield();
												
			} catch(Exception e) {
				// put the Exception in the event-queue to report it
				synchronized(resultEventQueueMonitor) {
					if(resultEventQueue != null) {
						Object src = task!=null ? (Object)task : (Object)this;
						resultEventQueue.post(new ExceptionEvent(src, e));
					}
				}
				if(debug) {
					printError("Exception while processing "
										+task
										+":");
					e.printStackTrace();
				}
			}
		}
	}
	
	// The rest of the code are just some standard-methods i commonly use for debugging-purposes:
	/** Name of the class, including package-name. Useful for knowing the classname even in static-methods. Used as the header for any properties.
		@see #debug
	*/
	private static String className = "de.jaetzold.util.TaskPerformer";
	
	/** Print className, this.toString() and message on stderr
		@see #className
		@see #printErrorStatic(String)
		@see #printInfo(String)
		@see #printInfoStatic(String)
	*/
	private void printError(String message) {
		System.err.println(className +"(" +this +")" +": " +message);
	}
	/** Print className, this.toString() and message on stdout
		@see #className
		@see #printError(String)
		@see #printErrorStatic(String)
		@see #printInfoStatic(String)
	*/
	private void printInfo(String message) {
		System.out.println(className +"(" +this +")" +": " +message);
	}
	/** Print className and message on stderr
		@see #className
		@see #printError(String)
		@see #printInfo(String)
		@see #printInfoStatic(String)
	*/
	private static void printErrorStatic(String message) {
		System.err.println(className +": " +message);
	}
	/** Print className and message on stdout
		@see #className
		@see #printError(String)
		@see #printErrorStatic(String)
		@see #printInfo(String)
	*/
	private static void printInfoStatic(String message) {
		System.out.println(className +": " +message);
	}
	
	/** Reflects the value of the property className.debug at the time this class had been initialized. If such a property does not exist enclosing packages are tried for this property.
		<p>Example:
		If the package name is a.b and the class is called C, debug will be true if the property
		<ul>
		<li>a.b.C.debug exists and is true</li>
		<li>a.b.debug exists and is true and a.b.C.debug doesn't exist</li>
		<li>a.debug exists and is true and a.b.debug as well as a.b.C.debug don't exist</li>
		<li>debug exists and is true and a.debug as well as a.b.debug and a.b.C.debug don't exist</li>
		</ul>
		In any other case debug will be set to false.</p>
		@see #className
		@see #printError(String)
		@see #printErrorStatic(String)
		@see #printInfo(String)
		@see #printInfoStatic(String)
	*/
	private static boolean debug;
	static {
		String v = null;
		v = System.getProperty(className+".debug");
		int index = className.length();
		while(v==null && (index=className.lastIndexOf('.', index-1)) != -1) {
			v = System.getProperty(className.substring(0,index)+".debug");
		}
		if(v==null) {
			v = System.getProperty("debug");
		}
		
		// results to false if v!="true" (ignoring case)
		debug = Boolean.valueOf(v).booleanValue();
	}
}
