package de.jaetzold.util;

import java.awt.*;
import java.awt.event.*;

/** TextOutput should be an interface later. For now it is everything what deals with user-interaction, also the Input.
The Interface "TextOutput" should provide the ability to:
"append"	Strings;
"clear"		the output, or
"set"	it to "";
maybe:
"get"		everything which is shown at the moment;
*/
public class TextOutput extends Panel implements CommandLine {
    /**	Main output. */
    protected TextArea	output;

    /**	This is where we get the Input-Events/Commands from. */
    protected TextField	input;

    /**	This is where we present the prompt. */
    protected Label		prompt;

    /**	This is our own buffer of the output. In Windows95 the TextArea cant contain an unlimited amount of text, so I try to limit it somehow. */
    protected StringBuffer	content;

    /**	The Array where the Commands for the history are stored. It has HISTLENGTH elements. */
    protected String[]	history;

    /** Thats where we are when we scroll through our history. */
    protected int			histindex;

    /**	The last index in History which contains a Command. This Command has not been send to our listener yet. */
    protected int			histtop;

    /**	It is the size of all fonts. */
    protected static final int FONTSIZE = 18;

    /**	Every prompt longer than PROMPTLENGTH will be cut on the left side. */
    protected static final int PROMPTLENGTH = 22;

    /**	The Buffer is truncated to SCROLLBACKMIN if it exeeds SCROLLBACKMAX. */
    protected static final int SCROLLBACKMAX = 12000;

    /**	A greater difference between min and max increases the time to the next call of setText() to the TextArea which always results in some flickering. */
    protected static final int SCROLLBACKMIN = 10000;

    /**	Maximum number of Command Strings in the History. */
    protected static final int HISTLENGTH = 200;

    /**	If this is true, successive equal Commands are only stored once. */
    protected static final boolean HISTSQUEEZE = true;

    /**	Constructs a new TextOutput whith default parameters. The output-aerea will have 30 lines & 80 columns. The Font is "Monospaced".
        The Windowtitle is set to the given parameter.
        */
    public TextOutput(int rows, int cols) {
        output = new TextArea("",rows,cols,TextArea.SCROLLBARS_VERTICAL_ONLY);
        input = new TextField();
        // PROMPTLENGTH Zeichen mindestens breit
        StringBuffer s = new StringBuffer();
        for (int i=0; i< PROMPTLENGTH ; i++)
            s.append(" ");
        prompt = new Label(s.toString(), Label.RIGHT);
        content = new StringBuffer(SCROLLBACKMAX);
        history = new String[HISTLENGTH];
        histtop = 0;
        histindex = histtop;

        output.setEditable(false);
        output.setFont(new Font("Monospaced", Font.BOLD, FONTSIZE));

        input.setFont(new Font("Monospaced", Font.PLAIN, FONTSIZE));

        prompt.setFont(new Font("Monospaced", Font.PLAIN, FONTSIZE));

        this.setLayout(new BorderLayout());
        this.add("Center", output);
        Panel p = new Panel(new BorderLayout());
        this.add("South", p);
        p.add("West", prompt);
        p.add("Center", input);
        if(output.isEditable()) {
            System.err.println("why?"); System.err.flush();
        }
        input.requestFocus();
    }

    /**	In Windows95 the TextArea cant contain an unlimited amount of text, so I try to limit it somehow.
        */
    public synchronized void append(Object obj) {
        String str = obj.toString();
        int len = str.length() + content.length();
        if (SCROLLBACKMAX < len) {
            char[]	cbuf = new char[SCROLLBACKMIN];
            if(str.length() <= SCROLLBACKMIN) {
                content.getChars(len - SCROLLBACKMIN, content.length(), cbuf, 0);
                str.getChars(0, str.length(), cbuf, SCROLLBACKMIN - str.length());
                } else {
                    str.getChars(str.length()-SCROLLBACKMIN, str.length(), cbuf, 0);
                    }
            str = new String(cbuf);
            content = new StringBuffer(str);
            output.setText(str);
            output.setCaretPosition(content.length());
            return;
            }

        content.append(str);
        output.append(str);
        output.setCaretPosition(content.length());
        return;
        }

    /**	Set str as prompt. The prompt will only be set to the string after the last occurence of any newline-character.
        */
    public void setPrompt(String str) {
        int n;
        if(PROMPTLENGTH < str.length())
            str = str.substring(str.length() - PROMPTLENGTH);
        if((n = str.lastIndexOf('\n')) != -1)
            str = str.substring(n);
        prompt.setText(str);
        return;
    }
    public String getPrompt() {
        return prompt.getText();
    }
    
    /** This sets the CommandLineListener. Though this method means "add", DON'T call it twice, I don't know what will happen. It registrates Action- and Key-listeners for our input-TextField which do the stuff with the history.
        */
    public synchronized void addCommandLineListener(CommandLineListener l) {
        final CommandLineListener fl = l;
        input.addActionListener(new ActionListener() {
            CommandLineListener l = fl;
            public void actionPerformed(ActionEvent e) {
                if (HISTSQUEEZE && (histtop > 0)) {
                    if(history[histtop-1].equals(e.getActionCommand())) {
                        histtop--;
                    }
                }
                history[histtop] = e.getActionCommand();
                if (history[histtop].trim().length() == 0) {
                    histtop--;
                }

                histtop ++;
                if (histtop >= HISTLENGTH) {
                    System.arraycopy(history, 1, history, 0, HISTLENGTH-1);
                    histtop --;
                }
                histindex = histtop;
                history[histtop] = "";

                input.setText("");
                l.processCommandLine(e.getActionCommand().trim());
            }

        });
        /*		input.addTextListener(new TextListener() {
            public void textValueChanged(TextEvent e) {
                System.out.println(e);
                System.out.println(e.paramString());
            }

        }); */
        input.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                if (e.getKeyCode() == KeyEvent.VK_UP) {
                    if (histindex == histtop) {
                        history[histindex] = input.getText();
                    }
                    if (histindex > 0) {
                        histindex--;
                        input.setText(history[histindex]);
                    }
                } else if (e.getKeyCode()== KeyEvent.VK_DOWN) {
                    if (histindex < histtop) {
                        histindex++;
                        input.setText(history[histindex]);
                    }
                } else if (e.getKeyCode()== KeyEvent.VK_TAB) {
                    for(int i=histtop-1; i>=0; i--) {
                        if(history[i].indexOf(input.getText()) == 0) {
                            histindex = i;
                            input.setText(history[histindex]);
                            input.requestFocus();
                            e.consume();
                        }
                    }
                }
            }

        });
        return;
   }
}
